library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;


entity Zarlink_SPI_Module is
    Port ( 

		-- WriteEn == Start 
		Zarlink_WriteEn			: in  STD_LOGIC; 

		Zarlink_Adress				: in  STD_LOGIC_VECTOR(7 downto 0);
		Zarlink_Data_Out			: in  STD_LOGIC_VECTOR(7 downto 0); 
		Zarlink_RequireValid_Out: out  STD_LOGIC; 
		
		
		Zarlink_Data_In			: out STD_LOGIC_VECTOR(7 downto 0);
		Zarlink_Valid_In			: out STD_LOGIC;
		
		Zarlink_BlockMode			: in  STD_LOGIC; 

		Zarlink_Wait				: out STD_LOGIC; 

		-- Connection to the Zarlink 
		SPI_ZL_CS 					: out STD_LOGIC;
		SPI_SDI 						: out STD_LOGIC;
		SPI_SDO 						: in  STD_LOGIC;
		SPI_CLK 						: out STD_LOGIC;		

		-- The usual ports 
		Reset 						: in  STD_LOGIC;				
		Clk 							: in  STD_LOGIC
	 );
end Zarlink_SPI_Module;

architecture Behavioral of Zarlink_SPI_Module is

	signal Counter 						: unsigned (3 downto 0) := (others => '0'); 

	signal Zarlink_BlockMode_Copy 	: std_logic := '0';
	signal Zarlink_Data_Out_Copy		: STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
	signal Zarlink_Adress_Copy 		: STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
	signal Zarlink_Data_In_Buffer		: STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
	signal Zarlink_Valid_In_Buffer 	: std_logic := '0';
	
	type STATE_TYPE is (
		-- IDLE State 
		STATE_IDLE,
		-- Start the SPI transfer
		--	(second&third start state for holding the CS down for a bit longer before begining with the ride),
		SPI_START_A,
		SPI_START_B,
		SPI_START_C,
		SPI_START_D,
		-- ADRESS
		SPI_ADDRESS_0_A,
		SPI_ADDRESS_0_B,
		SPI_ADDRESS_1_A,
		SPI_ADDRESS_1_B,
		SPI_ADDRESS_2_A,
		SPI_ADDRESS_2_B,
		SPI_ADDRESS_3_A,
		SPI_ADDRESS_3_B,
		SPI_ADDRESS_4_A,
		SPI_ADDRESS_4_B,
		SPI_ADDRESS_5_A,
		SPI_ADDRESS_5_B,
		SPI_ADDRESS_6_A,
		SPI_ADDRESS_6_B,
		SPI_ADDRESS_7_A,
		SPI_ADDRESS_7_B,
		-- DATA
		SPI_DATA_0_A,
		SPI_DATA_0_B,
		SPI_DATA_1_A,
		SPI_DATA_1_B,
		SPI_DATA_2_A,
		SPI_DATA_2_B,
		SPI_DATA_3_A,
		SPI_DATA_3_B,
		SPI_DATA_4_A,
		SPI_DATA_4_B,
		SPI_DATA_5_A,
		SPI_DATA_5_B,
		SPI_DATA_6_A,
		SPI_DATA_6_B,
		SPI_DATA_7_A,
		SPI_DATA_7_B,
		-- Ending the SPI transfer
		SPI_END_A,
		SPI_END_B,
		SPI_END_C,
		SPI_END_D,
		-- Cool down states for holding SPI_CS_B = 1 for > 250 ns (2x 125 = 375ns) 
		-- If neccessary, one of the cooldown states can (possibly) be removed 
		SPI_COOLDOWN_WAIT_A,
		SPI_COOLDOWN_WAIT_B,
		SPI_COOLDOWN_WAIT_C
	);
	signal FSM_STATE : STATE_TYPE;	
begin
	Zarlink_Data_In <= Zarlink_Data_In_Buffer;
	Zarlink_Valid_In <= Zarlink_Valid_In_Buffer;
	
	
	FSM : process(Reset,Clk)
	begin
		if (Reset = '1') then
			FSM_STATE <= STATE_IDLE;

			SPI_ZL_CS <= '1';
			SPI_CLK <= '0';
			SPI_SDI <= '0';
			
			Counter <= (others => '0');
			
			Zarlink_BlockMode_Copy <= '0';
			Zarlink_Adress_Copy <= (others => '0');
			Zarlink_Data_Out_Copy <= (others => '0');
			Zarlink_Data_In_Buffer <= (others => '0');
			Zarlink_Wait <= '0'; 
			
			Zarlink_Valid_In_Buffer <= '0';
			Zarlink_RequireValid_Out <= '0';
			
		elsif rising_edge(Clk) then
		
			-- Default:
			Counter <= Counter;
			SPI_ZL_CS <= '1';
			SPI_CLK <= '0';
			SPI_SDI <= '0';
			
			Zarlink_Wait <= '1';
			Zarlink_Adress_Copy <= Zarlink_Adress_Copy;
			Zarlink_BlockMode_Copy <= Zarlink_BlockMode_Copy;
			Zarlink_Data_Out_Copy <= Zarlink_Data_Out_Copy;

			Zarlink_Valid_In_Buffer <= '0';
			Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer;

			Zarlink_RequireValid_Out <= '0';
			
			case FSM_STATE is
					
				when STATE_IDLE =>
					if (Zarlink_WriteEn = '1') then
						FSM_STATE <= SPI_START_A;
					else
						FSM_STATE <= STATE_IDLE;
						Zarlink_Wait <= '0';
					end if;

					Zarlink_Adress_Copy <= Zarlink_Adress;
					Zarlink_BlockMode_Copy <= Zarlink_BlockMode;
					Zarlink_Data_Out_Copy <= Zarlink_Data_Out;	
					Counter <= (others => '0');					
			
				when SPI_START_A =>
					FSM_STATE <= SPI_START_B;
					SPI_ZL_CS <= '0';
					if (Zarlink_Adress_Copy(0) = '0') AND (Zarlink_BlockMode_Copy = '1') then 
						Zarlink_RequireValid_Out <= '1';
					else
						null;					
					end if;

				when SPI_START_B =>
					FSM_STATE <= SPI_START_C;
					SPI_ZL_CS <= '0';

				when SPI_START_C =>
					FSM_STATE <= SPI_START_D;
					SPI_ZL_CS <= '0';
					
				when SPI_START_D =>
					FSM_STATE <= SPI_ADDRESS_0_A;
					Counter <= (others => '0');
					SPI_ZL_CS <= '0';

-----------------

				when SPI_ADDRESS_0_A =>
					FSM_STATE <= SPI_ADDRESS_0_B;
					SPI_SDI <= Zarlink_Adress_Copy(0);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_0_B =>
					FSM_STATE <= SPI_ADDRESS_1_A;
					SPI_SDI <= Zarlink_Adress_Copy(0);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_1_A =>
					FSM_STATE <= SPI_ADDRESS_1_B;
					SPI_SDI <= Zarlink_Adress_Copy(1);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_1_B =>
					FSM_STATE <= SPI_ADDRESS_2_A;
					SPI_SDI <= Zarlink_Adress_Copy(1);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_2_A =>
					FSM_STATE <= SPI_ADDRESS_2_B;
					SPI_SDI <= Zarlink_Adress_Copy(2);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_2_B =>					
					FSM_STATE <= SPI_ADDRESS_3_A;
					SPI_SDI <= Zarlink_Adress_Copy(2);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_3_A =>
					FSM_STATE <= SPI_ADDRESS_3_B;
					SPI_SDI <= Zarlink_Adress_Copy(3);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_3_B =>
					FSM_STATE <= SPI_ADDRESS_4_A;
					SPI_SDI <= Zarlink_Adress_Copy(3);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_4_A =>
					FSM_STATE <= SPI_ADDRESS_4_B;
					SPI_SDI <= Zarlink_Adress_Copy(4);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_4_B =>
					FSM_STATE <= SPI_ADDRESS_5_A;
					SPI_SDI <= Zarlink_Adress_Copy(4);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_5_A =>
					FSM_STATE <= SPI_ADDRESS_5_B;
					SPI_SDI <= Zarlink_Adress_Copy(5);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_5_B =>
					FSM_STATE <= SPI_ADDRESS_6_A;
					SPI_SDI <= Zarlink_Adress_Copy(5);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_6_A =>
					FSM_STATE <= SPI_ADDRESS_6_B;
					SPI_SDI <= Zarlink_Adress_Copy(6);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_6_B =>
					FSM_STATE <= SPI_ADDRESS_7_A;
					SPI_SDI <= Zarlink_Adress_Copy(6);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
				when SPI_ADDRESS_7_A =>
					FSM_STATE <= SPI_ADDRESS_7_B;
					SPI_SDI <= Zarlink_Adress_Copy(7);
					SPI_ZL_CS <= '0';					
				when SPI_ADDRESS_7_B =>
					FSM_STATE <= SPI_DATA_0_A;
					SPI_SDI <= Zarlink_Adress_Copy(7);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					if (Zarlink_Adress_Copy(0) = '0') AND (Zarlink_BlockMode_Copy = '1') then 
						Zarlink_Data_Out_Copy <= Zarlink_Data_Out;
					else
						null;					
					end if;
					

-----------------

				when SPI_DATA_0_A =>
					FSM_STATE <= SPI_DATA_0_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(0);
					SPI_ZL_CS <= '0';
					if Zarlink_Adress_Copy(0) = '1' then 
						Zarlink_RequireValid_Out <= '0';
					elsif Zarlink_BlockMode_Copy = '0' then
						Zarlink_RequireValid_Out <= '0';
					elsif (Zarlink_BlockMode_Copy = '1') AND (Counter = to_unsigned(13,4)) then
						Zarlink_RequireValid_Out <= '0';
					else
						Zarlink_RequireValid_Out <= '1';					
					end if;
					
				when SPI_DATA_0_B =>
					FSM_STATE <= SPI_DATA_1_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(0);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= "0000000" & SPI_SDO;
				when SPI_DATA_1_A =>
					FSM_STATE <= SPI_DATA_1_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(1);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_1_B =>
					FSM_STATE <= SPI_DATA_2_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(1);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
				when SPI_DATA_2_A =>
					FSM_STATE <= SPI_DATA_2_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(2);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_2_B =>					
					FSM_STATE <= SPI_DATA_3_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(2);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
				when SPI_DATA_3_A =>
					FSM_STATE <= SPI_DATA_3_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(3);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_3_B =>
					FSM_STATE <= SPI_DATA_4_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(3);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
				when SPI_DATA_4_A =>
					FSM_STATE <= SPI_DATA_4_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(4);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_4_B =>
					FSM_STATE <= SPI_DATA_5_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(4);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
				when SPI_DATA_5_A =>
					FSM_STATE <= SPI_DATA_5_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(5);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_5_B =>
					FSM_STATE <= SPI_DATA_6_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(5);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
				when SPI_DATA_6_A =>
					FSM_STATE <= SPI_DATA_6_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(6);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_6_B =>
					FSM_STATE <= SPI_DATA_7_A;
					SPI_SDI <= Zarlink_Data_Out_Copy(6);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
				when SPI_DATA_7_A =>
					FSM_STATE <= SPI_DATA_7_B;
					SPI_SDI <= Zarlink_Data_Out_Copy(7);
					SPI_ZL_CS <= '0';					
				when SPI_DATA_7_B =>
					if (Zarlink_BlockMode_Copy = '1') AND (Counter = to_unsigned(13,4)) then
						FSM_STATE <= SPI_END_A;
						Counter <= (others => '0');
					elsif Zarlink_BlockMode_Copy = '0' then
						FSM_STATE <= SPI_END_A;
						Counter <= (others => '0');
					else
						FSM_STATE <= SPI_DATA_0_A;
						Counter <= Counter + 1;
					end if;
					SPI_SDI <= Zarlink_Data_Out_Copy(7);
					SPI_ZL_CS <= '0';					
					SPI_CLK <= '1';
					Zarlink_Data_In_Buffer <= Zarlink_Data_In_Buffer(6 downto 0) & SPI_SDO;
					Zarlink_Valid_In_Buffer	<= '1';
					Zarlink_Data_Out_Copy <= Zarlink_Data_Out;
-----------------
					
				when SPI_END_A =>
					FSM_STATE <= SPI_END_B;
					SPI_ZL_CS <= '0';

				when SPI_END_B =>
					FSM_STATE <= SPI_END_C;
					SPI_ZL_CS <= '0';
					
				when SPI_END_C =>
					FSM_STATE <= SPI_END_D;
					SPI_ZL_CS <= '0';					
					
				when SPI_END_D =>
					FSM_STATE <= SPI_COOLDOWN_WAIT_A;
					SPI_ZL_CS <= '0';
					
				when SPI_COOLDOWN_WAIT_A =>	 			
						FSM_STATE <= SPI_COOLDOWN_WAIT_B;

				when SPI_COOLDOWN_WAIT_B =>	 			
						FSM_STATE <= SPI_COOLDOWN_WAIT_C;

				when SPI_COOLDOWN_WAIT_C => 
						FSM_STATE <= STATE_IDLE;
					
				when others =>
					FSM_STATE <= STATE_IDLE;
					
			end case;
			
		end if;
	end process FSM;




end Behavioral;

