-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This module analyses the incomming commands and sets the internal
-- registers acordingling.
-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO

library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

-- RHD Register Setting sequence:
-- A direct connection to the RHA via SPI
-- (stops the data acquisition)
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_RHD_COMMAND : std_logic_vector(7 downto 0) := "10101010";
-- First Byte:
-- Bit MASK which of the RHD are meant
-- Second Byte:
-- RHD_COMMAND_DATA(15 downto 8)
-- Third Byte
-- RHD_COMMAND_DATA(7 downto 0)


-- Channel Mask Setting sequence:
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_Channel: std_logic_vector(7 downto 0) := "10101011";
-- 1. Byte Which RHD ? 0 -7
-- 2. Byte Channel Mask 31 downto 24
-- 3. Byte Channel Mask 23 downto 16
-- 4. Byte Channel Mask 15 downto  8
-- 5. Byte Channel Mask  7 downto  0


-- Aux Mask Setting sequence:
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_Aux: std_logic_vector(7 downto 0) := "10001010";
-- First Byte:
-- Which RHD ? 0 -7
-- Second Byte:
-- 3 Bit Mask which Aux Channel should be active

-- Test Mode sequence (Switches RHA Testgenerator on/off):
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_TestMode: std_logic_vector(7 downto 0) := "10101110";
-- First Byte: Input(0) = '1' On, Input(0) = '0' Off,

-- Work mode sequence (switches data aquisition on/off):
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_WORKMODE: std_logic_vector(7 downto 0) := "10111010";
-- First Byte: Input(0) = '1' On, Input(0) = '0' Off,

-- Sample rate sequence (Pause between two samples, measured in a 24 MHz Clock):
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_SampleRate  : std_logic_vector(7 downto 0) := "11101010";
-- First Byte:
-- SampleRate             : std_logic_vector(31 downto 24);
-- Second Byte:
-- SampleRate             : std_logic_vector(23 downto 16);
-- Third Byte:
-- SampleRate             : std_logic_vector(15 downto  8);
-- Fourth Byte:
-- SampleRate             : std_logic_vector( 7 downto  0);

-- New Bitmask sequence:
-- Defines which bit of a sample are collected
-- ------------------------------------
-- constant HeaderByte_1  : std_logic_vector(7 downto 0) := "00110000"; -- 48
-- constant HeaderByte_2  : std_logic_vector(7 downto 0) := "01101001"; -- 105
-- constant Implant_ID    : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
-- constant Command_BITMASK : std_logic_vector(7 downto 0) := "00101010";
-- First Byte:
-- BITMASK_SAMPLE_INTERNAL(15 downto 8)
-- Second Byte:
-- BITMASK_SAMPLE_INTERNAL(7 downto 0)


entity COMMAND_RECEIVER is
  Port (
    -- Parameters
    ChannelMask_00        : out std_logic_vector( 31 downto 0);
    ChannelMask_01        : out std_logic_vector( 31 downto 0);
    ChannelMask_02        : out std_logic_vector( 31 downto 0);
    ChannelMask_03        : out std_logic_vector( 31 downto 0);
    ChannelMask_04        : out std_logic_vector( 31 downto 0);
    ChannelMask_05        : out std_logic_vector( 31 downto 0);
    ChannelMask_06        : out std_logic_vector( 31 downto 0);
    ChannelMask_07        : out std_logic_vector( 31 downto 0);
    ChannelMask_Changed   : out std_logic;

    AuxMask_00            : out std_logic_vector( 2 downto 0);
    AuxMask_01            : out std_logic_vector( 2 downto 0);
    AuxMask_02            : out std_logic_vector( 2 downto 0);
    AuxMask_03            : out std_logic_vector( 2 downto 0);
    AuxMask_04            : out std_logic_vector( 2 downto 0);
    AuxMask_05            : out std_logic_vector( 2 downto 0);
    AuxMask_06            : out std_logic_vector( 2 downto 0);
    AuxMask_07            : out std_logic_vector( 2 downto 0);
    AuxMask_Changed       : out std_logic;

    -- Command to RHD
    RHD_COMMAND_DATA      : out std_logic_vector( 15 downto 0);
    RHD_COMMAND_VALID     : out std_logic_vector(7 downto 0);

    -- We also need a variable clock that is looking on this SampleRate
    SampleRate            : out std_logic_vector(31 downto 0);
    SampleRate_Changed    : out std_logic;

    -- RHD TEST MODE GENATOR (ON/OFF)
    TestMode              : out std_logic;
    TestMode_Changed      : out std_logic;

    -- Work MODE DATA ACQUISITION(ON/OFF)
    WorkMode              : out std_logic;
    WorkMode_Changed      : out std_logic;

    DeviceID              : out std_logic_vector(7 downto 0);

    -- Bitmask for the data selection from the 16 bit samples.
    BITMASK_SAMPLE        : out std_logic_vector(15 downto 0);
    BITMASK_SAMPLE_Changed: out std_logic;

    -- Data from the Zarlink:
    FIFO_OUT_DATA         :  in std_logic_vector(7 downto 0);
    FIFO_OUT_VALID        :  in std_logic;

    HeaderByte_1          :  in std_logic_vector(7 downto 0);
    HeaderByte_2          :  in std_logic_vector(7 downto 0);
    Implant_ID            :  in std_logic_vector(7 downto 0);
    Command_RHD_COMMAND   :  in std_logic_vector(7 downto 0);
    Command_Channel       :  in std_logic_vector(7 downto 0);
    Command_Aux           :  in std_logic_vector(7 downto 0);
    Command_TestMode      :  in std_logic_vector(7 downto 0);
    Command_WORKMODE      :  in std_logic_vector(7 downto 0);
    Command_SampleRate    :  in std_logic_vector(7 downto 0);
    Command_BITMASK       :  in std_logic_vector(7 downto 0);

    -- DEBUG SECTION
    DEBUG_M_FSM0_STATE    : out std_logic_vector(4 downto 0);

    Clk                   :  in std_logic;
    Reset                 :  in std_logic
  );
end COMMAND_RECEIVER;

architecture Behavioral of COMMAND_RECEIVER is


  signal SampleRate_INTERNAL        : std_logic_vector( 31 downto 0) := (others => '0');
  signal BITMASK_SAMPLE_INTERNAL    : std_logic_vector( 15 downto 0) := (others => '0');
  signal WORKMODE_INTERNAL          : std_logic := '0';
  signal TestMode_INTERAL           : std_logic := '0';

  signal ChannelMask_00_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_01_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_02_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_03_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_04_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_05_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_06_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_07_INTERNAL    : std_logic_vector( 31 downto 0) := (others => '0');

  signal AuxMask_00_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_01_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_02_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_03_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_04_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_05_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_06_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');
  signal AuxMask_07_INTERNAL        : std_logic_vector(  2 downto 0) := (others => '0');

  signal TEMP_WHICH_8BIT            : std_logic_vector(  7 downto 0) := (others => '0');
  signal TEMP_WHICH_RHD             : std_logic_vector(  7 downto 0) := (others => '0');
  signal TEMP_WHICH_VALUE           : std_logic_vector(  7 downto 0) := (others => '0');

  signal RHD_COMMAND_HIGHER         : std_logic_vector(  7 downto 0) := (others => '0');
  signal RHD_COMMAND_LOWER          : std_logic_vector(  7 downto 0) := (others => '0');
  signal RHD_COMMAND_VALID_INTERNAL : std_logic_vector(  7 downto 0) := (others => '0');
  signal ChannelMask_Temp           : std_logic_vector( 31 downto 0) := (others => '0');

  type STATE_Type is (
    SM_IDLE,
    SM_HEADER_CHECK_A,
    SM_HEADER_CHECK_B,
    SM_PROCESS,
    SM_COMMAND_SAMPLERATE_A,
    SM_COMMAND_SAMPLERATE_B,
    SM_COMMAND_SAMPLERATE_C,
    SM_COMMAND_SAMPLERATE_D,
    SM_COMMAND_WORKMODE,
    SM_COMMAND_TESTMODE,
    SM_COMMAND_BITMASK_A,
    SM_COMMAND_BITMASK_B,
    SM_COMMAND_CHANNEL_A,
    SM_COMMAND_CHANNEL_DATA_0,
    SM_COMMAND_CHANNEL_DATA_1,
    SM_COMMAND_CHANNEL_DATA_2,
    SM_COMMAND_CHANNEL_DATA_3,
    SM_COMMAND_CHANNEL_Sel,
    SM_COMMAND_AUX_A,
    SM_COMMAND_AUX_B,
    SM_COMMAND_AUX_Sel,
    SM_RHD_COMMAND_A,
    SM_RHD_COMMAND_B,
    SM_RHD_COMMAND_C,
    SM_RHD_COMMAND_D
    );

  signal SPIMACHINE_STATE : STATE_TYPE;

  signal FIFO_OUT_VALID_LAST : std_logic := '0';

begin
  TestMode <= TestMode_INTERAL;

  -- Export internal buffered signals
  ChannelMask_00 <= ChannelMask_00_INTERNAL;
  ChannelMask_01 <= ChannelMask_01_INTERNAL;
  ChannelMask_02 <= ChannelMask_02_INTERNAL;
  ChannelMask_03 <= ChannelMask_03_INTERNAL;
  ChannelMask_04 <= ChannelMask_04_INTERNAL;
  ChannelMask_05 <= ChannelMask_05_INTERNAL;
  ChannelMask_06 <= ChannelMask_06_INTERNAL;
  ChannelMask_07 <= ChannelMask_07_INTERNAL;

  AuxMask_00 <= AuxMask_00_INTERNAL;
  AuxMask_01 <= AuxMask_01_INTERNAL;
  AuxMask_02 <= AuxMask_02_INTERNAL;
  AuxMask_03 <= AuxMask_03_INTERNAL;
  AuxMask_04 <= AuxMask_04_INTERNAL;
  AuxMask_05 <= AuxMask_05_INTERNAL;
  AuxMask_06 <= AuxMask_06_INTERNAL;
  AuxMask_07 <= AuxMask_07_INTERNAL;

  RHD_COMMAND_DATA <= RHD_COMMAND_HIGHER & RHD_COMMAND_LOWER;
  RHD_COMMAND_VALID <= RHD_COMMAND_VALID_INTERNAL;
  SampleRate <= SampleRate_INTERNAL;
  DeviceID <= Implant_ID;

  BITMASK_SAMPLE <= BITMASK_SAMPLE_INTERNAL;

  WORKMODE <= NOT WORKMODE_INTERNAL;

  CMD_RECEIVER : process (Clk,Reset)
  begin
    if Reset = '1' then
      ChannelMask_00_INTERNAL <= (others => '0');
      ChannelMask_01_INTERNAL <= (others => '0');
      ChannelMask_02_INTERNAL <= (others => '0');
      ChannelMask_03_INTERNAL <= (others => '0');
      ChannelMask_04_INTERNAL <= (others => '0');
      ChannelMask_05_INTERNAL <= (others => '0');
      ChannelMask_06_INTERNAL <= (others => '0');
      ChannelMask_07_INTERNAL <= (others => '0');
      ChannelMask_Changed <= '0';

      AuxMask_00_INTERNAL <= (others => '0');
      AuxMask_01_INTERNAL <= (others => '0');
      AuxMask_02_INTERNAL <= (others => '0');
      AuxMask_03_INTERNAL <= (others => '0');
      AuxMask_04_INTERNAL <= (others => '0');
      AuxMask_05_INTERNAL <= (others => '0');
      AuxMask_06_INTERNAL <= (others => '0');
      AuxMask_07_INTERNAL <= (others => '0');
      AuxMask_Changed <= '0';

      SampleRate_INTERNAL <= (others => '0');
      SampleRate_Changed <= '0';

      BITMASK_SAMPLE_INTERNAL <= (others => '1');
      BITMASK_SAMPLE_Changed <= '0';

      TestMode_INTERAL <= '1';
      TestMode_Changed <= '0';

      WORKMODE_INTERNAL <= '1';
      WorkMode_Changed <= '0';

      SPIMACHINE_STATE <= SM_IDLE;
      FIFO_OUT_VALID_LAST <= '0';

      TEMP_WHICH_8BIT <= (others => '0');
      TEMP_WHICH_RHD <= (others => '0');
      TEMP_WHICH_VALUE <= (others => '0');

      RHD_COMMAND_HIGHER <= (others => '0');
      RHD_COMMAND_LOWER <= (others => '0');
      RHD_COMMAND_VALID_INTERNAL <= (others => '0');
      ChannelMask_Temp <= (others => '0');

    elsif rising_edge(Clk) then
      ChannelMask_00_INTERNAL <= ChannelMask_00_INTERNAL;
      ChannelMask_01_INTERNAL <= ChannelMask_01_INTERNAL;
      ChannelMask_02_INTERNAL <= ChannelMask_02_INTERNAL;
      ChannelMask_03_INTERNAL <= ChannelMask_03_INTERNAL;
      ChannelMask_04_INTERNAL <= ChannelMask_04_INTERNAL;
      ChannelMask_05_INTERNAL <= ChannelMask_05_INTERNAL;
      ChannelMask_06_INTERNAL <= ChannelMask_06_INTERNAL;
      ChannelMask_07_INTERNAL <= ChannelMask_07_INTERNAL;
      ChannelMask_Changed <= '0';

      AuxMask_00_INTERNAL <= AuxMask_00_INTERNAL;
      AuxMask_01_INTERNAL <= AuxMask_01_INTERNAL;
      AuxMask_02_INTERNAL <= AuxMask_02_INTERNAL;
      AuxMask_03_INTERNAL <= AuxMask_03_INTERNAL;
      AuxMask_04_INTERNAL <= AuxMask_04_INTERNAL;
      AuxMask_05_INTERNAL <= AuxMask_05_INTERNAL;
      AuxMask_06_INTERNAL <= AuxMask_06_INTERNAL;
      AuxMask_07_INTERNAL <= AuxMask_07_INTERNAL;
      AuxMask_Changed <= '0';

      TestMode_INTERAL <= TestMode_INTERAL;
      TestMode_Changed <= '0';

      WORKMODE_INTERNAL <= WORKMODE_INTERNAL;
      WorkMode_Changed <= '0';

      SampleRate_INTERNAL <= SampleRate_INTERNAL;
      SampleRate_Changed <= '0';

      BITMASK_SAMPLE_INTERNAL <= BITMASK_SAMPLE_INTERNAL;
      BITMASK_SAMPLE_Changed <= '0';

      TEMP_WHICH_8BIT <= TEMP_WHICH_8BIT;
      TEMP_WHICH_RHD <= TEMP_WHICH_RHD;
      TEMP_WHICH_VALUE <= TEMP_WHICH_VALUE;
      ChannelMask_Temp <= ChannelMask_Temp;

      RHD_COMMAND_HIGHER <= RHD_COMMAND_HIGHER;
      RHD_COMMAND_LOWER <= RHD_COMMAND_LOWER;
      RHD_COMMAND_VALID_INTERNAL <= (others => '0');

      FIFO_OUT_VALID_LAST <= FIFO_OUT_VALID;
      -- Failsafe !
      SPIMACHINE_STATE <= SM_IDLE;

      case SPIMACHINE_STATE is
        when SM_IDLE =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = HeaderByte_1) then
            SPIMACHINE_STATE <= SM_HEADER_CHECK_A;
          else
            SPIMACHINE_STATE <= SM_IDLE;
          end if;

        when SM_HEADER_CHECK_A =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = HeaderByte_2) then
            SPIMACHINE_STATE <= SM_HEADER_CHECK_B;
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
          else
            SPIMACHINE_STATE <= SM_HEADER_CHECK_A;
          end if;

        when SM_HEADER_CHECK_B =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Implant_ID) then
            SPIMACHINE_STATE <= SM_PROCESS;
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
          else
            SPIMACHINE_STATE <= SM_HEADER_CHECK_B;
          end if;

        -- Which command is it?
        when SM_PROCESS =>
          WORKMODE_INTERNAL <= '1';
          -- WORKMODE
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_WORKMODE) then
            SPIMACHINE_STATE <= SM_COMMAND_WORKMODE;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- CHANNEL MASK
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_Channel) then
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_A;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- AUX MASK
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_Aux) then
            SPIMACHINE_STATE <= SM_COMMAND_AUX_A;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- RHD COMMAND
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_RHD_COMMAND) then
            SPIMACHINE_STATE <= SM_RHD_COMMAND_A;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- SAMPLERATE
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_SampleRate) then
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_A;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- BITMASK
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_BITMASK) then
            SPIMACHINE_STATE <= SM_COMMAND_BITMASK_A;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- TESTMODE
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0')
           AND (FIFO_OUT_DATA = Command_TestMode) then
            SPIMACHINE_STATE <= SM_COMMAND_TESTMODE;
            if WORKMODE_INTERNAL = '0' then
              WorkMode_Changed <= '1';
            else
              WorkMode_Changed <= '0';
            end if;
          -- UNKNOWN COMMAND
          elsif (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            WorkMode_Changed <= '0';
          else
            SPIMACHINE_STATE <= SM_PROCESS;
            WorkMode_Changed <= '0';
            WORKMODE_INTERNAL <= WORKMODE_INTERNAL;
          end if;
-------------------------------------------------------------------------------
        -- Set work mode
        when SM_COMMAND_WORKMODE =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            WORKMODE_INTERNAL <= NOT FIFO_OUT_DATA(0);
            WorkMode_Changed <= '1';
          else
            SPIMACHINE_STATE <= SM_COMMAND_WORKMODE;
          end if;
-------------------------------------------------------------------------------
        -- Set test mode
        when SM_COMMAND_TESTMODE =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            TestMode_INTERAL <= FIFO_OUT_DATA(0);
            TestMode_Changed <= '1';
          else
            SPIMACHINE_STATE <= SM_COMMAND_TESTMODE;
          end if;
-------------------------------------------------------------------------------
          -- Set Sample Rate
          -- First Byte:
        when SM_COMMAND_SAMPLERATE_A =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_B;
            SampleRate_INTERNAL(31 downto 24) <= FIFO_OUT_DATA(7 downto 0);
          else
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_A;
          end if;

        -- Second Byte:
        when SM_COMMAND_SAMPLERATE_B =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_C;
            SampleRate_INTERNAL(23 downto 16) <= FIFO_OUT_DATA(7 downto 0);
          else
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_B;
          end if;

        -- Third Byte:
        when SM_COMMAND_SAMPLERATE_C =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_D;
            SampleRate_INTERNAL(15 downto 8) <= FIFO_OUT_DATA(7 downto 0);
          else
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_C;
          end if;

        -- Last Byte:
        when SM_COMMAND_SAMPLERATE_D =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            SampleRate_INTERNAL(7 downto 0) <= FIFO_OUT_DATA(7 downto 0);
            SampleRate_Changed <= '1';
          else
            SPIMACHINE_STATE <= SM_COMMAND_SAMPLERATE_D;
          end if;
-------------------------------------------------------------------------------
        -- Set the ADC sample resolution (Lower Bits)
        when SM_COMMAND_BITMASK_A =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_BITMASK_B;
            BITMASK_SAMPLE_INTERNAL(15 downto 8) <= FIFO_OUT_DATA(7 downto 0);
          else
            SPIMACHINE_STATE <= SM_COMMAND_BITMASK_A;
          end if;

        -- Set the ADC sample resolution (Higher Bits)
        when SM_COMMAND_BITMASK_B =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            BITMASK_SAMPLE_INTERNAL(7 downto 0) <= FIFO_OUT_DATA(7 downto 0);
            BITMASK_SAMPLE_Changed <= '1';
          else
            SPIMACHINE_STATE <= SM_COMMAND_BITMASK_B;
          end if;
-------------------------------------------------------------------------------
        -- RHD Commands
        when SM_RHD_COMMAND_A =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_RHD_COMMAND_B;
            TEMP_WHICH_RHD <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_RHD_COMMAND_A;
          end if;

        when SM_RHD_COMMAND_B =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_RHD_COMMAND_C;
            RHD_COMMAND_HIGHER <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_RHD_COMMAND_B;
          end if;

        when SM_RHD_COMMAND_C =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_RHD_COMMAND_D;
            RHD_COMMAND_LOWER <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_RHD_COMMAND_C;
          end if;

        when SM_RHD_COMMAND_D =>
          SPIMACHINE_STATE <= SM_IDLE;
          RHD_COMMAND_VALID_INTERNAL <= TEMP_WHICH_RHD;
-------------------------------------------------------------------------------
        -- CHANNEL MASK
        when SM_COMMAND_CHANNEL_A =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_0;
            TEMP_WHICH_RHD <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_A;
          end if;

        when SM_COMMAND_CHANNEL_DATA_0 =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_1;
            ChannelMask_Temp(31 downto 24) <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_0;
          end if;

        when SM_COMMAND_CHANNEL_DATA_1 =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_2;
            ChannelMask_Temp(23 downto 16) <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_1;
          end if;

        when SM_COMMAND_CHANNEL_DATA_2 =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_3;
            ChannelMask_Temp(15 downto  8) <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_2;
          end if;

        when SM_COMMAND_CHANNEL_DATA_3 =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_Sel;
            ChannelMask_Temp(7 downto 0) <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_CHANNEL_DATA_3;
          end if;

        when SM_COMMAND_CHANNEL_Sel =>
          SPIMACHINE_STATE <= SM_IDLE;
          ChannelMask_Changed <= '1';
          case TEMP_WHICH_RHD is
            when "00000000" =>
              ChannelMask_00_INTERNAL <= ChannelMask_Temp;
            when "00000001" =>
              ChannelMask_01_INTERNAL <= ChannelMask_Temp;
            when "00000010" =>
              ChannelMask_02_INTERNAL <= ChannelMask_Temp;
            when "00000011" =>
              ChannelMask_03_INTERNAL <= ChannelMask_Temp;
            when "00000100" =>
              ChannelMask_04_INTERNAL <= ChannelMask_Temp;
            when "00000101" =>
              ChannelMask_05_INTERNAL <= ChannelMask_Temp;
            when "00000110" =>
              ChannelMask_06_INTERNAL <= ChannelMask_Temp;
            when "00000111" =>
              ChannelMask_07_INTERNAL <= ChannelMask_Temp;
            when others =>
              ChannelMask_Changed <= '0';
          end case;
-------------------------------------------------------------------------------
        -- AUX MASK
        when SM_COMMAND_AUX_A =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_AUX_B;
            TEMP_WHICH_RHD <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_AUX_A;
          end if;

        when SM_COMMAND_AUX_B =>
          if (FIFO_OUT_VALID = '1') AND (FIFO_OUT_VALID_LAST = '0') then
            SPIMACHINE_STATE <= SM_COMMAND_AUX_Sel;
            TEMP_WHICH_VALUE <= FIFO_OUT_DATA;
          else
            SPIMACHINE_STATE <= SM_COMMAND_AUX_B;
          end if;

        when SM_COMMAND_AUX_Sel =>
          SPIMACHINE_STATE <= SM_IDLE;
          AuxMask_Changed <= '1';
          case TEMP_WHICH_RHD is
            when "00000000" =>
              AuxMask_00_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000001" =>
              AuxMask_01_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000010" =>
              AuxMask_02_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000011" =>
              AuxMask_03_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000100" =>
              AuxMask_04_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000101" =>
              AuxMask_05_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000110" =>
              AuxMask_06_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when "00000111" =>
              AuxMask_07_INTERNAL( 2 downto  0) <= TEMP_WHICH_VALUE(2 downto 0);
            when others =>
              AuxMask_Changed <= '0';
          end case;
-------------------------------------------------------------------------------
        when others =>
          WORKMODE_INTERNAL <= '1';
          SPIMACHINE_STATE <= SM_IDLE;
      end case;
    end if;

  end process CMD_RECEIVER;

  -- DEBUGGING FSM
  DEBUG_PROCESS : process (SPIMACHINE_STATE)
  begin
    case SPIMACHINE_STATE is
      when SM_IDLE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 0,5));
      when SM_HEADER_CHECK_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 1,5));
      when SM_HEADER_CHECK_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 2,5));
      when SM_PROCESS =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 3,5));
      when SM_COMMAND_SAMPLERATE_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 4,5));
      when SM_COMMAND_SAMPLERATE_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 5,5));
      when SM_COMMAND_SAMPLERATE_C =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 6,5));
      when SM_COMMAND_SAMPLERATE_D =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 7,5));
      when SM_COMMAND_WORKMODE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 8,5));
      when SM_COMMAND_TESTMODE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 9,5));
      when SM_COMMAND_BITMASK_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(10,5));
      when SM_COMMAND_BITMASK_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(11,5));
      when SM_COMMAND_CHANNEL_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(12,5));
      when SM_COMMAND_CHANNEL_DATA_0 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(13,5));
      when SM_COMMAND_CHANNEL_DATA_1 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(14,5));
      when SM_COMMAND_CHANNEL_DATA_2 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(15,5));
      when SM_COMMAND_CHANNEL_DATA_3 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(16,5));
      when SM_COMMAND_CHANNEL_Sel =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(17,5));
      when SM_COMMAND_AUX_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(18,5));
      when SM_COMMAND_AUX_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(19,5));
      when SM_COMMAND_AUX_Sel =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(20,5));
      when SM_RHD_COMMAND_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(21,5));
      when SM_RHD_COMMAND_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(22,5));
      when SM_RHD_COMMAND_C =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(23,5));
      when SM_RHD_COMMAND_D =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(24,5));
      when others =>
        DEBUG_M_FSM0_STATE <= "11111";
    end case;
  end process DEBUG_PROCESS;
end Behavioral;
