-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This binding module organizes the Zarlink side of the I/O
-- on the highest level.

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO

library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity CONTROL_NEXUS is
  Port (

    -- Connection to the Zarlink
    SPI_ZL_CS             : out std_logic;
    SPI_SDI               : out std_logic;
    SPI_SDO               :  in std_logic;
    SPI_CLK               : out std_logic;

    -- Parameters
    ChannelMask_00        : out std_logic_vector( 31 downto 0);
    ChannelMask_01        : out std_logic_vector( 31 downto 0);
    ChannelMask_02        : out std_logic_vector( 31 downto 0);
    ChannelMask_03        : out std_logic_vector( 31 downto 0);
    ChannelMask_04        : out std_logic_vector( 31 downto 0);
    ChannelMask_05        : out std_logic_vector( 31 downto 0);
    ChannelMask_06        : out std_logic_vector( 31 downto 0);
    ChannelMask_07        : out std_logic_vector( 31 downto 0);
    ChannelMask_Changed   : out std_logic;

    AuxMask_00            : out std_logic_vector( 2 downto 0);
    AuxMask_01            : out std_logic_vector( 2 downto 0);
    AuxMask_02            : out std_logic_vector( 2 downto 0);
    AuxMask_03            : out std_logic_vector( 2 downto 0);
    AuxMask_04            : out std_logic_vector( 2 downto 0);
    AuxMask_05            : out std_logic_vector( 2 downto 0);
    AuxMask_06            : out std_logic_vector( 2 downto 0);
    AuxMask_07            : out std_logic_vector( 2 downto 0);
    AuxMask_Changed       : out std_logic;

    -- Command to RHD
    RHD_COMMAND_DATA      : out std_logic_vector( 15 downto 0);
    RHD_COMMAND_VALID     : out std_logic_vector(7 downto 0);

    -- We also need a variable clock that is looking on this SampleRate
    SampleRate            : out std_logic_vector(31 downto 0);
    SampleRate_Changed    : out std_logic;

    -- RHD TEST MODE GENATOR (ON/OFF)
    TestMode              : out std_logic;
    TestMode_Changed      : out std_logic;

    -- Work MODE DATA ACQUISITION(ON/OFF)
    WorkMode              : out std_logic;
    WorkMode_Changed      : out std_logic;

    DeviceID              : out std_logic_vector(7 downto 0);

    -- Bitmask for the data selection from the 16 bit samples.
    BITMASK_SAMPLE        : out std_logic_vector(15 downto 0);
    BITMASK_SAMPLE_Changed: out std_logic;

    WAIT_PROCESS_PARAMETER:  in std_logic;

    -- Connection to the RHA DATA FIFO...
    -- Data
    FIFO_IN_DATA          :  in std_logic_vector(7 downto 0);
    -- At least 14 x 8 bit are availabe if this is one...
    FIFO_IN_VALID         :  in std_logic;
    -- Need new data
    FIFO_IN_READENABLE    : out std_logic;

    -- Constants defined eleswhere
    HeaderByte_1          :  in std_logic_vector(7 downto 0);
    HeaderByte_2          :  in std_logic_vector(7 downto 0);
    Implant_ID            :  in std_logic_vector(7 downto 0);
    Command_RHD_COMMAND   :  in std_logic_vector(7 downto 0);
    Command_Channel       :  in std_logic_vector(7 downto 0);
    Command_Aux           :  in std_logic_vector(7 downto 0);
    Command_TestMode      :  in std_logic_vector(7 downto 0);
    Command_WORKMODE      :  in std_logic_vector(7 downto 0);
    Command_SampleRate    :  in std_logic_vector(7 downto 0);
    Command_BITMASK       :  in std_logic_vector(7 downto 0);

    -- DEBUG SECTION
    DEBUG_S_COMREC_FSM0_STATE : out std_logic_vector(4 downto 0);
    DEBUG_S_ZARINI_FSM0_STATE : out std_logic_vector(5 downto 0);
    DEBUG_S_ORGANI_FSM0_STATE : out std_logic_vector(3 downto 0);
    DEBUG_S_ZARSPI_FSM0_STATE : out std_logic_vector(5 downto 0);

    -- The usual ports
    Reset                 :  in std_logic;
    Clock_4MHz            :  in std_logic;
    Clk                   :  in std_logic
  );
end CONTROL_NEXUS;

architecture Behavioral of CONTROL_NEXUS is

  signal Zarlink_WriteEn            : STD_LOGIC := '0';
  signal Zarlink_Adress             : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal Zarlink_Data_Out           : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal Zarlink_RequireValid_Out   : STD_LOGIC := '0';
  signal Zarlink_Data_In            : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal Zarlink_Valid_In           : STD_LOGIC := '0';
  signal Zarlink_BlockMode          : STD_LOGIC := '0';
  signal Zarlink_Wait               : STD_LOGIC := '0';

  -- ZarlinkConnectionInit
  signal ZarlinkConnectionInit_DO   : STD_LOGIC := '0';

  -- CheckConnection
  signal CheckConnection_CHECK      : STD_LOGIC := '0';
  signal CheckConnection_OK         : STD_LOGIC := '0';

  -- CheckIngoingBuffer
  signal CheckIngoingBuffer_CHECK   : STD_LOGIC := '0';
  signal CheckIngoingBuffer_OK      : STD_LOGIC := '0';
  signal CheckIngoingBuffer_SIZE    : STD_LOGIC_VECTOR (6 downto 0) := (others => '0');

  -- CheckOutgoingBuffer
  signal CheckOutgoingBuffer_CHECK  : STD_LOGIC := '0';
  signal CheckOutgoingBuffer_OK     : STD_LOGIC := '0';
  signal CheckOutgoingBuffer_SIZE   : STD_LOGIC_VECTOR (6 downto 0) := (others => '0');

  -- GetCommand
  signal GetCommand_GET             : STD_LOGIC := '0';
  signal GetCommand_DATA            : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal GetCommand_VALID           : STD_LOGIC := '0';

  -- Write data to Zarlink
  signal Zarlink_FIFO_WRITE         : STD_LOGIC := '0';
  signal Zarlink_FIFO_DATA          : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal Zarlink_FIFO_FILLED        : STD_LOGIC := '0';

  signal ProcessOngoing             : STD_LOGIC := '0';

  signal WorkMode_INTERNAL          : STD_LOGIC := '0';
  signal WaitSignal_ProcessOngoing  : STD_LOGIC := '0';

begin
  -- Export the internal buffered signals
  WorkMode <= WorkMode_INTERNAL;
  -- Connection to the RHA DATA FIFO...
  -- Data
  Zarlink_FIFO_DATA <=  FIFO_IN_DATA;
  -- At least 14 x 8 bit are availabe if this is one...
  Zarlink_FIFO_FILLED <= FIFO_IN_VALID;
  -- Need new data
  FIFO_IN_READENABLE <= Zarlink_RequireValid_Out;

  -- This module is responsible for interfacing the Zarlink hardware via SPI
  Inst_Zarlink_SPI_Module: entity work.Zarlink_SPI_Module
  PORT MAP(
    Zarlink_WriteEn => Zarlink_WriteEn,
    Zarlink_Adress => Zarlink_Adress,
    Zarlink_Data_Out => Zarlink_Data_Out,
    Zarlink_Data_In => Zarlink_Data_In,
    Zarlink_Valid_In => Zarlink_Valid_In,
    Zarlink_RequireValid_Out => Zarlink_RequireValid_Out,
    Zarlink_BlockMode => Zarlink_BlockMode,
    Zarlink_Wait => Zarlink_Wait,
    SPI_ZL_CS => SPI_ZL_CS,
    SPI_SDI => SPI_SDI,
    SPI_SDO => SPI_SDO,
    SPI_CLK => SPI_CLK,
    DEBUG_M_FSM0_STATE => DEBUG_S_ZARSPI_FSM0_STATE,
    Reset => Reset,
    Clk => Clock_4MHz
  );

  -- This module manages all the commands (connection initialization process, check connection status, read commands,
  -- copy data to the Zarlink, check the rx/tx buffer) concerning the Zarlink
  Inst_ZARLINK_CONNECT_INIT: entity work.ZARLINK_CONNECT_INIT
  PORT MAP(
    ZarlinkConnectionInit_DO => ZarlinkConnectionInit_DO,
    CheckConnection_CHECK => CheckConnection_CHECK,
    CheckConnection_OK => CheckConnection_OK,
    CheckIngoingBuffer_CHECK => CheckIngoingBuffer_CHECK,
    CheckIngoingBuffer_OK => CheckIngoingBuffer_OK,
    CheckIngoingBuffer_SIZE => CheckIngoingBuffer_SIZE,
    CheckOutgoingBuffer_CHECK => CheckOutgoingBuffer_CHECK,
    CheckOutgoingBuffer_OK => CheckOutgoingBuffer_OK,
    CheckOutgoingBuffer_SIZE => CheckOutgoingBuffer_SIZE,
    GetCommand_GET => GetCommand_GET,
    GetCommand_DATA => GetCommand_DATA,
    GetCommand_VALID => GetCommand_VALID,
    Zarlink_FIFO_WRITE => Zarlink_FIFO_WRITE,
    Zarlink_FIFO_DATA => Zarlink_FIFO_DATA,
    Zarlink_WriteEn => Zarlink_WriteEn,
    Zarlink_Adress => Zarlink_Adress,
    Zarlink_Data_Out => Zarlink_Data_Out,
    Zarlink_Data_In => Zarlink_Data_In,
    Zarlink_Valid_In => Zarlink_Valid_In,
    Zarlink_BlockMode => Zarlink_BlockMode,
    Zarlink_Wait => Zarlink_Wait,
    WaitSignal_ProcessOngoing => WaitSignal_ProcessOngoing,
    DEBUG_M_FSM0_STATE => DEBUG_S_ZARINI_FSM0_STATE,
    Reset => Reset,
    Clk => Clk
  );

  -- This module contains a FSM which organizes the behaviour of the implant.
  ProcessOngoing <= (WaitSignal_ProcessOngoing OR WAIT_PROCESS_PARAMETER);

  Inst_ORGANIZER: entity work.ORGANIZER
  PORT MAP(
    CheckConnection_CHECK => CheckConnection_CHECK,
    CheckConnection_OK => CheckConnection_OK,
    ZarlinkConnectionInit_DO => ZarlinkConnectionInit_DO,
    CheckIngoingBuffer_CHECK => CheckIngoingBuffer_CHECK,
    CheckIngoingBuffer_OK => CheckIngoingBuffer_OK,
    CheckIngoingBuffer_SIZE => CheckIngoingBuffer_SIZE,
    CheckOutgoingBuffer_CHECK => CheckOutgoingBuffer_CHECK,
    CheckOutgoingBuffer_OK => CheckOutgoingBuffer_OK,
    CheckOutgoingBuffer_SIZE => CheckOutgoingBuffer_SIZE,
    Zarlink_FIFO_FILLED => Zarlink_FIFO_FILLED,
    Zarlink_FIFO_WRITE => Zarlink_FIFO_WRITE,
    GetCommand_GET => GetCommand_GET,
    ProcessOngoing => ProcessOngoing,
    DEBUG_M_FSM0_STATE => DEBUG_S_ORGANI_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- This module analyzes the incomming data from the Zarlink and reacts to these commands
  Inst_COMMAND_RECEIVER: entity work.COMMAND_RECEIVER
  PORT MAP(
    ChannelMask_00 => ChannelMask_00,
    ChannelMask_01 => ChannelMask_01,
    ChannelMask_02 => ChannelMask_02,
    ChannelMask_03 => ChannelMask_03,
    ChannelMask_04 => ChannelMask_04,
    ChannelMask_05 => ChannelMask_05,
    ChannelMask_06 => ChannelMask_06,
    ChannelMask_07 => ChannelMask_07,
    ChannelMask_Changed => ChannelMask_Changed,
    AuxMask_00 => AuxMask_00,
    AuxMask_01 => AuxMask_01,
    AuxMask_02 => AuxMask_02,
    AuxMask_03 => AuxMask_03,
    AuxMask_04 => AuxMask_04,
    AuxMask_05 => AuxMask_05,
    AuxMask_06 => AuxMask_06,
    AuxMask_07 => AuxMask_07,
    AuxMask_Changed => AuxMask_Changed,
    RHD_COMMAND_DATA => RHD_COMMAND_DATA,
    RHD_COMMAND_VALID => RHD_COMMAND_VALID,
    SampleRate => SampleRate,
    SampleRate_Changed => SampleRate_Changed,
    TestMode => TestMode,
    TestMode_Changed => TestMode_Changed,
    WorkMode => WorkMode_INTERNAL,
    WorkMode_Changed => WorkMode_Changed,
    DeviceID => DeviceID,
    BITMASK_SAMPLE => BITMASK_SAMPLE,
    BITMASK_SAMPLE_Changed	=> BITMASK_SAMPLE_Changed,
    FIFO_OUT_DATA => GetCommand_DATA,
    FIFO_OUT_VALID => GetCommand_VALID,
    HeaderByte_1 => HeaderByte_1,
    HeaderByte_2 => HeaderByte_2,
    Implant_ID => Implant_ID,
    Command_RHD_COMMAND => Command_RHD_COMMAND,
    Command_Channel => Command_Channel,
    Command_Aux => Command_Aux,
    Command_TestMode => Command_TestMode,
    Command_WORKMODE => Command_WORKMODE,
    Command_SampleRate => Command_SampleRate,
    Command_BITMASK => Command_BITMASK,
    DEBUG_M_FSM0_STATE  => DEBUG_S_COMREC_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

end architecture Behavioral;
