-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This binding module organizes the RHD side of the I/O
-- on the highest level.

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO

library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity DATA_ACQUISITION_BLOCK is
  Port (
    -- For the data stream
    RHD_MOSI_OUTSIDE                  : out STD_LOGIC_VECTOR(7 downto 0);
    RHD_MISO_OUTSIDE                  :  in STD_LOGIC_VECTOR(7 downto 0);
    RHD_Clk_OUTSIDE                   : out STD_LOGIC_VECTOR(7 downto 0);
    RHD_CS_OUTSIDE                    : out STD_LOGIC_VECTOR(7 downto 0);

    -- This goes into the Zarlink / FIFO direction:
    FIFO_OUT_DATA                     : out STD_LOGIC_VECTOR(7 downto 0);
    FIFO_OUT_VALID                    : out STD_LOGIC;
    -- We need the FIFO to have (at least)
    -- 280 x 8 bit (2240 bits)
    -- of free space for the next block to be on the safe side.
    ExportNext280ChannelBlock         :  in STD_LOGIC;

    -- Parameters
    ChannelMask_00                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_01                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_02                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_03                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_04                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_05                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_06                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_07                    :  in std_logic_vector( 31 downto 0);
    ChannelMask_Changed               :  in std_logic;

    AuxMask_00                        :  in std_logic_vector( 2 downto 0);
    AuxMask_01                        :  in std_logic_vector( 2 downto 0);
    AuxMask_02                        :  in std_logic_vector( 2 downto 0);
    AuxMask_03                        :  in std_logic_vector( 2 downto 0);
    AuxMask_04                        :  in std_logic_vector( 2 downto 0);
    AuxMask_05                        :  in std_logic_vector( 2 downto 0);
    AuxMask_06                        :  in std_logic_vector( 2 downto 0);
    AuxMask_07                        :  in std_logic_vector( 2 downto 0);
    AuxMask_Changed                   :  in std_logic;

    -- Command to RHD
    RHD_COMMAND_DATA                  :  in std_logic_vector( 15 downto 0);
    RHD_COMMAND_VALID                 :  in std_logic_vector(7 downto 0);

    -- We also need a variable clock that is looking on this SampleRate
    SampleRate                        :  in std_logic_vector(31 downto 0);
    SampleRate_Changed                :  in std_logic;

    -- RHD TEST MODE GENATOR (ON/OFF)
    TestMode                          :  in STD_LOGIC;
    TestMode_Changed                  :  in STD_LOGIC;

    -- Work MODE DATA ACQUISITION(ON/OFF)
    WorkMode                          :  in STD_LOGIC;
    WorkMode_Changed                  :  in STD_LOGIC;

    DeviceID                          :  in std_logic_vector(7 downto 0);

    -- Bitmask for the data selection from the 16 bit samples.
    BITMASK_SAMPLE                    :  in std_logic_vector(15 downto 0);
    BITMASK_SAMPLE_Changed            :  in std_logic;

    -- If we do something with the parameters, stop the further processing of other parameters
    WAIT_PROCESS_PARAMETER            : out std_logic;

    -- Error flags
    ERROR_ALL_OUT                     : out STD_LOGIC;

    Command_RHD_COMMAND               : in  std_logic_vector(7 downto 0);
    Command_Channel                   : in  std_logic_vector(7 downto 0);
    Command_Aux                       : in  std_logic_vector(7 downto 0);
    Command_TestMode                  : in  std_logic_vector(7 downto 0);
    Command_WORKMODE                  : in  std_logic_vector(7 downto 0);
    Command_SampleRate                : in  std_logic_vector(7 downto 0);
    Command_BITMASK                   : in  std_logic_vector(7 downto 0);

    -- DEBUG SECTION
    DEBUG_S_RHDTES_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_RHDTES_FSM1_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_RHD2ZL_FSM0_STATE         : out std_logic_vector(3 downto 0);
    DEBUG_S_RHDARR_FSM0_STATE         : out std_logic_vector(3 downto 0);
    DEBUG_S_DACOL0_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL1_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL2_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL3_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL4_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL5_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL6_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL7_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_P_DAACBL_1                  : out std_logic_vector(8 downto 0);
    DEBUG_P_DAACBL_2                  : out std_logic_vector(8 downto 0);
    Reset                             : in  STD_LOGIC;
    Clk                               : in  STD_LOGIC
  );
end DATA_ACQUISITION_BLOCK;

architecture Behavioral of DATA_ACQUISITION_BLOCK is

  -- The collected output, brocken into single values
  signal DATAOUT_DATA_OUT         : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal DATAOUT_VALID_OUT        : STD_LOGIC := '0';
  signal NewBlock                 : STD_LOGIC := '0';
  signal ReadRequest              : STD_LOGIC := '0';

  signal RHD_MOSI_TEST            : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal RHD_MISO_TEST            : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal RHD_Clk_TEST             : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal RHD_CS_TEST              : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');

  signal RHD_MOSI_ARRAY           : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal RHD_MISO_ARRAY           : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal RHD_Clk_ARRAY            : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal RHD_CS_ARRAY             : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');

  -- Error Flags
  signal ERROR_BUFFEROVERFLOW_INTERNAL : std_logic := '0';

  -- For handling a change in the AUX MASK or CHANNEL MASK
  signal ChannelMask_ReadyForUpdate : std_logic := '0';
  signal AuxMask_ReadyForUpdate     : std_logic := '0';
  signal RHD_COMMAND_ReadyForUpdate : std_logic := '0';
  signal SampleRate_ReadyForUpdate  : std_logic := '0';
  signal TestMode_ReadyForUpdate    : std_logic := '0';
  signal WorkMode_ReadyForUpdate    : std_logic := '0';
  signal BITMASK_SAMPLE_ReadyForUpdate : std_logic := '0';

  signal Parameter_UpdateDone     : std_logic := '0';
  signal WAIT_PROCESS_PARAMETER_INTERNAL : std_logic := '0';

  signal RHD_COMMAND_DATA_Copy    : std_logic_vector( 15 downto 0) := (others => '0');
  signal RHD_COMMAND_VALID_Copy   : std_logic_vector(7 downto 0) := (others => '0');

  signal CommandInfos_Data        : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal CommandInfos_Valid       : STD_LOGIC := '0';
  signal CommandInfos_NextData    : STD_LOGIC := '0';

  --signal DATAIN_DATA_EXT : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  --signal DATAIN_VALID_EXT : STD_LOGIC_VECTOR(7 downto 0)  := (others => '0');
  signal DATAIN_USED_EXT          : STD_LOGIC := '0';
  signal Command_Processed        : STD_LOGIC := '0';
  signal TEST_INFO                : STD_LOGIC := '0';

begin

  WAIT_PROCESS_PARAMETER <= WAIT_PROCESS_PARAMETER_INTERNAL;
  CommandInfos_Valid <= WAIT_PROCESS_PARAMETER_INTERNAL;
  Parameter_UpdateDone <= CommandInfos_NextData;

  -- We want to detect any changes in the parameters
  -- and stop the Command Reciever while we
  -- process the stuff...
  -- We should have time until the check_ingoing buffer status was done
  DEBUG_P_DAACBL_1 <= WAIT_PROCESS_PARAMETER_INTERNAL & RHD_COMMAND_VALID;
  DEBUG_P_DAACBL_2 <= WAIT_PROCESS_PARAMETER_INTERNAL & ChannelMask_Changed
    & AuxMask_Changed & SampleRate_Changed & TestMode_Changed & WorkMode_Changed
    & BITMASK_SAMPLE_Changed & Parameter_UpdateDone & "0";
  CheckChange_PROCESS : process (Reset,Clk)
  begin
    if Reset = '1' then
      WAIT_PROCESS_PARAMETER_INTERNAL <= '0';
    elsif rising_edge(Clk) then
      if    ChannelMask_Changed = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif AuxMask_Changed = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif NOT (RHD_COMMAND_VALID = "00000000") then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif SampleRate_Changed = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif TestMode_Changed = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif WorkMode_Changed = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif BITMASK_SAMPLE_Changed = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '1';
      elsif Parameter_UpdateDone = '1' then
        WAIT_PROCESS_PARAMETER_INTERNAL <= '0';
      else
        WAIT_PROCESS_PARAMETER_INTERNAL <= WAIT_PROCESS_PARAMETER_INTERNAL;
      end if;
    end if;
  end process CheckChange_PROCESS;

  -- Here we set the response message to the ingoing commands
  Comand_Info_Process : process (Reset,ChannelMask_ReadyForUpdate,AuxMask_ReadyForUpdate,
   RHD_COMMAND_ReadyForUpdate, SampleRate_ReadyForUpdate, TestMode_ReadyForUpdate,
   WorkMode_ReadyForUpdate, BITMASK_SAMPLE_ReadyForUpdate, Command_Channel,
   Command_Aux, Command_RHD_COMMAND, Command_SampleRate, Command_TestMode,
   Command_WORKMODE, Command_BITMASK)
  begin
    if Reset = '1' then
      CommandInfos_Data <= (others => '0');
    elsif ChannelMask_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_Channel & "00000000";
    elsif AuxMask_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_Aux & "00000000";
    elsif RHD_COMMAND_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_RHD_COMMAND & "00000000";
    elsif SampleRate_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_SampleRate & "00000000";
    elsif TestMode_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_TestMode & "00000000";
    elsif WorkMode_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_WORKMODE & "00000000";
    elsif BITMASK_SAMPLE_ReadyForUpdate = '1' then
      CommandInfos_Data <= Command_BITMASK & "00000000";
    else
      CommandInfos_Data <= (others => '0');
    end if;
  end process Comand_Info_Process;

  -- Is there an update on the parameters?
  Update_PROCESS : process (Reset,Clk)
  begin
    if Reset = '1' then
      ChannelMask_ReadyForUpdate <= '0';
      AuxMask_ReadyForUpdate <= '0';
      RHD_COMMAND_ReadyForUpdate <= '0';
      SampleRate_ReadyForUpdate <= '0';
      TestMode_ReadyForUpdate <= '0';
      WorkMode_ReadyForUpdate <= '0';
      BITMASK_SAMPLE_ReadyForUpdate <= '0';

      RHD_COMMAND_DATA_Copy <= (others => '0');
      RHD_COMMAND_VALID_Copy <= (others => '0');

    elsif rising_edge(Clk) then
      -- Channel Mask
      ChannelMask_ReadyForUpdate <= ChannelMask_ReadyForUpdate;

      if ChannelMask_Changed = '1' then
        ChannelMask_ReadyForUpdate <= '1';
      elsif Parameter_UpdateDone = '1' then
        ChannelMask_ReadyForUpdate <= '0';
      else
        null;
      end if;

      -- Aux Mask
      AuxMask_ReadyForUpdate <= AuxMask_ReadyForUpdate;

      if AuxMask_Changed = '1' then
        AuxMask_ReadyForUpdate <= '1';
      elsif Parameter_UpdateDone = '1' then
        AuxMask_ReadyForUpdate <= '0';
      else
        null;
      end if;

      -- RHD Command
      RHD_COMMAND_ReadyForUpdate <= RHD_COMMAND_ReadyForUpdate;
      RHD_COMMAND_DATA_Copy <= RHD_COMMAND_DATA_Copy;
      RHD_COMMAND_VALID_Copy <= RHD_COMMAND_VALID_Copy;

      if NOT (RHD_COMMAND_VALID = "00000000") then
        RHD_COMMAND_ReadyForUpdate <= '1';
        RHD_COMMAND_DATA_Copy <= RHD_COMMAND_DATA;
        RHD_COMMAND_VALID_Copy <= RHD_COMMAND_VALID;
      elsif Parameter_UpdateDone = '1' then
        RHD_COMMAND_ReadyForUpdate <= '0';
        RHD_COMMAND_VALID_Copy <= (others => '0');
      else
        null;
      end if;

      -- Sample Rate
      SampleRate_ReadyForUpdate <= SampleRate_ReadyForUpdate;
      if SampleRate_Changed = '1' then
        SampleRate_ReadyForUpdate <= '1';
      elsif Parameter_UpdateDone = '1' then
        SampleRate_ReadyForUpdate <= '0';
      else
        null;
      end if;

      -- Test mode
      TestMode_ReadyForUpdate <= TestMode_ReadyForUpdate;
      if TestMode_Changed = '1' then
        TestMode_ReadyForUpdate <= '1';
      elsif Parameter_UpdateDone = '1' then
        TestMode_ReadyForUpdate <= '0';
      else
        null;
      end if;

      -- Work mode
      WorkMode_ReadyForUpdate <= WorkMode_ReadyForUpdate;
      if WorkMode_Changed = '1' then
        WorkMode_ReadyForUpdate <= '1';
      elsif Parameter_UpdateDone = '1' then
        WorkMode_ReadyForUpdate <= '0';
      else
        null;
      end if;

      -- Bitmask
      BITMASK_SAMPLE_ReadyForUpdate <= BITMASK_SAMPLE_ReadyForUpdate;
      if BITMASK_SAMPLE_Changed = '1' then
        BITMASK_SAMPLE_ReadyForUpdate <= '1';
      elsif Parameter_UpdateDone = '1' then
        BITMASK_SAMPLE_ReadyForUpdate <= '0';
      else
        null;
      end if;

    end if;
  end process Update_PROCESS;

  -- This module joins the output all 8 RHAs ICs and has
  -- sub-modules for handling the SPI connection with the
  -- RHDs.
  Inst_RHD_ARRAY: entity work.RHD_ARRAY
  PORT MAP(
    RHD_MOSI_OUTSIDE => RHD_MOSI_ARRAY,
    RHD_MISO_OUTSIDE => RHD_MISO_ARRAY,
    RHD_Clk_OUTSIDE => RHD_Clk_ARRAY,
    RHD_CS_OUTSIDE => RHD_CS_ARRAY,
    ChannelMask_00 => ChannelMask_00,
    ChannelMask_01 => ChannelMask_01,
    ChannelMask_02 => ChannelMask_02,
    ChannelMask_03 => ChannelMask_03,
    ChannelMask_04 => ChannelMask_04,
    ChannelMask_05 => ChannelMask_05,
    ChannelMask_06 => ChannelMask_06,
    ChannelMask_07 => ChannelMask_07,
    AuxMask_00 => AuxMask_00,
    AuxMask_01 => AuxMask_01,
    AuxMask_02 => AuxMask_02,
    AuxMask_03 => AuxMask_03,
    AuxMask_04 => AuxMask_04,
    AuxMask_05 => AuxMask_05,
    AuxMask_06 => AuxMask_06,
    AuxMask_07 => AuxMask_07,
    DATAOUT_DATA_EXT => DATAOUT_DATA_OUT,
    DATAOUT_VALID_EXT	=> DATAOUT_VALID_OUT,
    SampleRate => SampleRate,
    DATAIN_DATA_EXT => RHD_COMMAND_DATA_Copy,
    DATAIN_VALID_EXT => RHD_COMMAND_VALID_Copy,
    DATAIN_USED_EXT => DATAIN_USED_EXT,
    NewBlock	=> NewBlock,
    Command_Processed	 => Command_Processed,
    ReadRequest	=> ReadRequest,
    ERROR_BUFFEROVERFLOW	 => ERROR_BUFFEROVERFLOW_INTERNAL,
    EnableDataCollection => WorkMode,
    TEST_INFO => TEST_INFO,
    DEBUG_M_FSM0_STATE => DEBUG_S_RHDARR_FSM0_STATE,
    DEBUG_S_DACOL0_FSM0_STATE => DEBUG_S_DACOL0_FSM0_STATE,
    DEBUG_S_DACOL1_FSM0_STATE => DEBUG_S_DACOL1_FSM0_STATE,
    DEBUG_S_DACOL2_FSM0_STATE => DEBUG_S_DACOL2_FSM0_STATE,
    DEBUG_S_DACOL3_FSM0_STATE => DEBUG_S_DACOL3_FSM0_STATE,
    DEBUG_S_DACOL4_FSM0_STATE => DEBUG_S_DACOL4_FSM0_STATE,
    DEBUG_S_DACOL5_FSM0_STATE => DEBUG_S_DACOL5_FSM0_STATE,
    DEBUG_S_DACOL6_FSM0_STATE => DEBUG_S_DACOL6_FSM0_STATE,
    DEBUG_S_DACOL7_FSM0_STATE => DEBUG_S_DACOL7_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  Inst_RHD_TO_ZL_CONVERTER: entity work.RHD_TO_ZL_CONVERTER
  PORT MAP(
    DATAOUT_DATA_OUT => DATAOUT_DATA_OUT,
    DATAOUT_VALID_OUT => DATAOUT_VALID_OUT,
    NewBlock => NewBlock,
    Command_Processed => Command_Processed,
    ReadRequest => ReadRequest,
    FIFO_OUT_DATA => FIFO_OUT_DATA,
    FIFO_OUT_VALID => FIFO_OUT_VALID,
    ERROR_ALL_OUT => ERROR_BUFFEROVERFLOW_INTERNAL,
    DeviceID => DeviceID,
    BITMASK_SAMPLE => BITMASK_SAMPLE,
    CommandInfos_Data	=> CommandInfos_Data,
    CommandInfos_Valid => CommandInfos_Valid,
    CommandInfos_NextData => CommandInfos_NextData,
    ExportNext280ChannelBlock => ExportNext280ChannelBlock,
    DEBUG_M_FSM0_STATE => DEBUG_S_RHD2ZL_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  Inst_RHD_TESTMODULE: entity work.RHD_TESTMODULE
  PORT MAP(
    RHD_MOSI_OUTSIDE => RHD_MOSI_TEST,
    RHD_MISO_OUTSIDE => RHD_MISO_TEST,
    RHD_Clk_OUTSIDE => RHD_Clk_TEST,
    RHD_CS_OUTSIDE => RHD_CS_TEST,
    TEST_INFO => TEST_INFO,
    DEBUG_S_RHDTES_FSM0_STATE => DEBUG_S_RHDTES_FSM0_STATE,
    DEBUG_S_RHDTES_FSM1_STATE => DEBUG_S_RHDTES_FSM1_STATE,
    Clk => Clk,
    Reset => Reset
  );

	-- RHA TESTMODE
	RHD_MOSI_TEST <= RHD_MOSI_ARRAY;
	RHD_Clk_TEST <= RHD_Clk_ARRAY;
	RHD_CS_TEST <= RHD_CS_ARRAY;
	RHD_MOSI_OUTSIDE <= RHD_MOSI_ARRAY;
	RHD_Clk_OUTSIDE <= RHD_Clk_ARRAY;
	RHD_CS_OUTSIDE <= RHD_CS_ARRAY;

	TestMode_Process : process (TestMode,RHD_MISO_TEST,RHD_MISO_OUTSIDE,RHD_MISO_ARRAY)
	begin
		if TestMode = '1' then
			RHD_MISO_ARRAY <= RHD_MISO_TEST;
		else
			RHD_MISO_ARRAY <= RHD_MISO_OUTSIDE;
		end if;
	end process TestMode_Process;

	ERROR_ALL_OUT <= ERROR_BUFFEROVERFLOW_INTERNAL;

end architecture Behavioral;
