-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This module combines all of the sub modules into one. This is the TOP module
-- concerning the functionaly of the implant.

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO
library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity IMPLANT_TOP is
  Port (
    -- Connection to the Zarlink
    SPI_ZL_CS             : out STD_LOGIC;
    SPI_SDI               : out STD_LOGIC;
    SPI_SDO               :  in STD_LOGIC;
    SPI_CLK               : out STD_LOGIC;
    -- For the data stream
    RHD_MOSI              : out STD_LOGIC_VECTOR(7 downto 0);
    RHD_MISO              :  in STD_LOGIC_VECTOR(7 downto 0);
    RHD_Clk               : out STD_LOGIC_VECTOR(7 downto 0);
    RHD_CS                : out STD_LOGIC_VECTOR(7 downto 0);

    -- Error flags
    ERROR_ALL_OUT         : out STD_LOGIC;

    -- DEBUG SECTION
    DEBUG_S_RHDTES_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_RHDTES_FSM1_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_RHD2ZL_FSM0_STATE         : out std_logic_vector(3 downto 0);
    DEBUG_S_RHDARR_FSM0_STATE         : out std_logic_vector(3 downto 0);
    DEBUG_S_DACOL0_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL1_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL2_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL3_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL4_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL5_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL6_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL7_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_CLOCK_ORIGINAL              : out std_logic;
    DEBUG_CLOCK_HALF                  : out std_logic;
    DEBUG_CLOCK_4MHz                  : out std_logic;
    DEBUG_S_COMREC_FSM0_STATE         : out std_logic_vector(4 downto 0);
    DEBUG_S_ZARINI_FSM0_STATE         : out std_logic_vector(5 downto 0);
    DEBUG_S_ORGANI_FSM0_STATE         : out std_logic_vector(3 downto 0);
    DEBUG_S_ZARSPI_FSM0_STATE         : out std_logic_vector(5 downto 0);
    DEBUG_P_DAACBL_1                  : out std_logic_vector(8 downto 0);
    DEBUG_P_DAACBL_2                  : out std_logic_vector(8 downto 0);
    
    -- The usual ports
    Reset                 : in  STD_LOGIC;
    Clk_FAST              : in  STD_LOGIC
  );
end IMPLANT_TOP;

architecture Behavioral of IMPLANT_TOP is

  -- Parameters
  signal ChannelMask_00           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_01           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_02           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_03           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_04           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_05           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_06           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_07           : std_logic_vector( 31 downto 0) := (others => '0');
  signal ChannelMask_Changed      : std_logic := '0';

  signal AuxMask_00               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_01               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_02               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_03               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_04               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_05               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_06               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_07               : std_logic_vector( 2 downto 0) := (others => '0');
  signal AuxMask_Changed          : std_logic := '0';

  -- Command to RHD
  signal RHD_COMMAND_DATA         : std_logic_vector( 15 downto 0) := (others => '0');
  signal RHD_COMMAND_VALID        : std_logic_vector(7 downto 0) := (others => '0');

  -- We also need a variable clock that is looking on this SampleRate
  signal SampleRate               : std_logic_vector(31 downto 0) := (others => '0');
  signal SampleRate_Changed       : std_logic := '0';

  -- RHD TEST MODE GENATOR (ON/OFF)
  signal TestMode                 : STD_LOGIC := '0';
  signal TestMode_Changed         : STD_LOGIC := '0';

  -- Work MODE DATA ACQUISITION(ON/OFF)
  signal WorkMode                 : STD_LOGIC := '0';
  signal NOTWorkMode              : STD_LOGIC := '0';
  signal WorkMode_Changed         : STD_LOGIC := '0';

  signal DeviceID                 : std_logic_vector(7 downto 0) := (others => '0');

  -- Bitmask for the data selection from the 16 bit samples.
  signal BITMASK_SAMPLE           : std_logic_vector(15 downto 0) := (others => '0');
  signal BITMASK_SAMPLE_Changed   : std_logic := '0';

  signal WAIT_PROCESS_PARAMETER   : std_logic := '0';

  -- Connection to the RHA DATA FIFO...
  signal FIFO_IN_DATA             : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal FIFO_IN_VALID            : STD_LOGIC := '0';
  signal FIFO_IN_READENABLE       : STD_LOGIC := '0';

  -- This goes into the Zarlink / FIFO direction:
  signal FIFO_OUT_DATA            : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal FIFO_OUT_VALID           : STD_LOGIC := '0';
  -- We need the FIFO to have (at least)
  -- 21 x 112 bit (2352 its)
  -- of free space for the next block to be on the safe side.
  signal ExportNext280ChannelBlock: STD_LOGIC := '0';


  signal FIFO_AFULL               : STD_LOGIC := '0';
  signal FIFO_AEMPTY              : STD_LOGIC := '0';
  signal FIFO_EMPTY               : STD_LOGIC := '0';
  signal FIFO_FULL                : STD_LOGIC := '0';
  -- Domain Crossing!
  signal FIFO_IN_READENABLE_DX    : STD_LOGIC := '0';
  signal FIFO_IN_READENABLE_LAST  : STD_LOGIC := '0';
  signal dout                     : STD_LOGIC_VECTOR(7 DOWNTO 0) := (others => '0');


  -- Decouple
  signal din_DECOUPLE             : STD_LOGIC_VECTOR(7 DOWNTO 0) := (others => '0');
  signal wr_en_DECOUPLE           : STD_LOGIC := '0';
  signal rd_en_DECOUPLE           : STD_LOGIC := '0';
  signal dout_DECOUPLE            : STD_LOGIC_VECTOR(7 DOWNTO 0) := (others => '0');
  signal full_DECOUPLE            : STD_LOGIC := '0';
  signal empty_DECOUPLE           : STD_LOGIC := '0';
  signal prog_full_DECOUPLE       : STD_LOGIC := '0';
  signal prog_empty_DECOUPLE      : STD_LOGIC := '0';

  signal SPI_ZL_CS_BUFFER         : STD_LOGIC := '0';
  signal SPI_SDI_BUFFER           : STD_LOGIC := '0';
  signal SPI_CLK_BUFFER           : STD_LOGIC := '0';

  constant HeaderByte_1           : std_logic_vector(7 downto 0) := "00110000"; -- 48
  constant HeaderByte_2           : std_logic_vector(7 downto 0) := "01101001"; -- 105
  constant Implant_ID             : std_logic_vector(7 downto 0) := "00110101"; -- Implant ID 53
  constant Command_RHD_COMMAND    : std_logic_vector(7 downto 0) := "10101010";
  constant Command_Channel        : std_logic_vector(7 downto 0) := "10101011";
  constant Command_Aux            : std_logic_vector(7 downto 0) := "10001010";
  constant Command_TestMode       : std_logic_vector(7 downto 0) := "10101110";
  constant Command_WORKMODE       : std_logic_vector(7 downto 0) := "10111010";
  constant Command_SampleRate     : std_logic_vector(7 downto 0) := "11101010";
  constant Command_BITMASK        : std_logic_vector(7 downto 0) := "00101010";


  signal Clk                      : std_logic := '0';
  signal Clock_4MHz               : STD_LOGIC := '0';
  signal Counter_4MHz             : unsigned(1 downto 0) := (others => '0');


begin

  SPI_ZL_CS <= SPI_ZL_CS_BUFFER;
  SPI_SDI <= SPI_SDI_BUFFER;
  SPI_CLK <= SPI_CLK_BUFFER;

  NOTWorkMode <= NOT WorkMode;
  -- 32 channel plus 3 aux channels = 35
  -- 8 RHA x 35 channels = 280 samples = 4480 bits
  -- Plus a 31bit header = 4511 bits = 564 bytes (rounded up)
  -- This results in upto 41 x 14byte blocks = 574 bytes.
  MyFIFO : entity work.FIFO
  PORT MAP (
    rst => Reset,
    wr_clk => Clk,
    rd_clk => Clk,
    din => din_DECOUPLE,
    wr_en => wr_en_DECOUPLE,
    rd_en => rd_en_DECOUPLE,
    dout => dout_DECOUPLE,
    full => full_DECOUPLE,
    empty => empty_DECOUPLE,
    prog_full => prog_full_DECOUPLE,
    prog_empty => prog_empty_DECOUPLE
  );

  DECOUPLE_FIFO : process (Reset,Clk)
  begin
    if Reset = '1' then
      -- Reset
      din_DECOUPLE <= (others => '0');
      wr_en_DECOUPLE <= '0';

      dout <= (others => '0');
      rd_en_DECOUPLE <= '0';

      FIFO_AFULL <= '1';
      FIFO_FULL <= '1';

      FIFO_AEMPTY <= '1';
      FIFO_EMPTY <= '1';

    elsif rising_edge(Clk) then

      -- din_DECOUPLE <= "01010101";
      din_DECOUPLE <= FIFO_OUT_DATA;
      wr_en_DECOUPLE <= FIFO_OUT_VALID;

      dout <= dout_DECOUPLE;
      rd_en_DECOUPLE <= FIFO_IN_READENABLE_DX;

      FIFO_AFULL <= prog_full_DECOUPLE;
      FIFO_FULL <= full_DECOUPLE;

      FIFO_AEMPTY <= prog_empty_DECOUPLE;
      FIFO_EMPTY <= empty_DECOUPLE;
    end if;
  end process DECOUPLE_FIFO;

  -- We need to reverse the bit order otherwise the results is in the wrong order...
  FIFO_IN_DATA(7) <= dout(0);
  FIFO_IN_DATA(6) <= dout(1);
  FIFO_IN_DATA(5) <= dout(2);
  FIFO_IN_DATA(4) <= dout(3);
  FIFO_IN_DATA(3) <= dout(4);
  FIFO_IN_DATA(2) <= dout(5);
  FIFO_IN_DATA(1) <= dout(6);
  FIFO_IN_DATA(0) <= dout(7);

  -- If there is still space then '1'
  ExportNext280ChannelBlock <= NOT FIFO_AFULL;
  -- If there is are at least 14 x 8 bit left, be a '1'
  FIFO_IN_VALID <= NOT FIFO_AEMPTY;

  -- We need to cross the read signal from the
  -- slow Zarlink module to the fast fifo.
  WR_EN_FIFO : process (Clk,Reset)
  begin
    if Reset = '1' then
      FIFO_IN_READENABLE_DX <= '0';
    elsif rising_edge(Clk) then
      FIFO_IN_READENABLE_DX <= '0';
      FIFO_IN_READENABLE_LAST <= FIFO_IN_READENABLE;

      if (FIFO_IN_READENABLE_LAST = '0') AND (FIFO_IN_READENABLE = '1') then
        FIFO_IN_READENABLE_DX <= '1';
      else
        null;
      end if;

    end if;
  end process WR_EN_FIFO;

  Inst_CONTROL_NEXUS: entity work.CONTROL_NEXUS
  PORT MAP(
    SPI_ZL_CS => SPI_ZL_CS_BUFFER,
    SPI_SDI => SPI_SDI_BUFFER,
    SPI_SDO => SPI_SDO,
    SPI_CLK => SPI_CLK_BUFFER,
    ChannelMask_00 => ChannelMask_00,
    ChannelMask_01 => ChannelMask_01,
    ChannelMask_02 => ChannelMask_02,
    ChannelMask_03 => ChannelMask_03,
    ChannelMask_04 => ChannelMask_04,
    ChannelMask_05 => ChannelMask_05,
    ChannelMask_06 => ChannelMask_06,
    ChannelMask_07 => ChannelMask_07,
    ChannelMask_Changed => ChannelMask_Changed,
    AuxMask_00 => AuxMask_00,
    AuxMask_01 => AuxMask_01,
    AuxMask_02 => AuxMask_02,
    AuxMask_03 => AuxMask_03,
    AuxMask_04 => AuxMask_04,
    AuxMask_05 => AuxMask_05,
    AuxMask_06 => AuxMask_06,
    AuxMask_07 => AuxMask_07,
    AuxMask_Changed => AuxMask_Changed,
    RHD_COMMAND_DATA => RHD_COMMAND_DATA,
    RHD_COMMAND_VALID => RHD_COMMAND_VALID,
    SampleRate => SampleRate,
    SampleRate_Changed => SampleRate_Changed,
    TestMode => TestMode,
    TestMode_Changed => TestMode_Changed,
    WorkMode => WorkMode,
    WorkMode_Changed => WorkMode_Changed,
    DeviceID => DeviceID,
    BITMASK_SAMPLE => BITMASK_SAMPLE,
    BITMASK_SAMPLE_Changed	=> BITMASK_SAMPLE_Changed,
    WAIT_PROCESS_PARAMETER => WAIT_PROCESS_PARAMETER,
    FIFO_IN_DATA => FIFO_IN_DATA,
    FIFO_IN_VALID => FIFO_IN_VALID,
    FIFO_IN_READENABLE => FIFO_IN_READENABLE,
    HeaderByte_1 => HeaderByte_1,
    HeaderByte_2 => HeaderByte_2,
    Implant_ID => Implant_ID,
    Command_RHD_COMMAND => Command_RHD_COMMAND,
    Command_Channel => Command_Channel,
    Command_Aux => Command_Aux,
    Command_TestMode => Command_TestMode,
    Command_WORKMODE => Command_WORKMODE,
    Command_SampleRate => Command_SampleRate,
    Command_BITMASK => Command_BITMASK,
    DEBUG_S_COMREC_FSM0_STATE => DEBUG_S_COMREC_FSM0_STATE,
    DEBUG_S_ZARINI_FSM0_STATE => DEBUG_S_ZARINI_FSM0_STATE,
    DEBUG_S_ORGANI_FSM0_STATE => DEBUG_S_ORGANI_FSM0_STATE,
    DEBUG_S_ZARSPI_FSM0_STATE => DEBUG_S_ZARSPI_FSM0_STATE,
    Clock_4MHz => Clock_4MHz,
    Reset => Reset,
    Clk => Clk
  );

  Inst_DATA_ACQUISITION_BLOCK: entity work.DATA_ACQUISITION_BLOCK
  PORT MAP(
    RHD_MOSI_OUTSIDE => RHD_MOSI,
    RHD_MISO_OUTSIDE	=> RHD_MISO,
    RHD_Clk_OUTSIDE => RHD_Clk,
    RHD_CS_OUTSIDE => RHD_CS,
    FIFO_OUT_DATA => FIFO_OUT_DATA,
    FIFO_OUT_VALID => FIFO_OUT_VALID,
    ExportNext280ChannelBlock => ExportNext280ChannelBlock,
    ChannelMask_00 => ChannelMask_00,
    ChannelMask_01 => ChannelMask_01,
    ChannelMask_02 => ChannelMask_02,
    ChannelMask_03 => ChannelMask_03,
    ChannelMask_04 => ChannelMask_04,
    ChannelMask_05 => ChannelMask_05,
    ChannelMask_06 => ChannelMask_06,
    ChannelMask_07 => ChannelMask_07,
    ChannelMask_Changed => ChannelMask_Changed,
    AuxMask_00 => AuxMask_00,
    AuxMask_01 => AuxMask_01,
    AuxMask_02 => AuxMask_02,
    AuxMask_03 => AuxMask_03,
    AuxMask_04 => AuxMask_04,
    AuxMask_05 => AuxMask_05,
    AuxMask_06 => AuxMask_06,
    AuxMask_07 => AuxMask_07,
    AuxMask_Changed => AuxMask_Changed,
    RHD_COMMAND_DATA => RHD_COMMAND_DATA,
    RHD_COMMAND_VALID => RHD_COMMAND_VALID,
    SampleRate => SampleRate,
    SampleRate_Changed => SampleRate_Changed,
    TestMode => TestMode,
    TestMode_Changed => TestMode_Changed,
    WorkMode => WorkMode,
    WorkMode_Changed => WorkMode_Changed,
    DeviceID => DeviceID,
    BITMASK_SAMPLE => BITMASK_SAMPLE,
    BITMASK_SAMPLE_Changed	=> BITMASK_SAMPLE_Changed,
    WAIT_PROCESS_PARAMETER => WAIT_PROCESS_PARAMETER,
    ERROR_ALL_OUT => ERROR_ALL_OUT,
    Command_RHD_COMMAND => Command_RHD_COMMAND,
    Command_Channel => Command_Channel,
    Command_Aux => Command_Aux,
    Command_TestMode => Command_TestMode,
    Command_WORKMODE => Command_WORKMODE,
    Command_SampleRate => Command_SampleRate,
    Command_BITMASK => Command_BITMASK,
    DEBUG_S_RHDTES_FSM0_STATE => DEBUG_S_RHDTES_FSM0_STATE,
    DEBUG_S_RHDTES_FSM1_STATE => DEBUG_S_RHDTES_FSM1_STATE,
    DEBUG_S_RHD2ZL_FSM0_STATE => DEBUG_S_RHD2ZL_FSM0_STATE,
    DEBUG_S_RHDARR_FSM0_STATE => DEBUG_S_RHDARR_FSM0_STATE,
    DEBUG_S_DACOL0_FSM0_STATE => DEBUG_S_DACOL0_FSM0_STATE,
    DEBUG_S_DACOL1_FSM0_STATE => DEBUG_S_DACOL1_FSM0_STATE,
    DEBUG_S_DACOL2_FSM0_STATE => DEBUG_S_DACOL2_FSM0_STATE,
    DEBUG_S_DACOL3_FSM0_STATE => DEBUG_S_DACOL3_FSM0_STATE,
    DEBUG_S_DACOL4_FSM0_STATE => DEBUG_S_DACOL4_FSM0_STATE,
    DEBUG_S_DACOL5_FSM0_STATE => DEBUG_S_DACOL5_FSM0_STATE,
    DEBUG_S_DACOL6_FSM0_STATE => DEBUG_S_DACOL6_FSM0_STATE,
    DEBUG_S_DACOL7_FSM0_STATE => DEBUG_S_DACOL7_FSM0_STATE,
    DEBUG_P_DAACBL_1 => DEBUG_P_DAACBL_1,
    DEBUG_P_DAACBL_2 => DEBUG_P_DAACBL_2,
    Reset => Reset,
    Clk => Clk
  );

   -- We reduce the 24 MHz to a 4 MHz clock
   CLK_4MHZ_PROCESS : process (Reset,Clk_FAST)
   begin
       if Reset = '1' then
           Counter_4MHz <= to_unsigned(0,2);
           Clock_4MHz <= '0';
       elsif rising_edge(Clk_FAST) then
           Clock_4MHz <= Clock_4MHz;
           Counter_4MHz <= Counter_4MHz;
           if std_logic_vector(Counter_4MHz) = "10" then
               Clock_4MHz <= NOT Clock_4MHz;
               Counter_4MHz <= to_unsigned(0,2);
           else
               Counter_4MHz <= Counter_4MHz + 1;
           end if;
       end if;
   end process CLK_4MHZ_PROCESS;

  CLK_2 : process (Clk_FAST,Reset)
  begin
    if Reset = '1' then
      Clk <= '0';
    elsif rising_edge(Clk_FAST) then
      Clk <= NOT Clk;
    end if;
  end process CLK_2;

  -- Export the clocks for debug purposes
  DEBUG_CLOCK_ORIGINAL <= Clk_FAST;
  DEBUG_CLOCK_HALF <= Clk;
  DEBUG_CLOCK_4MHz <= Clock_4MHz;

end architecture Behavioral;
