-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This module organizes the Zarlink side of the I/O. Check and initializes the
-- wireless connection, checks the buffer, etc...

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO

library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;


entity ORGANIZER is
  Port (
    -- CheckConnection
    CheckConnection_CHECK       : out STD_LOGIC;
    CheckConnection_OK          : in  STD_LOGIC;

    -- ZarlinkConnectionInit
    ZarlinkConnectionInit_DO    : out STD_LOGIC;

    -- CheckIngoingBuffer
    CheckIngoingBuffer_CHECK    : out STD_LOGIC;
    CheckIngoingBuffer_OK       : in  STD_LOGIC;
    CheckIngoingBuffer_SIZE     : in  STD_LOGIC_VECTOR (6 downto 0);

    -- CheckOutgoingBuffer
    CheckOutgoingBuffer_CHECK   : out STD_LOGIC;
    CheckOutgoingBuffer_OK      : in  STD_LOGIC;
    CheckOutgoingBuffer_SIZE    : in  STD_LOGIC_VECTOR (6 downto 0);

    -- Copy data from fifo to Zarlink SPI
    Zarlink_FIFO_FILLED         : in  STD_LOGIC;
    Zarlink_FIFO_WRITE          : out STD_LOGIC;

    -- GetCommand
    GetCommand_GET              : out STD_LOGIC;

    ProcessOngoing              : in STD_LOGIC;

    -- DEBUG SECTION
    DEBUG_M_FSM0_STATE          : out std_logic_vector (3 downto 0);

    Clk                         : in  STD_LOGIC;
    Reset                       : in  STD_LOGIC
  );
end ORGANIZER;

architecture Behavioral of ORGANIZER is

  signal CheckConnection_CHECK_INTERNAL                 : std_logic := '0';
  signal ZarlinkConnectionInit_CHECK_INTERNAL           : std_logic := '0';
  signal CheckIngoingBuffer_CHECK_INTERNAL              : std_logic := '0';
  signal CheckOutgoingBuffer_CHECK_INTERNAL             : std_logic := '0';
  signal Zarlink_FIFO_WRITE_INTERNAL                    : std_logic := '0';
  signal LASTSTATEBUFFERCHECK                           : std_logic := '0';

  signal GetCommand_GET_INTERNAL                        : std_logic := '0';

  -- Clock pro sec = 24*10^6; Size of 40 blocks: (14*8*40); bits pro sec = 500000
  -- 24*10^6*(14*8*40)/500000
  constant WaitForXCyles : unsigned (17 downto 0) := to_unsigned(215040,18);
  -- USE THIS FOR THE SIMULATION:
--  constant WaitForXCyles : unsigned (17 downto 0) := to_unsigned(200,18);
  signal WaitForXCyles_Counter : unsigned (17 downto 0) := (others => '0');

  signal CheckOutgoingBuffer_SIZE_COPY : unsigned(6 downto 0) := (others => '0');
  constant MAX_BUFFER_SIZE_OUTGOING : unsigned(6 downto 0) := to_unsigned(60,7);
  -- signal DEBUG_STRING_INTERNAL : std_logic_vector (3 downto 0) := (others => '0');

  type STATE_Type is (
    SM_RESET,
    SM_INIT,
    SM_CHECK_CONNECTION_START,
    SM_CHECK_CONNECTION,
    SM_ZARLINK_INIT_START,
    SM_ZARLINK_INIT,
    SM_CHECK_INGOING_BUFFER_START,
    SM_CHECK_INGOING_BUFFER,
    SM_GET_DATA_FROM_ZARLINK_START,
    SM_GET_DATA_FROM_ZARLINK,
    SM_CHECK_OUTGOING_BUFFER_START,
    SM_CHECK_OUTGOING_BUFFER,
    SM_SEND_DATA_TO_ZARLINK_START,
    SM_SEND_DATA_TO_ZARLINK,
    SM_WAIT_OUTGOING_BUFFER_FULL
  );
  signal SPIMACHINE_STATE : STATE_TYPE;

begin
  CheckConnection_CHECK <= CheckConnection_CHECK_INTERNAL;
  ZarlinkConnectionInit_DO <= ZarlinkConnectionInit_CHECK_INTERNAL;
  CheckIngoingBuffer_CHECK <= CheckIngoingBuffer_CHECK_INTERNAL;
  CheckOutgoingBuffer_CHECK <= CheckOutgoingBuffer_CHECK_INTERNAL;
  Zarlink_FIFO_WRITE <= Zarlink_FIFO_WRITE_INTERNAL;

  GetCommand_GET <= GetCommand_GET_INTERNAL;

  FSM : process (Clk,Reset)
  begin
    if Reset = '1' then
      SPIMACHINE_STATE <= SM_RESET;
      CheckConnection_CHECK_INTERNAL <= '0';
      ZarlinkConnectionInit_CHECK_INTERNAL <= '0';
      CheckIngoingBuffer_CHECK_INTERNAL <= '0';
      CheckOutgoingBuffer_CHECK_INTERNAL <= '0';
      WaitForXCyles_Counter <= (others => '0');
      CheckOutgoingBuffer_SIZE_COPY <= (others => '0');
      Zarlink_FIFO_WRITE_INTERNAL <= '0';
      LASTSTATEBUFFERCHECK <= '0';

      GetCommand_GET_INTERNAL <= '0';

    elsif rising_edge(Clk) then
      -- Fallback
      SPIMACHINE_STATE <= SM_INIT;
      CheckConnection_CHECK_INTERNAL <= '0';
      ZarlinkConnectionInit_CHECK_INTERNAL <= '0';
      CheckIngoingBuffer_CHECK_INTERNAL <= '0';
      CheckOutgoingBuffer_CHECK_INTERNAL <= '0';
      WaitForXCyles_Counter <= (others => '0');
      CheckOutgoingBuffer_SIZE_COPY <= CheckOutgoingBuffer_SIZE_COPY;
      Zarlink_FIFO_WRITE_INTERNAL <= '0';
      LASTSTATEBUFFERCHECK <= '0';

      GetCommand_GET_INTERNAL <= '0';

      case SPIMACHINE_STATE is
        when SM_RESET =>
          SPIMACHINE_STATE <= SM_ZARLINK_INIT_START;

        when SM_INIT =>
          SPIMACHINE_STATE <= SM_CHECK_CONNECTION_START;

        -- Check if there is a Zarlink connection
        when SM_CHECK_CONNECTION_START =>
          if ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_CHECK_CONNECTION_START;
          else
            SPIMACHINE_STATE <= SM_CHECK_CONNECTION;
            CheckConnection_CHECK_INTERNAL <= '1';
          end if;

        -- If there is a connection then check the ingoing buffer
        -- else initiate the Zarlink connection
        when SM_CHECK_CONNECTION =>
          if CheckConnection_CHECK_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_CHECK_CONNECTION;
          elsif ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_CHECK_CONNECTION;
          elsif CheckConnection_OK = '1' then
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER_START;
          else
            SPIMACHINE_STATE <= SM_ZARLINK_INIT_START;
          end if;

        -- If there is no zarlink connection then create one...
        when SM_ZARLINK_INIT_START =>
          if ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_ZARLINK_INIT_START;
          else
            SPIMACHINE_STATE <= SM_ZARLINK_INIT;
            ZarlinkConnectionInit_CHECK_INTERNAL <= '1';
          end if;

        -- If we ran through the zarlink connection module
        -- then check if there is a connection...
        when SM_ZARLINK_INIT =>
          if ZarlinkConnectionInit_CHECK_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_ZARLINK_INIT;
          elsif ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_ZARLINK_INIT;
          else
            SPIMACHINE_STATE <= SM_CHECK_CONNECTION_START;
          end if;

        -- Yes, there is a Zarlink connection. What now?
        -- Are there commands for me?
        -- Check ingoing Zarlink buffer
        when SM_CHECK_INGOING_BUFFER_START =>
          if ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER_START;
          else
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER;
            CheckIngoingBuffer_CHECK_INTERNAL <= '1';
          end if;

        -- And here are the results...
        -- We found something in that ingoing buffer -> let's analyse it
        -- We found nothing and the recording was started -> let's export the measured data
        -- Some thing else happend -> Get's go to the beginning
        when SM_CHECK_INGOING_BUFFER =>
          if CheckIngoingBuffer_CHECK_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER;
          elsif ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER;
          elsif (CheckIngoingBuffer_OK = '1') AND (NOT (CheckIngoingBuffer_SIZE = "0000000")) then
            SPIMACHINE_STATE <= SM_GET_DATA_FROM_ZARLINK_START;
          elsif (CheckIngoingBuffer_OK = '1') AND (CheckIngoingBuffer_SIZE = "0000000") then
            SPIMACHINE_STATE <= SM_CHECK_OUTGOING_BUFFER_START;
          else
            SPIMACHINE_STATE <= SM_INIT;
          end if;

        -- If there are commands from the outside then get them and process them
        when SM_GET_DATA_FROM_ZARLINK_START =>
          if ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_GET_DATA_FROM_ZARLINK_START;
          else
            SPIMACHINE_STATE <= SM_GET_DATA_FROM_ZARLINK;
            GetCommand_GET_INTERNAL <= '1';
          end if;

        when SM_GET_DATA_FROM_ZARLINK =>
        if GetCommand_GET_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_GET_DATA_FROM_ZARLINK;
          elsif ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_GET_DATA_FROM_ZARLINK;
          else
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER_START;
          end if;

        -- Is there space in the outgoing buffer?
        when SM_CHECK_OUTGOING_BUFFER_START =>
          if ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_CHECK_OUTGOING_BUFFER_START;
          else
            SPIMACHINE_STATE <= SM_CHECK_OUTGOING_BUFFER;
            CheckOutgoingBuffer_CHECK_INTERNAL <= '1';
          end if;

        -- Is there still space left in the outgoing buffer?
        -- Lets get the actual fill status
        when SM_CHECK_OUTGOING_BUFFER =>
          if CheckOutgoingBuffer_CHECK_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_CHECK_OUTGOING_BUFFER;
          elsif ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_CHECK_OUTGOING_BUFFER;
          elsif (CheckOutgoingBuffer_OK = '1') then
            CheckOutgoingBuffer_SIZE_COPY <= unsigned(CheckOutgoingBuffer_SIZE);
            SPIMACHINE_STATE <= SM_SEND_DATA_TO_ZARLINK_START;
          else
            SPIMACHINE_STATE <= SM_INIT;
          end if;
          LASTSTATEBUFFERCHECK <= '1';

        -- If there is still space, then copy more
        when SM_SEND_DATA_TO_ZARLINK_START =>
          -- Full after checking buffer. Thus wait and check connection...
          if (CheckOutgoingBuffer_SIZE_COPY = MAX_BUFFER_SIZE_OUTGOING) AND (LASTSTATEBUFFERCHECK = '1') then
            SPIMACHINE_STATE <= SM_WAIT_OUTGOING_BUFFER_FULL;
            -- Full after writing connection. Check for incomming messages and then continue
          elsif (CheckOutgoingBuffer_SIZE_COPY = MAX_BUFFER_SIZE_OUTGOING) AND (LASTSTATEBUFFERCHECK = '0') then
            SPIMACHINE_STATE <= SM_CHECK_INGOING_BUFFER_START;
          elsif CheckOutgoingBuffer_SIZE_COPY > MAX_BUFFER_SIZE_OUTGOING then
            SPIMACHINE_STATE <= SM_WAIT_OUTGOING_BUFFER_FULL;
          elsif (Zarlink_FIFO_FILLED = '1') AND (ProcessOngoing = '0') then
            SPIMACHINE_STATE <= SM_SEND_DATA_TO_ZARLINK;
            Zarlink_FIFO_WRITE_INTERNAL <= '1';
          elsif (Zarlink_FIFO_FILLED = '1') AND (ProcessOngoing = '1') then
            SPIMACHINE_STATE <= SM_SEND_DATA_TO_ZARLINK_START;
          else
            SPIMACHINE_STATE <= SM_INIT;
          end if;

        when SM_SEND_DATA_TO_ZARLINK =>
          if Zarlink_FIFO_WRITE_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_SEND_DATA_TO_ZARLINK;
          elsif ProcessOngoing = '1' then
            SPIMACHINE_STATE <= SM_SEND_DATA_TO_ZARLINK;
          else
            CheckOutgoingBuffer_SIZE_COPY <= CheckOutgoingBuffer_SIZE_COPY + 1;
            SPIMACHINE_STATE <= SM_SEND_DATA_TO_ZARLINK_START;
          end if;

        -- If buffer is full then wait for a given time and then
        -- check the connection and co. Until the export of data
        -- can be continued...
        when SM_WAIT_OUTGOING_BUFFER_FULL =>
          if WaitForXCyles_Counter = WaitForXCyles then
            SPIMACHINE_STATE <= SM_INIT;
          else
            WaitForXCyles_Counter <= WaitForXCyles_Counter + 1;
            SPIMACHINE_STATE <= SM_WAIT_OUTGOING_BUFFER_FULL;
          end if;

        when others =>
          null;
      end case;
    end if;
  end process FSM;

  -- DEBUGGING FSM
  DEBUG_PROCESS : process (SPIMACHINE_STATE)
  begin
    case SPIMACHINE_STATE is
      when SM_RESET =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 0,4));
      when SM_INIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 1,4));
      when SM_CHECK_CONNECTION_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 2,4));
      when SM_CHECK_CONNECTION =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 3,4));
      when SM_ZARLINK_INIT_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 4,4));
      when SM_ZARLINK_INIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 5,4));
      when SM_CHECK_INGOING_BUFFER_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 6,4));
      when SM_CHECK_INGOING_BUFFER =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 7,4));
      when SM_GET_DATA_FROM_ZARLINK_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 8,4));
      when SM_GET_DATA_FROM_ZARLINK =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 9,4));
      when SM_CHECK_OUTGOING_BUFFER_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(10,4));
      when SM_CHECK_OUTGOING_BUFFER =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(11,4));
      when SM_SEND_DATA_TO_ZARLINK_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(12,4));
      when SM_SEND_DATA_TO_ZARLINK =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(13,4));
      when SM_WAIT_OUTGOING_BUFFER_FULL =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(14,4));
      when others =>
        DEBUG_M_FSM0_STATE <= "1111";
      end case;
  end process DEBUG_PROCESS;
end Behavioral;
