-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This module operates the RHD statemachines and collects from there data.
-- Then the data is cut into single words and send out.

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO
library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity RHD_ARRAY is
  Port (
    -- For the data stream
    RHD_MOSI_OUTSIDE      : out STD_LOGIC_VECTOR(7 downto 0);
    RHD_MISO_OUTSIDE      :  in STD_LOGIC_VECTOR(7 downto 0);
    RHD_Clk_OUTSIDE       : out STD_LOGIC_VECTOR(7 downto 0);
    RHD_CS_OUTSIDE        : out STD_LOGIC_VECTOR(7 downto 0);

    ChannelMask_00        :  in std_logic_vector( 31 downto 0);
    ChannelMask_01        :  in std_logic_vector( 31 downto 0);
    ChannelMask_02        :  in std_logic_vector( 31 downto 0);
    ChannelMask_03        :  in std_logic_vector( 31 downto 0);
    ChannelMask_04        :  in std_logic_vector( 31 downto 0);
    ChannelMask_05        :  in std_logic_vector( 31 downto 0);
    ChannelMask_06        :  in std_logic_vector( 31 downto 0);
    ChannelMask_07        :  in std_logic_vector( 31 downto 0);

    AuxMask_00            :  in std_logic_vector( 2 downto 0);
    AuxMask_01            :  in std_logic_vector( 2 downto 0);
    AuxMask_02            :  in std_logic_vector( 2 downto 0);
    AuxMask_03            :  in std_logic_vector( 2 downto 0);
    AuxMask_04            :  in std_logic_vector( 2 downto 0);
    AuxMask_05            :  in std_logic_vector( 2 downto 0);
    AuxMask_06            :  in std_logic_vector( 2 downto 0);
    AuxMask_07            :  in std_logic_vector( 2 downto 0);

    -- The collected output, brocken into single values
    DATAOUT_DATA_EXT      : out STD_LOGIC_VECTOR(15 downto 0);
    -- VALID is checking if this is from an active RHA channel (according to the
    -- channel mask).
    DATAOUT_VALID_EXT     : out STD_LOGIC;

    SampleRate            :  in std_logic_vector(31 downto 0);

    -- The ports for sending the data out
    DATAIN_DATA_EXT       :  in STD_LOGIC_VECTOR(15 downto 0);
    DATAIN_VALID_EXT      :  in STD_LOGIC_VECTOR(7 downto 0);
    DATAIN_USED_EXT       : out STD_LOGIC;

    NewBlock              : out STD_LOGIC;
    Command_Processed     : out STD_LOGIC;

    ReadRequest           :  in STD_LOGIC;
    TEST_INFO             : out STD_LOGIC;

    -- Was the readout too slow?
    ERROR_BUFFEROVERFLOW  : out STD_LOGIC;

    EnableDataCollection  :  in STD_LOGIC;

    -- DEBUG SECTION
    DEBUG_M_FSM0_STATE         : out std_logic_vector(3 downto 0);
    DEBUG_S_DACOL0_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL1_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL2_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL3_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL4_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL5_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL6_FSM0_STATE  : out std_logic_vector(4 downto 0);
    DEBUG_S_DACOL7_FSM0_STATE  : out std_logic_vector(4 downto 0);

    Clk                   :  in STD_LOGIC;
    Reset                 :  in STD_LOGIC

  );
end RHD_ARRAY;

architecture Behavioral of RHD_ARRAY is

  -- Buffering of the data
  signal DATAOUT_DATA_INTERNAL      : STD_LOGIC_VECTOR(127 downto 0) := (others => '0');
  signal DATAOUT_VALIDSTRING        : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal BufferOverflow_INTERNAL    : STD_LOGIC := '0';

  -- Connections to the RHD SPI machines
  signal DATAOUT_DATA               : STD_LOGIC_VECTOR(127 downto 0) := (others => '0');
  signal DATAOUT_VALID              : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal DATAIN_DATA                : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal DATAIN_VALID               : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal DATAIN_USED                : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal END_PAUSE                  : std_logic := '0';

  -- Extended Channel Mask (Aux & Channel Mask)
  signal ChannelMask_Sub_00         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_01         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_02         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_03         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_04         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_05         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_06         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');
  signal ChannelMask_Sub_07         : STD_LOGIC_VECTOR(34 downto 0) := (others => '0');

  -- Counting up the requested channels
  signal Channel_Counter            : unsigned (5 downto 0) := (others => '0');
  signal DATAIN_USED_LAST           : std_logic_vector(7 downto 0) := (others => '0');

  -- Counters for controlling ENDPAUSE
  signal SampleRate_Copy            : unsigned(31 downto 0) := (others => '0');
  signal SampleRate_Counter         : unsigned(31 downto 0) := (others => '0');
  constant SampleRate_SmallCounter_Max  : unsigned(4 downto 0) := to_unsigned(24,5);
  signal SampleRate_SmallCounter    : unsigned(4 downto 0) := (others => '0');

  -- Information about NewBlocks (Channel Counter = 0) and Commands are
  -- synchronized to the output
  signal NewBlock_IN                : STD_LOGIC := '0';
  signal NewBlock_OUT               : STD_LOGIC := '0';
  signal Command_IN                 : STD_LOGIC := '0';
  signal Command_OUT                : STD_LOGIC := '0';
  signal EnableDataCollection_Copy  : STD_LOGIC := '0';

  -- FSM for the EndPause Process
  type STATE_Type_SR is (
    SM_SMALL,
    SM_LONG
  );
  signal SPIMACHINE_STATE_SR : STATE_TYPE_SR;

  -- State Machine that controls requesting the data from the channels
  type STATE_Type is (
    SM_COMMAND,
    SM_COUNT
  );
  signal SPIMACHINE_STATE : STATE_TYPE;

begin
  ERROR_BUFFEROVERFLOW <= BufferOverflow_INTERNAL;

  -- 0
  RHD0 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(0),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(0),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(0),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(0),
    DATAOUT_DATA => DATAOUT_DATA(15 downto 0),
    DATAOUT_VALID => DATAOUT_VALID(0),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(0),
    DATAIN_USED => DATAIN_USED(0),
    END_PAUSE => END_PAUSE,
    NewBlock_IN	=> NewBlock_IN,
    NewBlock_OUT => NewBlock_OUT,
    Command_IN	=> Command_IN,
    Command_OUT	=> Command_OUT,
    TEST_INFO => TEST_INFO,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL0_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 1
  RHD1 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(1),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(1),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(1),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(1),
    DATAOUT_DATA => DATAOUT_DATA(31 downto 16),
    DATAOUT_VALID => DATAOUT_VALID(1),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(1),
    DATAIN_USED => DATAIN_USED(1),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL1_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 2
  RHD2 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(2),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(2),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(2),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(2),
    DATAOUT_DATA => DATAOUT_DATA(47 downto 32),
    DATAOUT_VALID => DATAOUT_VALID(2),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(2),
    DATAIN_USED => DATAIN_USED(2),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL2_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 3
  RHD3 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(3),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(3),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(3),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(3),
    DATAOUT_DATA => DATAOUT_DATA(63 downto 48),
    DATAOUT_VALID => DATAOUT_VALID(3),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(3),
    DATAIN_USED => DATAIN_USED(3),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL3_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 4
  RHD4 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(4),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(4),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(4),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(4),
    DATAOUT_DATA => DATAOUT_DATA(79 downto 64),
    DATAOUT_VALID => DATAOUT_VALID(4),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(4),
    DATAIN_USED => DATAIN_USED(4),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL4_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 5
  RHD5 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(5),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(5),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(5),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(5),
    DATAOUT_DATA => DATAOUT_DATA(95 downto 80),
    DATAOUT_VALID => DATAOUT_VALID(5),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(5),
    DATAIN_USED => DATAIN_USED(5),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL5_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 6
  RHD6 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(6),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(6),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(6),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(6),
    DATAOUT_DATA => DATAOUT_DATA(111 downto 96),
    DATAOUT_VALID => DATAOUT_VALID(6),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(6),
    DATAIN_USED => DATAIN_USED(6),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL6_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );

  -- 7
  RHD7 : entity work.RHD_DATA_COLLECTOR
  Port MAP (
    RHD_MOSI_OUTSIDE => RHD_MOSI_OUTSIDE(7),
    RHD_MISO_OUTSIDE => RHD_MISO_OUTSIDE(7),
    RHD_Clk_OUTSIDE => RHD_Clk_OUTSIDE(7),
    RHD_CS_OUTSIDE => RHD_CS_OUTSIDE(7),
    DATAOUT_DATA => DATAOUT_DATA(127 downto 112),
    DATAOUT_VALID => DATAOUT_VALID(7),
    DATAIN_DATA => DATAIN_DATA,
    DATAIN_VALID => DATAIN_VALID(7),
    DATAIN_USED => DATAIN_USED(7),
    NewBlock_IN	=> '0',
    NewBlock_OUT => open,
    Command_IN	=> '0',
    Command_OUT	=> open,
    END_PAUSE => END_PAUSE,
    TEST_INFO => open,
    DEBUG_M_FSM0_STATE => DEBUG_S_DACOL7_FSM0_STATE,
    Clk => Clk,
    Reset => Reset
  );
  ---------------------------------------------------------------------------------------------------------
  REQUEST_PROCESS : process (Clk,Reset)
  begin
    if Reset = '1' then
      DATAIN_USED_EXT <= '0';
      SPIMACHINE_STATE <= SM_COMMAND;
      Channel_Counter <= (others => '0');
      NewBlock_IN <= '0';
      Command_IN <= '0';
      DATAIN_USED_LAST <= (others => '0');

      -- This is what we write to the SPI machines
      DATAIN_DATA <= (others => '0');
      DATAIN_VALID <= (others => '0');

      ChannelMask_Sub_00 <= (others => '0');
      ChannelMask_Sub_01 <= (others => '0');
      ChannelMask_Sub_02 <= (others => '0');
      ChannelMask_Sub_03 <= (others => '0');
      ChannelMask_Sub_04 <= (others => '0');
      ChannelMask_Sub_05 <= (others => '0');
      ChannelMask_Sub_06 <= (others => '0');
      ChannelMask_Sub_07 <= (others => '0');
      EnableDataCollection_Copy <= '0';
    elsif rising_edge(Clk) then
      -- Default:
      DATAIN_USED_EXT <= '0';
      SPIMACHINE_STATE <= SM_COMMAND;
      Channel_Counter <= Channel_Counter;
      NewBlock_IN <= NewBlock_IN;
      Command_IN <= Command_IN;
      DATAIN_USED_LAST <= DATAIN_USED;

      -- To the SPI machines
      DATAIN_DATA <= DATAIN_DATA;
      DATAIN_VALID <= DATAIN_VALID;

      -- Keep the Channel Masks alive
      ChannelMask_Sub_00 <= ChannelMask_Sub_00;
      ChannelMask_Sub_01 <= ChannelMask_Sub_01;
      ChannelMask_Sub_02 <= ChannelMask_Sub_02;
      ChannelMask_Sub_03 <= ChannelMask_Sub_03;
      ChannelMask_Sub_04 <= ChannelMask_Sub_04;
      ChannelMask_Sub_05 <= ChannelMask_Sub_05;
      ChannelMask_Sub_06 <= ChannelMask_Sub_06;
      ChannelMask_Sub_07 <= ChannelMask_Sub_07;
      EnableDataCollection_Copy <= EnableDataCollection_Copy;

      case SPIMACHINE_STATE is
        when SM_COMMAND =>
          -- We have a command
          NewBlock_IN <= '0';
          Command_IN <= '1';

          -- If there is new imput required than we deliver it.
          if (DATAIN_USED_LAST = "00000000") AND (NOT(DATAIN_USED = "00000000")) then
            SPIMACHINE_STATE <= SM_COUNT;
            NewBlock_IN <= EnableDataCollection;
            Command_IN <= '0';
          else
            SPIMACHINE_STATE <= SM_COMMAND;
          end if;
          -- We copy an external command to the RHD SPIs
          DATAIN_DATA <= DATAIN_DATA_EXT;
          DATAIN_VALID <= DATAIN_VALID_EXT;


          -- Dont' request a new command...
          DATAIN_USED_EXT <= '0';

          -- We want to restart the channel counter
          Channel_Counter <= (others => '0');

          -- Here we compose the extended channel mask
          ChannelMask_Sub_00 <= AuxMask_00 & ChannelMask_00;
          ChannelMask_Sub_01 <= AuxMask_01 & ChannelMask_01;
          ChannelMask_Sub_02 <= AuxMask_02 & ChannelMask_02;
          ChannelMask_Sub_03 <= AuxMask_03 & ChannelMask_03;
          ChannelMask_Sub_04 <= AuxMask_04 & ChannelMask_04;
          ChannelMask_Sub_05 <= AuxMask_05 & ChannelMask_05;
          ChannelMask_Sub_06 <= AuxMask_06 & ChannelMask_06;
          ChannelMask_Sub_07 <= AuxMask_07 & ChannelMask_07;
          EnableDataCollection_Copy <= EnableDataCollection;

        when SM_COUNT =>

          Command_IN <= '0';
          -- If there is new imput required than we deliver it.
          -- This is the first of many channels. But this informs the later modules that a new
          -- set of channels starts
          if (DATAIN_USED_LAST = "00000000") AND (NOT(DATAIN_USED = "00000000"))
           AND (Channel_Counter = "000000") then
            SPIMACHINE_STATE <= SM_COUNT;
            Channel_Counter <= Channel_Counter + 1;
            NewBlock_IN <= '0';
            -- One channel was used. Make next one ready.
            ChannelMask_Sub_00 <= "0" & ChannelMask_Sub_00(34 downto 1);
            ChannelMask_Sub_01 <= "0" & ChannelMask_Sub_01(34 downto 1);
            ChannelMask_Sub_02 <= "0" & ChannelMask_Sub_02(34 downto 1);
            ChannelMask_Sub_03 <= "0" & ChannelMask_Sub_03(34 downto 1);
            ChannelMask_Sub_04 <= "0" & ChannelMask_Sub_04(34 downto 1);
            ChannelMask_Sub_05 <= "0" & ChannelMask_Sub_05(34 downto 1);
            ChannelMask_Sub_06 <= "0" & ChannelMask_Sub_06(34 downto 1);
            ChannelMask_Sub_07 <= "0" & ChannelMask_Sub_07(34 downto 1);
            -- We went through all possile channels, now we check if there are new commands to process
          elsif (DATAIN_USED_LAST = "00000000") AND (NOT(DATAIN_USED = "00000000"))
           AND (Channel_Counter = to_unsigned(34,6)) then
            SPIMACHINE_STATE <= SM_COMMAND;
            -- We don't need to update the channel mask because we don't use it anymore
            -- The normal increment to the next channel (no bounderies)
          elsif (DATAIN_USED_LAST = "00000000") AND (NOT(DATAIN_USED = "00000000")) then
            SPIMACHINE_STATE <= SM_COUNT;
            Channel_Counter <= Channel_Counter + 1;
            -- One channel was used. Make next one ready.
            ChannelMask_Sub_00 <= "0" & ChannelMask_Sub_00(34 downto 1);
            ChannelMask_Sub_01 <= "0" & ChannelMask_Sub_01(34 downto 1);
            ChannelMask_Sub_02 <= "0" & ChannelMask_Sub_02(34 downto 1);
            ChannelMask_Sub_03 <= "0" & ChannelMask_Sub_03(34 downto 1);
            ChannelMask_Sub_04 <= "0" & ChannelMask_Sub_04(34 downto 1);
            ChannelMask_Sub_05 <= "0" & ChannelMask_Sub_05(34 downto 1);
            ChannelMask_Sub_06 <= "0" & ChannelMask_Sub_06(34 downto 1);
            ChannelMask_Sub_07 <= "0" & ChannelMask_Sub_07(34 downto 1);
          else
            SPIMACHINE_STATE <= SM_COUNT;
          end if;
          -- We used the old comand...
          DATAIN_USED_EXT <= '1';

          -- We use the channel mask to request only the channels which are active.
          -- Otherwise we do nothing.
          if EnableDataCollection_Copy = '1' then
            DATAIN_VALID <= ChannelMask_Sub_07(0) & ChannelMask_Sub_06(0) & ChannelMask_Sub_05(0)
            & ChannelMask_Sub_04(0) & ChannelMask_Sub_03(0) & ChannelMask_Sub_02(0)
            & ChannelMask_Sub_01(0) & ChannelMask_Sub_00(0);
          else
            DATAIN_VALID <= "00000000";
          end if;

          -- This needs one clock cylce to be correct but we have about 20 cycles time.
          -- Thus no problem.
          DATAIN_DATA <= "00" & std_logic_vector(Channel_Counter) & "00000000";

          when others =>
          SPIMACHINE_STATE <= SM_COMMAND;
      end case;
    end if;
  end process REQUEST_PROCESS;
------------------------------------------------------------------------------
  -- We wait until a user defined time (SampleRate, which is given in number of
  -- 24MHz Clock cycles) plus the minimum required time for
  -- processing the 16 bit SPI transfer.
  -- This allows to be SampleRate set to zero without any problems...
  SLOW_SAMPLING_DOWN_PROCESS : process (Clk,Reset)
  begin
    if Reset = '1' then
      END_PAUSE <= '0';
      SPIMACHINE_STATE_SR <= SM_SMALL;
      SampleRate_Copy <= (others => '0');
      SampleRate_Counter <= (others => '0');
      SampleRate_SmallCounter <= (others => '0');

    elsif rising_edge(Clk) then
      -- Default
      END_PAUSE <= '0';
      SPIMACHINE_STATE_SR <= SM_SMALL;

      SampleRate_Copy <= SampleRate_Copy;
      SampleRate_Counter <= SampleRate_Counter;
      SampleRate_SmallCounter <= SampleRate_SmallCounter;

      case SPIMACHINE_STATE_SR is
        when SM_SMALL =>
          if SampleRate_SmallCounter = SampleRate_SmallCounter_Max then
            SPIMACHINE_STATE_SR <= SM_LONG;
            SampleRate_SmallCounter <= (others => '0');
          else
            SPIMACHINE_STATE_SR <= SM_SMALL;
            SampleRate_SmallCounter <= SampleRate_SmallCounter + 1;
          end if;
          SampleRate_Copy <= unsigned(SampleRate);

        when SM_LONG =>
          if SampleRate_Counter = SampleRate_Copy then
            SPIMACHINE_STATE_SR <= SM_SMALL;
            SampleRate_Counter <= (others => '0');
            END_PAUSE <= '1';
          else
            SPIMACHINE_STATE_SR <= SM_LONG;
            SampleRate_Counter <= SampleRate_Counter + 1;
          end if;

        when others =>
          SPIMACHINE_STATE_SR <= SM_SMALL;
      end case;
    end if;
  end process SLOW_SAMPLING_DOWN_PROCESS;
------------------------------------------------------------------------------
  COPY_VALID_DATA : process (Clk,Reset)
  begin
    if Reset = '1' then
      -- Internal Buffer
      DATAOUT_DATA_INTERNAL <= (others => '0');
      DATAOUT_VALIDSTRING <= (others => '0');

      -- External Signals
      DATAOUT_DATA_EXT <= (others => '0');
      DATAOUT_VALID_EXT <= '0';
      BufferOverflow_INTERNAL <= '0';

      Command_Processed <= '0';
      NewBlock	<= '0';

    elsif rising_edge(Clk) then

      -- Default:
      DATAOUT_DATA_INTERNAL <= DATAOUT_DATA_INTERNAL;
      DATAOUT_VALIDSTRING <= DATAOUT_VALIDSTRING;

      DATAOUT_DATA_EXT <= (others => '0');
      DATAOUT_VALID_EXT <= '0';

      NewBlock	<= NewBlock_OUT;
      Command_Processed <= Command_OUT;

      BufferOverflow_INTERNAL <= BufferOverflow_INTERNAL;

      if (NOT (DATAOUT_VALID = "00000000")) then
        -- If there is new data, then we store it in a buffer...
        DATAOUT_DATA_INTERNAL <= DATAOUT_DATA;
        DATAOUT_VALIDSTRING <= DATAOUT_VALID;

        -- Check for Buffer-Overruns
        if DATAOUT_VALIDSTRING = "00000000" then
          BufferOverflow_INTERNAL <= '0';
        else
          BufferOverflow_INTERNAL <= '1';
        end if;

      elsif ReadRequest = '1' then
        -- Bit shift the used data away
        -- The VALIDSTRING takes care of invalid data
        DATAOUT_DATA_INTERNAL <= "0000000000000000" & DATAOUT_DATA_INTERNAL(127 downto 16);
        DATAOUT_VALIDSTRING <= "0" & DATAOUT_VALIDSTRING(7 downto 1);

        -- This is were the output is generated
        -- NEEDS TO BE REVERSED
        -- DATAOUT_DATA_EXT <= DATAOUT_DATA_INTERNAL(15 downto 0);
        DATAOUT_DATA_EXT <= DATAOUT_DATA_INTERNAL( 0) & DATAOUT_DATA_INTERNAL( 1) & DATAOUT_DATA_INTERNAL( 2) & DATAOUT_DATA_INTERNAL( 3)
         & DATAOUT_DATA_INTERNAL( 4) & DATAOUT_DATA_INTERNAL( 5) & DATAOUT_DATA_INTERNAL( 6) & DATAOUT_DATA_INTERNAL( 7)
         & DATAOUT_DATA_INTERNAL( 8) & DATAOUT_DATA_INTERNAL( 9) & DATAOUT_DATA_INTERNAL(10) & DATAOUT_DATA_INTERNAL(11)
         & DATAOUT_DATA_INTERNAL(12) & DATAOUT_DATA_INTERNAL(13) & DATAOUT_DATA_INTERNAL(14) & DATAOUT_DATA_INTERNAL(15);
        DATAOUT_VALID_EXT <= DATAOUT_VALIDSTRING(0);
      else
        null;
      end if;
    end if;
  end process COPY_VALID_DATA;

  -- DEBUGGING FSM
  DEBUG_PROCESS : process (SPIMACHINE_STATE)
  begin
    case SPIMACHINE_STATE is
      when SM_COMMAND =>
        DEBUG_M_FSM0_STATE(1 downto 0) <= "00";
      when SM_COUNT =>
        DEBUG_M_FSM0_STATE(1 downto 0) <= "01";
      when others =>
        DEBUG_M_FSM0_STATE(1 downto 0) <= "11";
      end case;
  end process DEBUG_PROCESS;

  DEBUG_PROCESS_SR : process (SPIMACHINE_STATE_SR)
  begin
    case SPIMACHINE_STATE_SR is
      when SM_SMALL =>
        DEBUG_M_FSM0_STATE(3 downto 2) <= "00";
      when SM_LONG =>
        DEBUG_M_FSM0_STATE(3 downto 2) <= "01";
      when others =>
        DEBUG_M_FSM0_STATE(3 downto 2) <= "11";
      end case;
  end process DEBUG_PROCESS_SR;

end architecture Behavioral;
