-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This PHY-module contains a state machine which collects the data directly
-- from the RHD and deserializes it.
-- A problem is the delay between sending the command and the response.
-- We compensate it by a pipeline for the valid, newblock and command flag.

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO

library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;


entity RHD_DATA_COLLECTOR is
  Port (
    -- For the data stream
    RHD_MOSI_OUTSIDE        : out STD_LOGIC;
    RHD_MISO_OUTSIDE        :  in STD_LOGIC;
    RHD_Clk_OUTSIDE         : out STD_LOGIC;
    RHD_CS_OUTSIDE          : out STD_LOGIC;

    -- Results we get back
    DATAOUT_DATA            : out STD_LOGIC_VECTOR(15 downto 0);
    DATAOUT_VALID           : out STD_LOGIC;

    -- The ports for sending the data out
    DATAIN_DATA             :  in STD_LOGIC_VECTOR(15 downto 0);
    DATAIN_VALID            :  in STD_LOGIC;
    DATAIN_USED             : out STD_LOGIC;

    NewBlock_IN             :  in STD_LOGIC;
    NewBlock_OUT            : out STD_LOGIC;

    Command_IN              : in  STD_LOGIC;
    Command_OUT             : out STD_LOGIC;

    -- The whole test thing fails with the IGLOO FPGA
    -- Thus we need to work on Clk instead of on RHD_Clk_OUTSIDE
    TEST_INFO               : out STD_LOGIC;

    END_PAUSE               :  in STD_LOGIC;

  -- DEBUG SECTION
  DEBUG_M_FSM0_STATE          : out std_logic_vector(4 downto 0);

    -- The usual stuff (CLk = 24MHz)
    Clk                     : in  STD_LOGIC;
    Reset                   : in  STD_LOGIC
  );
end RHD_DATA_COLLECTOR;

architecture Behavioral of RHD_DATA_COLLECTOR is

  type STATE_Type is (
    SM_INIT,
    SM_WAIT,
    SM_CS_UP_A,
    SM_CS_UP_B,
    SM_CS_UP_C,
    SM_WARM_UP_A,
    SM_WARM_UP_B,
    SM_WARM_UP_C,
    SM_BIT_00,
    SM_BIT_01,
    SM_BIT_02,
    SM_BIT_03,
    SM_BIT_04,
    SM_BIT_05,
    SM_BIT_06,
    SM_BIT_07,
    SM_BIT_08,
    SM_BIT_09,
    SM_BIT_10,
    SM_BIT_11,
    SM_BIT_12,
    SM_BIT_13,
    SM_BIT_14,
    SM_BIT_15,
    SM_COOLDOWN_A
  );
  signal SPIMACHINE_STATE : STATE_TYPE;

  signal DATAOUT_DATA_Copy      : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal DATAIN_DATA_Copy       : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal NewBlock_Import        : STD_LOGIC := '0';
  signal NewBlock_Hold          : STD_LOGIC := '0';
  signal NewBlock_Export        : STD_LOGIC := '0';
  signal Valid_Import           : STD_LOGIC := '0';
  signal Valid_Hold             : STD_LOGIC := '0';
  signal Valid_Export           : STD_LOGIC := '0';
  signal Command_Import         : STD_LOGIC := '0';
  signal Command_Hold           : STD_LOGIC := '0';
  signal Command_Export         : STD_LOGIC := '0';
  signal TEST_INFO_INTERNAL     : STD_LOGIC := '0';

begin

  DATAOUT_DATA <= DATAOUT_DATA_Copy;
  TEST_INFO <= TEST_INFO_INTERNAL;

  FSM : process (Clk,Reset)
  begin

    if Reset = '1' then
      SPIMACHINE_STATE <= SM_INIT;
      RHD_MOSI_OUTSIDE <= '0';
      RHD_CS_OUTSIDE <= '1';
      DATAIN_USED <= '0';
      NewBlock_Import <= '0';
      NewBlock_Hold	<= '0';
      NewBlock_Export <= '0';
      Valid_Import <= '0';
      Valid_Hold <= '0';
      Valid_Export <= '0';
      Command_Import	<= '0';
      Command_Hold <= '0';
      Command_Export <= '0';
      TEST_INFO_INTERNAL <= '0';

    elsif falling_edge(Clk) then
      -- Default:
      SPIMACHINE_STATE <= SM_INIT;
      RHD_MOSI_OUTSIDE <= '0';
      RHD_CS_OUTSIDE <= '0';
      DATAIN_USED <= '0';
      NewBlock_Import <= NewBlock_Import;
      NewBlock_Hold	<= NewBlock_Hold;
      NewBlock_Export <= NewBlock_Export;
      Valid_Import <= Valid_Import;
      Valid_Hold <= Valid_Hold;
      Valid_Export <= Valid_Export;
      Command_Import	<= Command_Import;
      Command_Hold <= Command_Hold;
      Command_Export <= Command_Export;
      TEST_INFO_INTERNAL <= '0';

      case SPIMACHINE_STATE is
        when SM_INIT =>
          SPIMACHINE_STATE <= SM_WAIT;
          DATAIN_DATA_Copy <= DATAIN_DATA;
          RHD_CS_OUTSIDE <= '1';

          -- We need this for compensating the delay of the RHD
          NewBlock_Import <= NewBlock_IN;
          NewBlock_Hold	<= NewBlock_Import;
          NewBlock_Export <= NewBlock_Hold;

          Valid_Import <= DATAIN_VALID;
          Valid_Hold <= Valid_Import;
          Valid_Export <= Valid_Hold;

          Command_Import	<= Command_IN;
          Command_Hold <= Command_Import;
          Command_Export <= Command_Hold;

        when SM_WAIT =>
          if END_PAUSE = '1' then
            SPIMACHINE_STATE <= SM_CS_UP_A;
          else
            SPIMACHINE_STATE <= SM_WAIT;
          end if;
          RHD_CS_OUTSIDE <= '1';


        when SM_CS_UP_A =>
          SPIMACHINE_STATE <= SM_CS_UP_B;
          RHD_CS_OUTSIDE <= '1';
          DATAIN_USED <= '1';

        when SM_CS_UP_B =>
          SPIMACHINE_STATE <= SM_CS_UP_C;
          RHD_CS_OUTSIDE <= '1';

        when SM_CS_UP_C =>
          SPIMACHINE_STATE <= SM_WARM_UP_A;
          RHD_CS_OUTSIDE <= '1';

        when SM_WARM_UP_A =>
          SPIMACHINE_STATE <= SM_WARM_UP_B;

        when SM_WARM_UP_B =>
          SPIMACHINE_STATE <= SM_WARM_UP_C;

        when SM_WARM_UP_C =>
          SPIMACHINE_STATE <= SM_BIT_15;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy(15);
          TEST_INFO_INTERNAL <= '1';

        when SM_BIT_15 =>
          SPIMACHINE_STATE <= SM_BIT_14;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy(14);

        when SM_BIT_14 =>
          SPIMACHINE_STATE <= SM_BIT_13;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy(13);

        when SM_BIT_13 =>
          SPIMACHINE_STATE <= SM_BIT_12;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy(12);

        when SM_BIT_12 =>
          SPIMACHINE_STATE <= SM_BIT_11;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy(11);

        when SM_BIT_11 =>
          SPIMACHINE_STATE <= SM_BIT_10;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy(10);

        when SM_BIT_10 =>
          SPIMACHINE_STATE <= SM_BIT_09;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 9);

        when SM_BIT_09 =>
          SPIMACHINE_STATE <= SM_BIT_08;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 8);

        when SM_BIT_08 =>
          SPIMACHINE_STATE <= SM_BIT_07;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 7);

        when SM_BIT_07 =>
          SPIMACHINE_STATE <= SM_BIT_06;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 6);

        when SM_BIT_06 =>
          SPIMACHINE_STATE <= SM_BIT_05;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 5);

        when SM_BIT_05 =>
          SPIMACHINE_STATE <= SM_BIT_04;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 4);

        when SM_BIT_04 =>
          SPIMACHINE_STATE <= SM_BIT_03;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 3);

        when SM_BIT_03 =>
          SPIMACHINE_STATE <= SM_BIT_02;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 2);

        when SM_BIT_02 =>
          SPIMACHINE_STATE <= SM_BIT_01;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 1);

        when SM_BIT_01 =>
          SPIMACHINE_STATE <= SM_BIT_00;
          RHD_MOSI_OUTSIDE <= DATAIN_DATA_Copy( 0);

        when SM_BIT_00 =>
          SPIMACHINE_STATE <= SM_COOLDOWN_A;

        when SM_COOLDOWN_A =>
          SPIMACHINE_STATE <= SM_INIT;

        when others =>
          SPIMACHINE_STATE <= SM_INIT;
      end case;
    end if;
  end process FSM;
-----------------------------------------------------------------------------
  DATACOLLECT : process (Clk,Reset)
  begin

    if Reset = '1' then
      DATAOUT_DATA_Copy <= (others => '0');
      DATAOUT_VALID <= '0';
      NewBlock_OUT <= '0';
      Command_OUT <= '0';

    elsif rising_edge(Clk) then
      -- Default:
      DATAOUT_DATA_Copy <= DATAOUT_DATA_Copy;
      DATAOUT_VALID <= '0';
      NewBlock_OUT <= '0';
      Command_OUT <= '0';

      case SPIMACHINE_STATE is
        when SM_WAIT =>
          null;

        when SM_CS_UP_A =>
          null;

        when SM_CS_UP_B =>
          null;

        when SM_CS_UP_C =>
          null;

        when SM_WARM_UP_A =>
          null;

        when SM_WARM_UP_B =>
          null;

        when SM_WARM_UP_C =>
          null;

        when SM_BIT_15 =>
          DATAOUT_DATA_Copy(15) <= RHD_MISO_OUTSIDE;

        when SM_BIT_14 =>
          DATAOUT_DATA_Copy(14) <= RHD_MISO_OUTSIDE;

        when SM_BIT_13 =>
          DATAOUT_DATA_Copy(13) <= RHD_MISO_OUTSIDE;

        when SM_BIT_12 =>
          DATAOUT_DATA_Copy(12) <= RHD_MISO_OUTSIDE;

        when SM_BIT_11 =>
          DATAOUT_DATA_Copy(11) <= RHD_MISO_OUTSIDE;

        when SM_BIT_10 =>
          DATAOUT_DATA_Copy(10) <= RHD_MISO_OUTSIDE;

        when SM_BIT_09 =>
          DATAOUT_DATA_Copy( 9) <= RHD_MISO_OUTSIDE;

        when SM_BIT_08 =>
          DATAOUT_DATA_Copy( 8) <= RHD_MISO_OUTSIDE;

        when SM_BIT_07 =>
          DATAOUT_DATA_Copy( 7) <= RHD_MISO_OUTSIDE;

        when SM_BIT_06 =>
          DATAOUT_DATA_Copy( 6) <= RHD_MISO_OUTSIDE;

        when SM_BIT_05 =>
          DATAOUT_DATA_Copy( 5) <= RHD_MISO_OUTSIDE;

        when SM_BIT_04 =>
          DATAOUT_DATA_Copy( 4) <= RHD_MISO_OUTSIDE;

        when SM_BIT_03 =>
          DATAOUT_DATA_Copy( 3) <= RHD_MISO_OUTSIDE;

        when SM_BIT_02 =>
          DATAOUT_DATA_Copy( 2) <= RHD_MISO_OUTSIDE;

        when SM_BIT_01 =>
          DATAOUT_DATA_Copy( 1) <= RHD_MISO_OUTSIDE;

        when SM_BIT_00 =>
          DATAOUT_DATA_Copy( 0) <= RHD_MISO_OUTSIDE;
          NewBlock_OUT <= NewBlock_Export;
          Command_OUT <= Command_Export;
          DATAOUT_VALID <= Valid_Export;

        when SM_COOLDOWN_A =>
          null;

        when others =>
          null;
      end case;
    end if;
  end process DATACOLLECT;
-----------------------------------------------------------------------------
  SPI_CLOCK_SWITCH : process (Clk,SPIMACHINE_STATE)
  begin
    case SPIMACHINE_STATE is
      when SM_WAIT =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_CS_UP_A =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_CS_UP_B =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_CS_UP_C =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_WARM_UP_A =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_WARM_UP_B =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_WARM_UP_C =>
        RHD_Clk_OUTSIDE <= '0';

      when SM_BIT_15 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_14 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_13 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_12 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_11 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_10 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_09 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_08 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_07 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_06 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_05 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_04 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_03 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_02 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_01 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_BIT_00 =>
        RHD_Clk_OUTSIDE <= Clk;

      when SM_COOLDOWN_A =>
        RHD_Clk_OUTSIDE <= '0';

      when others =>
        RHD_Clk_OUTSIDE <= '0';
    end case;
  end process SPI_CLOCK_SWITCH;
-----------------------------------------------------------------------------
  -- DEBUGGING FSM
  DEBUG_PROCESS : process (SPIMACHINE_STATE)
  begin
    case SPIMACHINE_STATE is
      when SM_INIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 0,5));
      when SM_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 1,5));
      when SM_CS_UP_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 2,5));
      when SM_CS_UP_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 3,5));
      when SM_CS_UP_C =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 4,5));
      when SM_WARM_UP_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 5,5));
      when SM_WARM_UP_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 6,5));
      when SM_WARM_UP_C =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 7,5));
      when SM_BIT_00 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 8,5));
      when SM_BIT_01 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 9,5));
      when SM_BIT_02 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(10,5));
      when SM_BIT_03 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(11,5));
      when SM_BIT_04 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(12,5));
      when SM_BIT_05 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(13,5));
      when SM_BIT_06 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(14,5));
      when SM_BIT_07 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(15,5));
      when SM_BIT_08 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(16,5));
      when SM_BIT_09 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(17,5));
      when SM_BIT_10 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(18,5));
      when SM_BIT_11 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(19,5));
      when SM_BIT_12 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(20,5));
      when SM_BIT_13 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(21,5));
      when SM_BIT_14 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(22,5));
      when SM_BIT_15 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(23,5));
      when SM_COOLDOWN_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(24,5));
      when others =>
        DEBUG_M_FSM0_STATE <= "11111";
      end case;
  end process DEBUG_PROCESS;
end architecture Behavioral;
