-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This module cuts the data from the RHDs and reformates them for the
-- wireless data transmission

-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO
library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity RHD_TO_ZL_CONVERTER is
  Port (

    -- This we get from the RHA side:
    DATAOUT_DATA_OUT          :  in STD_LOGIC_VECTOR(15 downto 0);
    DATAOUT_VALID_OUT         :  in STD_LOGIC;

    NewBlock                  :  in STD_LOGIC;
    Command_Processed         :  in STD_LOGIC;

    ReadRequest               : out STD_LOGIC;

    -- This goes into the Zarlink / FIFO direction:
    FIFO_OUT_DATA             : out STD_LOGIC_VECTOR(7 downto 0);
    FIFO_OUT_VALID            : out STD_LOGIC;

    ERROR_ALL_OUT             :  in STD_LOGIC;

    -- Parameters for the header
    DeviceID                  :  in std_logic_vector(7 downto 0);

    -- Bitmask for the data selection from the 16 bit samples.
    BITMASK_SAMPLE            :  in std_logic_vector(15 downto 0);

    CommandInfos_Data         :  in STD_LOGIC_VECTOR(15 downto 0);
    CommandInfos_Valid        :  in STD_LOGIC;
    CommandInfos_NextData     : out STD_LOGIC;

    -- 32 channel plus 3 aux channels = 35
    -- 8 RHA x 35 channels = 280 samples = 4480 bits
    -- Plus a 31bit header = 4511 bits = 564 bytes (rounded up)
    -- This results in upto 41 x 14byte blocks = 574 bytes.
    ExportNext280ChannelBlock :  in STD_LOGIC;

    -- DEBUG SECTION
    DEBUG_M_FSM0_STATE          : out std_logic_vector (3 downto 0);

    Clk                       :  in STD_LOGIC;
    Reset                     :  in STD_LOGIC
  );
end RHD_TO_ZL_CONVERTER;

architecture Behavioral of RHD_TO_ZL_CONVERTER is

  -- For the headers...

  signal Header_DATA              : std_logic_vector( 30 downto 0) := (others => '0');
  signal Timestamp                : unsigned(15 downto 0) := (others => '0');

  signal Header_CMDRESPONSE       : std_logic_vector( 30 downto 0) := (others => '0');
  signal CMDRESPONSE_COUNTER      : unsigned(15 downto 0) := (others => '0');

  constant PacketHeader_DATA      : std_logic_vector(3 downto 0) := "0001";
  constant PacketHeader_CMDRESPONSE : std_logic_vector(3 downto 0) := "0011";

  signal FIFO_OUT_DATA_INTERNAL   : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal FIFO_OUT_VALID_INTERNAL  : STD_LOGIC := '0';

  signal BITMASK_SAMPLE_COPY      : std_logic_vector(15 downto 0) := (others => '0');
  signal DATA_COPY                : std_logic_vector(15 downto 0) := (others => '0');
  signal BITS_TO_PROCESS          : std_logic_vector(15 downto 0) := (others => '0');

  -- Needs to count up to 112
  signal COUNTER_BITS_PROCESSED   : unsigned(6 downto 0) := (others => '0');
  -- Needs to count up to 8
  signal COUNTER_BITS_OF_A_BYTE   : unsigned(3 downto 0) := (others => '0');

  signal NewBlock_LAST            : STD_LOGIC := '0';
  signal Command_Processed_LAST   : STD_LOGIC := '0';

  signal ReadRequest_INTERNAL     : STD_LOGIC := '0';
  signal ReadRequest_INTERNAL_LAST : STD_LOGIC := '0';
  signal ReadRequest_INTERNAL_LAST2 : STD_LOGIC := '0';
  signal ReadRequest_INTERNAL_LAST3 : STD_LOGIC := '0';

  signal DATAOUT_DATA_OUT_BUFFER  : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
  signal DATAOUT_VALID_OUT_BUFFER : STD_LOGIC := '0';
  signal DATA_USED                : STD_LOGIC := '0';

  signal NewBlock_Off             : STD_LOGIC := '0';
  signal NewBlock_INTERNAL        : STD_LOGIC := '0';

  signal Command_Processed_Off    : STD_LOGIC := '0';
  signal Command_Processed_INTERNAL : STD_LOGIC := '0';
  signal Command_Processed_SUB_INTERNAL : STD_LOGIC := '0';
  signal ERROR_ALL_OUT_INTERNAL   : STD_LOGIC := '0';

  signal NextCOMMAND              : STD_LOGIC := '0';
  signal NextBLOCK                : STD_LOGIC := '0';

  signal CommandResponseMode      : std_logic := '0';

  type STATE_Type is (
    -- Wait until there is enougth space left and a new block begins
    SM_IDLE,
    -- Header_Short(7 downto 0)
    SM_HEADER_DATA_01,
    -- Header_Short(15 downto 8)
    SM_HEADER_DATA_02,
    -- Header_Short(23 downto 16)
    SM_HEADER_DATA_03,
    -- Header_Short(30 downto 24)
    SM_HEADER_DATA_UNFINISHED,
    -- Header_CMDRESPONSE(7 downto 0)
    SM_HEADER_CMDRESPONSE_01,
    -- Header_CMDRESPONSE(15 downto 8)
    SM_HEADER_CMDRESPONSE_02,
    -- Header_CMDRESPONSE(23 downto 16)
    SM_HEADER_CMDRESPONSE_03,
    -- Header_CMDRESPONSE(31 downto 24)
    SM_HEADER_CMDRESPONSE_04,
    -- Header_CMDRESPONSE(39 downto 32)
    SM_HEADER_CMDRESPONSE_05,
    -- Header_CMDRESPONSE(46 downto 40)
    SM_HEADER_CMDRESPONSE_UNFINISHED,
    SM_COPY_DATA_BITS_REQUEST,
    SM_COPY_DATA_BITS_START,
    SM_COPY_DATA_BITS_NEWLINE,
    SM_CLEAN_UP
  );
  signal SPIMACHINE_STATE         : STATE_TYPE;

begin
  -- Export internal and buffered signals
  FIFO_OUT_DATA <= FIFO_OUT_DATA_INTERNAL;
  FIFO_OUT_VALID <= FIFO_OUT_VALID_INTERNAL;
  --	FIFO_OUT_VALID <= '0';

  ERROR_ALL_OUT_INTERNAL <= ERROR_ALL_OUT;

  -- Here we compile the headers
  Header_DATA(0) <= '1'; -- Begin of a data set which contains all the selected electrodes
  Header_DATA(4 downto 1) <= PacketHeader_DATA; -- Marks the type of the header (short)
  Header_DATA(12 downto 5) <= DeviceID; -- ID of the ASIC
  Header_DATA(28 downto 13) <= std_logic_vector(Timestamp); -- unsigned(15 downto 0)
  Header_DATA(29) <= ERROR_ALL_OUT_INTERNAL; -- unused/control bit
  Header_DATA(30) <= '0'; -- unused/control bit

  -- Here we compile the headers
  Header_CMDRESPONSE(0) <= '1'; -- Begin of a data set which contains all the selected electrodes
  Header_CMDRESPONSE(4 downto 1) <= PacketHeader_CMDRESPONSE; -- Marks the type of the header (short)
  Header_CMDRESPONSE(12 downto 5) <= DeviceID; -- ID of the ASIC
  Header_CMDRESPONSE(28 downto 13) <= std_logic_vector(CMDRESPONSE_COUNTER); -- unsigned(15 downto 0)
  Header_CMDRESPONSE(29) <= ERROR_ALL_OUT_INTERNAL; -- unused/control bit
  Header_CMDRESPONSE(30) <= '0'; -- unused/control bit

  ReadRequest <= ReadRequest_INTERNAL;
  -- We need to decouple the data request from the rest
  DATA_IMPORTING_PROCESS : process (Clk,Reset)
  begin
    if Reset = '1' then
      ReadRequest_INTERNAL <= '0';
      DATAOUT_DATA_OUT_BUFFER <= (others => '0');
      DATAOUT_VALID_OUT_BUFFER <= '0';
      ReadRequest_INTERNAL_LAST <= '0';
      ReadRequest_INTERNAL_LAST2 <= '0';
      ReadRequest_INTERNAL_LAST3 <= '0';
    elsif rising_edge(Clk) then

      ReadRequest_INTERNAL <= '0';
      DATAOUT_DATA_OUT_BUFFER <= DATAOUT_DATA_OUT_BUFFER;
      DATAOUT_VALID_OUT_BUFFER <= DATAOUT_VALID_OUT_BUFFER;

      ReadRequest_INTERNAL_LAST <= ReadRequest_INTERNAL;
      ReadRequest_INTERNAL_LAST2 <= ReadRequest_INTERNAL_LAST;
      ReadRequest_INTERNAL_LAST3 <= ReadRequest_INTERNAL_LAST2;

      if DATAOUT_VALID_OUT_BUFFER = '1' then
        null;
      elsif (ReadRequest_INTERNAL = '0') AND (ReadRequest_INTERNAL_LAST = '0')
       AND (ReadRequest_INTERNAL_LAST2 = '0') AND (ReadRequest_INTERNAL_LAST3 = '0') then
        ReadRequest_INTERNAL <= '1';
      else
        null;
      end if;

      if DATAOUT_VALID_OUT = '1' then
        DATAOUT_DATA_OUT_BUFFER <= DATAOUT_DATA_OUT;
        DATAOUT_VALID_OUT_BUFFER <= '1';
      elsif DATA_USED = '1' then
        DATAOUT_DATA_OUT_BUFFER <= (others => '0');
        DATAOUT_VALID_OUT_BUFFER <= '0';
      else
        null;
      end if;

    end if;
  end process DATA_IMPORTING_PROCESS;
-----------------------------------------------------------------------------
  -- We check when a new block arrives and count the timestamp one up.
  TIME_STAMP_PROCESS : process (Clk,Reset)
  begin
    if Reset = '1' then
      Timestamp <= (others => '0');
      CMDRESPONSE_COUNTER <= (others => '0');

    elsif rising_edge(Clk) then
      NewBlock_LAST <= NewBlock;
      Command_Processed_LAST <= Command_Processed;

      Timestamp <= Timestamp;
      CMDRESPONSE_COUNTER <= CMDRESPONSE_COUNTER;

      -- Time stamp
      if (NewBlock = '1') AND (NewBlock_LAST = '0') then
        Timestamp <= Timestamp + 1;
      else
        null;
      end if;

      -- Command Counter
      if (Command_Processed = '1') AND (Command_Processed_LAST = '0')
       AND (CommandInfos_Valid = '1') then
        CMDRESPONSE_COUNTER <= CMDRESPONSE_COUNTER + 1;
      else
        null;
      end if;

    end if;
  end process TIME_STAMP_PROCESS;
-----------------------------------------------------------------------------
  FSM_BLOCK : process (Clk,Reset)
  begin
    if Reset = '1' then
      FIFO_OUT_DATA_INTERNAL <= (others => '0');
      FIFO_OUT_VALID_INTERNAL <= '0';
      COUNTER_BITS_PROCESSED <= (others => '0');
      BITMASK_SAMPLE_COPY <= (others => '0');
      DATA_COPY <= (others => '0');
      BITS_TO_PROCESS <= (others => '0');
      COUNTER_BITS_OF_A_BYTE <= (others => '0');

      NewBlock_Off <= '0';
      NewBlock_INTERNAL <= '0';

      Command_Processed_Off <= '0';
      Command_Processed_INTERNAL <= '0';
      Command_Processed_SUB_INTERNAL <= '0';

      DATA_USED <= '0';

      NextCOMMAND <= '0';
      NextBLOCK <= '0';

      CommandInfos_NextData <= '0';
      CommandResponseMode <= '0';

      SPIMACHINE_STATE <= SM_IDLE;

    elsif rising_edge(Clk) then
      FIFO_OUT_DATA_INTERNAL <= FIFO_OUT_DATA_INTERNAL;
      FIFO_OUT_VALID_INTERNAL <= '0';
      COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED;
      BITMASK_SAMPLE_COPY <= BITMASK_SAMPLE_COPY;
      DATA_COPY <= DATA_COPY;
      BITS_TO_PROCESS <= BITS_TO_PROCESS;

      COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE;

      NewBlock_Off <= '0';
      Command_Processed_Off <= '0';

      DATA_USED <= '0';

      NextCOMMAND <= NextCOMMAND;
      NextBLOCK <= NextBLOCK;
      CommandResponseMode <= CommandResponseMode;

      -- Failsafe !
      SPIMACHINE_STATE <= SM_IDLE;

      if (NewBlock = '1') AND (NewBlock_LAST = '0') AND (ExportNext280ChannelBlock = '1') then
        NewBlock_INTERNAL <= '1';
      elsif NewBlock_Off = '1' then
        NewBlock_INTERNAL <= '0';
      else
        NewBlock_INTERNAL <= NewBlock_INTERNAL;
      end if;

      if (Command_Processed = '1') AND (Command_Processed_LAST = '0')
       AND (CommandInfos_Valid = '1') AND (ExportNext280ChannelBlock = '1') then
        Command_Processed_INTERNAL <= '1';
      elsif Command_Processed_Off = '1' then
        Command_Processed_INTERNAL <= '0';
      else
        Command_Processed_INTERNAL <= Command_Processed_INTERNAL;
      end if;

      if (Command_Processed = '1') AND (Command_Processed_LAST = '0') then
        Command_Processed_SUB_INTERNAL <= '1';
      elsif Command_Processed_Off = '1' then
        Command_Processed_SUB_INTERNAL <= '0';
      else
        Command_Processed_SUB_INTERNAL <= Command_Processed_SUB_INTERNAL;
      end if;

      CommandInfos_NextData <= '0';

      case SPIMACHINE_STATE is

          -- Wait until something has to be done. Now we decide which header we need to use...
          -- Use the long one only once!
        when SM_IDLE =>
          -- We find a command
          if (Command_Processed_INTERNAL = '1') AND (CommandInfos_Valid = '1')
           AND (ExportNext280ChannelBlock = '1')  then
            SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_01;
            Command_Processed_Off <= '1';
            NewBlock_Off <= '1';
            -- We find a new block of data
          elsif (NewBlock_INTERNAL = '1') AND (ExportNext280ChannelBlock = '1')  then
            SPIMACHINE_STATE <= SM_HEADER_DATA_01;
            Command_Processed_Off <= '1';
            NewBlock_Off <= '1';
          else
            SPIMACHINE_STATE <= SM_IDLE;
          end if;

          -- We need to wait for the next new block or new command
          if (DATAOUT_VALID_OUT_BUFFER = '1') AND (NewBlock_INTERNAL = '0') then
            DATA_USED <= '1';
-- Changed because fifo-fall through
--        elsif (DATAOUT_VALID_OUT_BUFFER = '1') AND (Command_Processed_INTERNAL = '0') then
--          DATA_USED <= '1';
--        elsif (DATAOUT_VALID_OUT_BUFFER = '1') AND (Command_Processed_INTERNAL = '1')
--         AND (CommandInfos_Valid = '0') then
--          DATA_USED <= '1';
            -- We need to discard the newblock because we have no space in the fifo left
          elsif (DATAOUT_VALID_OUT_BUFFER = '1') AND (ExportNext280ChannelBlock = '0') then
            DATA_USED <= '1';
          else
            null;
          end if;

-----------------------------------------------------------------------------
          -- DATA HEADER
        when SM_HEADER_DATA_01 =>
          SPIMACHINE_STATE <= SM_HEADER_DATA_02;
          FIFO_OUT_DATA_INTERNAL <= Header_DATA(7 downto 0);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_DATA_02 =>
          SPIMACHINE_STATE <= SM_HEADER_DATA_03;
          FIFO_OUT_DATA_INTERNAL <= Header_DATA(15 downto 8);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_DATA_03 =>
          SPIMACHINE_STATE <= SM_HEADER_DATA_UNFINISHED;
          FIFO_OUT_DATA_INTERNAL <= Header_DATA(23 downto 16);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_DATA_UNFINISHED =>
          SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
          FIFO_OUT_DATA_INTERNAL <= Header_DATA(30 downto 24) & "0";
          COUNTER_BITS_PROCESSED <= to_unsigned(31,7);
          COUNTER_BITS_OF_A_BYTE <= to_unsigned(7,4);
          CommandResponseMode <= '0';

-----------------------------------------------------------------------------
          -- COMMAND HEADER
        when SM_HEADER_CMDRESPONSE_01 =>
          SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_02;
          FIFO_OUT_DATA_INTERNAL <= Header_CMDRESPONSE(7 downto 0);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_CMDRESPONSE_02 =>
          SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_03;
          FIFO_OUT_DATA_INTERNAL <= Header_CMDRESPONSE(15 downto 8);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_CMDRESPONSE_03 =>
          SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_04;
          FIFO_OUT_DATA_INTERNAL <= Header_CMDRESPONSE(23 downto 16);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_CMDRESPONSE_04 =>
          SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_05;
          FIFO_OUT_DATA_INTERNAL <= CommandInfos_Data(0) & Header_CMDRESPONSE(30 downto 24);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_CMDRESPONSE_05 =>
          SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_UNFINISHED;
          FIFO_OUT_DATA_INTERNAL <= CommandInfos_Data(8 downto 1);
          FIFO_OUT_VALID_INTERNAL <= '1';

        when SM_HEADER_CMDRESPONSE_UNFINISHED =>
          SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
          FIFO_OUT_DATA_INTERNAL <= CommandInfos_Data(15 downto 9) & "0";
          COUNTER_BITS_PROCESSED <= to_unsigned(47,7);
          COUNTER_BITS_OF_A_BYTE <= to_unsigned(7,4);
          CommandInfos_NextData <= '1';
          CommandResponseMode <= '1';

-----------------------------------------------------------------------------
        -- For the data:
        -- Here we request and wait for the next sample
        when SM_COPY_DATA_BITS_REQUEST =>
          -- If we get during getting the new sample,
          -- a newblock signal or a new command then we have to close
          -- the old block
          -- New command
          -- if Command_Processed_INTERNAL = '1' then
          if Command_Processed_SUB_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_CLEAN_UP;
            Command_Processed_Off <= '1';
            NextCOMMAND <= '1';
            NextBLOCK <= '0';
            -- New Block of data
          elsif NewBlock_INTERNAL = '1' then
            SPIMACHINE_STATE <= SM_CLEAN_UP;
            NewBlock_Off <= '1';
            NextCOMMAND <= '0';
            NextBLOCK <= '1';
            -- New data discovered. Do something with it.
          elsif DATAOUT_VALID_OUT_BUFFER = '1' then
            SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
            BITMASK_SAMPLE_COPY <= BITMASK_SAMPLE;
            DATA_COPY <= DATAOUT_DATA_OUT_BUFFER;
            BITS_TO_PROCESS <= (others => '1');
            DATA_USED <= '1';
            -- Wait for new data in this state
          else
            SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
          end if;

        when SM_COPY_DATA_BITS_START =>
          -- All the data was used, we need to request new data
          if BITS_TO_PROCESS(0) = '0' then
            SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
            -- The Zarlink string is full and a new one needs to be created
            -- export the finished string to the fifo
          elsif COUNTER_BITS_OF_A_BYTE = to_unsigned(8,4) then
            SPIMACHINE_STATE <= SM_COPY_DATA_BITS_NEWLINE;
            FIFO_OUT_VALID_INTERNAL <= '1';
            -- We found a bit to process
          elsif (BITS_TO_PROCESS(0) = '1') AND ((BITMASK_SAMPLE_COPY(0) = '1')
           OR (CommandResponseMode = '1')) then
            SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
            FIFO_OUT_DATA_INTERNAL <= DATA_COPY(0) & FIFO_OUT_DATA_INTERNAL(7 downto 1);
            COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED + 1;
            COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE + 1;
            BITMASK_SAMPLE_COPY <= "0" & BITMASK_SAMPLE_COPY(15 downto 1);
            DATA_COPY <= "0" & DATA_COPY(15 downto 1);
            BITS_TO_PROCESS <= "0" & BITS_TO_PROCESS(15 downto 1);
            -- This bit shouldn't be processed
          else
            SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
            BITMASK_SAMPLE_COPY <= "0" & BITMASK_SAMPLE_COPY(15 downto 1);
            DATA_COPY <= "0" & DATA_COPY(15 downto 1);
            BITS_TO_PROCESS <= "0" & BITS_TO_PROCESS(15 downto 1);
          end if;

        -- Here we begin a new Zarlink string.
        when SM_COPY_DATA_BITS_NEWLINE =>
          SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
          if COUNTER_BITS_PROCESSED = to_unsigned(112,7) then
            -- It can only start with one "0"
            -- because it is never a header
            FIFO_OUT_DATA_INTERNAL <= (others => '0');
            COUNTER_BITS_PROCESSED <= to_unsigned(1,7);
            COUNTER_BITS_OF_A_BYTE <= to_unsigned(1,4);
          else
            -- We only finished a 8 bit block...
            FIFO_OUT_DATA_INTERNAL <= (others => '0');
            COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED;
            COUNTER_BITS_OF_A_BYTE <= to_unsigned(0,4);
          end if;

        -- Fill up the last 112 bits with zeros...
        when SM_CLEAN_UP =>
          -- If everything is filled up with zeros then...
          -- go here
          if (COUNTER_BITS_PROCESSED = to_unsigned(112,7)) AND (ExportNext280ChannelBlock = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            FIFO_OUT_VALID_INTERNAL <= '1';
            -- or here
            -- The next one will be data
          elsif (COUNTER_BITS_PROCESSED = to_unsigned(112,7)) AND (ExportNext280ChannelBlock = '1')
           AND (NextBLOCK = '1') then
            SPIMACHINE_STATE <= SM_HEADER_DATA_01;
            FIFO_OUT_VALID_INTERNAL <= '1';
            -- The next one will be a command response
          elsif (COUNTER_BITS_PROCESSED = to_unsigned(112,7)) AND (ExportNext280ChannelBlock = '1')
           AND (NextCOMMAND = '1') AND (CommandInfos_Valid = '1') then
            SPIMACHINE_STATE <= SM_HEADER_CMDRESPONSE_01;
            FIFO_OUT_VALID_INTERNAL <= '1';
            -- There is no real command reponse required only a unused husk.
          elsif (COUNTER_BITS_PROCESSED = to_unsigned(112,7)) AND (ExportNext280ChannelBlock = '1')
           AND (NextCOMMAND = '1') AND (CommandInfos_Valid = '0') then
            SPIMACHINE_STATE <= SM_IDLE;
            FIFO_OUT_VALID_INTERNAL <= '1';
          elsif COUNTER_BITS_OF_A_BYTE = to_unsigned(8,4) then
            -- Send another 8 bits on their way
            SPIMACHINE_STATE <= SM_CLEAN_UP;
            FIFO_OUT_VALID_INTERNAL <= '1';
            COUNTER_BITS_OF_A_BYTE <= (others => '0');
          elsif (COUNTER_BITS_OF_A_BYTE = to_unsigned(0,4)) AND (FIFO_OUT_VALID_INTERNAL = '1') then
            SPIMACHINE_STATE <= SM_CLEAN_UP;
            FIFO_OUT_DATA_INTERNAL <= (others => '0');
            COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED + 8;
            COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE + 8;
          else
            -- if zeros are missing continue here and stay in this state
            SPIMACHINE_STATE <= SM_CLEAN_UP;
            FIFO_OUT_DATA_INTERNAL <= "0" & FIFO_OUT_DATA_INTERNAL(7 downto 1);
            COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED + 1;
            COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE + 1;
          end if;

        when others =>
          SPIMACHINE_STATE <= SM_IDLE;
      end case;
    end if;
  end process FSM_BLOCK;

  -- DEBUGGING FSM
  DEBUG_PROCESS : process (SPIMACHINE_STATE)
  begin
    case SPIMACHINE_STATE is
      when SM_IDLE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 0,4));
      when SM_HEADER_DATA_01 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 1,4));
      when SM_HEADER_DATA_02 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 2,4));
      when SM_HEADER_DATA_03 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 3,4));
      when SM_HEADER_DATA_UNFINISHED =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 4,4));
      when SM_HEADER_CMDRESPONSE_01 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 5,4));
      when SM_HEADER_CMDRESPONSE_02 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 6,4));
      when SM_HEADER_CMDRESPONSE_03 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 7,4));
      when SM_HEADER_CMDRESPONSE_04 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 8,4));
      when SM_HEADER_CMDRESPONSE_05 =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 9,4));
      when SM_HEADER_CMDRESPONSE_UNFINISHED =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(10,4));
      when SM_COPY_DATA_BITS_REQUEST =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(11,4));
      when SM_COPY_DATA_BITS_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(12,4));
      when SM_COPY_DATA_BITS_NEWLINE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(13,4));
      when SM_CLEAN_UP =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(14,4));
      when others =>
        DEBUG_M_FSM0_STATE <= "1111";
      end case;
  end process DEBUG_PROCESS;

end architecture Behavioral;
