-- Written by David Rotermund (davrot@neuro.uni-bremen.de)
-- University of Bremen, Institute for theoretical physiscs
-- (Use Atom.IO editor for correct indents)

-- This module organizes the Zarlink connection and handing of the fifo
-- for command receiving and data export
-- Version 0.0 (23.07.14) : Original source code was written
-- Version 0.1 (03.07.15) : Added debug info, reformating with Atom.IO

library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity ZARLINK_CONNECT_INIT is
  Port (

    -- ZarlinkConnectionInit
    ZarlinkConnectionInit_DO    :  in STD_LOGIC;

    -- CheckConnection
    CheckConnection_CHECK       :  in STD_LOGIC;
    CheckConnection_OK          : out STD_LOGIC;

    -- CheckIngoingBuffer
    CheckIngoingBuffer_CHECK    :  in STD_LOGIC;
    CheckIngoingBuffer_OK       : out STD_LOGIC;
    CheckIngoingBuffer_SIZE     : out STD_LOGIC_VECTOR (6 downto 0);

    -- CheckOutgoingBuffer
    CheckOutgoingBuffer_CHECK   :  in STD_LOGIC;
    CheckOutgoingBuffer_OK      : out STD_LOGIC;
    CheckOutgoingBuffer_SIZE    : out STD_LOGIC_VECTOR (6 downto 0);

    -- GetCommand
    GetCommand_GET              :  in STD_LOGIC;
    GetCommand_DATA             : out STD_LOGIC_VECTOR(7 downto 0);
    GetCommand_VALID            : out STD_LOGIC;

    -- Write data to Zarlink
    Zarlink_FIFO_WRITE          :  in STD_LOGIC;
    Zarlink_FIFO_DATA           :  in STD_LOGIC_VECTOR(7 downto 0);

    -- Connection to the Zarlink
    Zarlink_WriteEn             : out STD_LOGIC;
    Zarlink_Adress              : out STD_LOGIC_VECTOR(7 downto 0);
    Zarlink_Data_Out            : out STD_LOGIC_VECTOR(7 downto 0);
    Zarlink_Data_In             :  in STD_LOGIC_VECTOR(7 downto 0);
    Zarlink_Valid_In            :  in STD_LOGIC;
    Zarlink_BlockMode           : out STD_LOGIC;
    Zarlink_Wait                :  in STD_LOGIC;

    WaitSignal_ProcessOngoing   : out STD_LOGIC;

    -- DEBUG SECTION
    DEBUG_M_FSM0_STATE          : out std_logic_vector(5 downto 0);

    -- The usual ports
    Reset                       :  in STD_LOGIC;
    Clk                         :  in STD_LOGIC
  );
end ZARLINK_CONNECT_INIT;

architecture Behavioral of ZARLINK_CONNECT_INIT is

  type STATE_TYPE is (
    SM_WAIT,
    ZC_FIRST_RESET,
    ZC_INITIATE_RESET,
    -- Wait until the software reset is done
    ZC_SECOND_RESET,
    -- SET SPI SPEED
    ZC_SPI_SPEED_A,
    -- Let's disable the watchdog
    ZC_WATCHDOG_OFF,
    -- SET SPI SPEED (again) We don't trust the watchdog...
    ZC_SPI_SPEED_B,
    -- Wait for the radio ready flag
    ZC_WAIT_IRQ_RADIOREADY_A_REQ,
    ZC_WAIT_IRQ_RADIOREADY_A_WAIT,
    ZC_WAIT_IRQ_RADIOREADY_A_PAUSE,
    ZC_REG_MAC_MODUSER,
    ZC_REG_MAC_CHANNEL,
    ZC_REG_REG_TXBUFFER_BSIZE_A,
    ZC_REG_REG_RXBUFFER_BSIZE_A,
    ZC_REG_TXBUFF_MAXPACKSIZE_A,
    ZC_ALLOW_HOUSEKEEPINGMESSAGES,
    ZC_IMD_COPY_ID_A,
    ZC_IMD_COPY_ID_B,
    ZC_IMD_COPY_ID_C,
    ZC_IMD_ID_A,
    ZC_IMD_ID_B,
    ZC_IMD_ID_C,
    ZC_REG_MAC_INITCOM_A_IMD,
    -- Wait for the link ready flag
    ZC_WAIT_IRQ_LINKREADY_A_REQ,
    ZC_WAIT_IRQ_LINKREADY_A_WAIT,
    ZC_WAIT_IRQ_LINKREADY_A_PAUSE,
    -- Check connection
    CC_CHECKHARDWARE_REG,
    CC_CHECKHARDWARE_WAIT,
    CC_WAIT_IRQ_LINKREADY_A_REQ,
    CC_WAIT_IRQ_LINKREADY_A_WAIT,
    -- Check the fill status of the ingoing buffer,
    CB_INGOING_START,
    CB_INGOING_WAIT,
    -- Check the fill status of the outgoing buffer,
    CB_OUTGOING_START,
    CB_OUTGOING_WAIT,
    -- Get Command
    GC_START,
    GC_WAIT,
    -- Write Data
    WD_START,
    WD_DOIT
  );
  signal FSM_STATE : STATE_TYPE;

  signal ZDOR                   : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal Z_RD                   : STD_LOGIC := '0';
  signal Z_Address              : STD_LOGIC_VECTOR(6 downto 0) := (others => '0');
  signal Z_BM                   : STD_LOGIC := '0';
  signal Z_WE                   : STD_LOGIC := '0';
  signal Z_Done                 : STD_LOGIC := '0';

  -- 500 ms with 24 MHz clock
  constant Zarlink_WaitForReset : unsigned(23 downto 0) := to_unsigned(12000000,24);
  -- USE THIS FOR THE SIMULATION:
--  constant Zarlink_WaitForReset : unsigned(23 downto 0) := to_unsigned(12,24);
  signal Zarlink_ResetCounter   : unsigned (23 downto 0) := (others => '0');

  -- Time until the connection is declared failed!
  -- 2 mins with a 24 MHz clock = 2880000000
  constant Zarlink_ConnectionTimeout : unsigned (31 downto 0) := "10101011101010010101000000000000";
  signal Zarlink_ConnectionTimeoutCounter : unsigned (31 downto 0) := (others => '0');

  signal Zarlink_TimeOut        : std_logic := '0';

  -- Wait 50 ms until we retry it...
  constant Zarlink_Retry        : unsigned(16 downto 0) := to_unsigned(120000,17);

  -- USE THIS FOR THE SIMULATION:
--  constant Zarlink_Retry : unsigned(16 downto 0) := to_unsigned(7,17);
  signal Zarlink_Retry_Counter  : unsigned(16 downto 0) := (others => '0');

  -- Some other constants...
  -- for the reg_mac_moduser (standard is 4fsk tx/rx)
  constant Zarlink_Connection_Mode : STD_LOGIC_VECTOR(3 downto 0) := "1111";
  constant Zarlink_SPISpeedFlags: STD_LOGIC_VECTOR(1 downto 0) := "11";

  -- AA = 170 = "10101010"
  constant IMD_A                : STD_LOGIC_VECTOR(7 downto 0) := "10101010";
  constant IMD_B                : STD_LOGIC_VECTOR(7 downto 0) := "10101010";
  constant IMD_C                : STD_LOGIC_VECTOR(7 downto 0) := "10101010";

  constant USERDEFINEDBITS      : STD_LOGIC_VECTOR(2 downto 0) := "101";

  signal CheckConnection_OK_INTERNAL      : std_logic := '0';

  signal CheckIngoingBuffer_OK_INTERNAL   : std_logic := '0';
  signal CheckIngoingBuffer_SIZE_INTERNAL : STD_LOGIC_VECTOR (6 downto 0) := (others => '0');
  signal CheckOutgoingBuffer_OK_INTERNAL  : std_logic := '0';
  signal CheckOutgoingBuffer_SIZE_INTERNAL: STD_LOGIC_VECTOR (6 downto 0) := (others => '0');

  signal GetCommand_DATA_INTERNAL         : STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
  signal GetCommand_VALID_INTERNAL        : STD_LOGIC := '0';
  signal WaitSignal                       : STD_LOGIC := '0';

begin

  -- Resort the Zarlink_Data_Out_Internal in Zarlink-Fashion
  Zarlink_Data_Out <= ZDOR(0) & ZDOR(1) & ZDOR(2) & ZDOR(3) & ZDOR(4) & ZDOR(5) & ZDOR(6) & ZDOR(7);
  Zarlink_Adress <= Z_Address(0) & Z_Address(1) & Z_Address(2) & Z_Address(3)
    & Z_Address(4) & Z_Address(5) & Z_Address(6) & Z_RD;

  Zarlink_BlockMode         <= Z_BM;
  Zarlink_WriteEn           <= Z_WE;

  WaitSignal_ProcessOngoing <= WaitSignal;

  CheckConnection_OK        <= CheckConnection_OK_INTERNAL;

  CheckIngoingBuffer_OK     <= CheckIngoingBuffer_OK_INTERNAL;
  CheckIngoingBuffer_SIZE   <= CheckIngoingBuffer_SIZE_INTERNAL;
  CheckOutgoingBuffer_OK    <= CheckOutgoingBuffer_OK_INTERNAL;
  CheckOutgoingBuffer_SIZE  <= CheckOutgoingBuffer_SIZE_INTERNAL;

  GetCommand_DATA           <= GetCommand_DATA_INTERNAL;
  GetCommand_VALID          <= GetCommand_VALID_INTERNAL;

  FSM : process (Clk,Reset)
  begin
    if Reset = '1' then
      ZDOR <= (others => '0');
      Z_RD <= '0';
      Z_Address <= (others => '0');
      Z_BM <= '0';
      Z_WE <= '0';
      Z_Done <= '0';
      FSM_STATE <= SM_WAIT;

      WaitSignal <= '0';

      Zarlink_ResetCounter <= (others => '0');
      Zarlink_ConnectionTimeoutCounter <= (others => '0');
      Zarlink_Retry_Counter <= (others => '0');
      CheckConnection_OK_INTERNAL <= '0';

      CheckIngoingBuffer_OK_INTERNAL <= '0';
      CheckIngoingBuffer_SIZE_INTERNAL <= (others => '0');
      CheckOutgoingBuffer_OK_INTERNAL <= '0';
      CheckOutgoingBuffer_SIZE_INTERNAL <= (others => '0');

      GetCommand_DATA_INTERNAL <= (others => '0');
      GetCommand_VALID_INTERNAL <= '0';

      Zarlink_TimeOut <= '0';

    elsif rising_edge(Clk) then
      -- Default:
      ZDOR <= (others => '0');
      Z_RD <= '0';
      Z_Address <= (others => '0');
      Z_BM <= '0';
      Z_WE <= '0';
      Z_Done <= Z_Done;
      -- Fallback:
      FSM_STATE <= SM_WAIT;
      -- Outputs:

      Zarlink_ResetCounter <= (others => '0');

      Zarlink_ConnectionTimeoutCounter <= Zarlink_ConnectionTimeoutCounter + 1;
      Zarlink_Retry_Counter <= Zarlink_Retry_Counter + 1;
      CheckConnection_OK_INTERNAL <= CheckConnection_OK_INTERNAL;

      CheckIngoingBuffer_OK_INTERNAL <= CheckIngoingBuffer_OK_INTERNAL;
      CheckIngoingBuffer_SIZE_INTERNAL <= CheckIngoingBuffer_SIZE_INTERNAL;
      CheckOutgoingBuffer_OK_INTERNAL <= CheckOutgoingBuffer_OK_INTERNAL;
      CheckOutgoingBuffer_SIZE_INTERNAL <= CheckOutgoingBuffer_SIZE_INTERNAL;

      GetCommand_DATA_INTERNAL <= GetCommand_DATA_INTERNAL;
      GetCommand_VALID_INTERNAL <= GetCommand_VALID_INTERNAL;

      WaitSignal <= '1';

      case FSM_STATE is
        -- Lets wait until something happens...
        when SM_WAIT =>
          -- We are asked to do a Zarlink Connection Init!
          if ZarlinkConnectionInit_DO = '1' then
            FSM_STATE <= ZC_FIRST_RESET;
            CheckConnection_OK_INTERNAL <= '0';
          elsif CheckConnection_CHECK = '1' then
            FSM_STATE <= CC_CHECKHARDWARE_REG;
            CheckConnection_OK_INTERNAL <= '0';
          elsif CheckIngoingBuffer_CHECK = '1' then
            FSM_STATE <= CB_INGOING_START;
            CheckIngoingBuffer_OK_INTERNAL <= '0';
          elsif CheckOutgoingBuffer_CHECK = '1' then
            FSM_STATE <= CB_OUTGOING_START;
            CheckOutgoingBuffer_OK_INTERNAL <= '0';
          elsif GetCommand_GET = '1' then
            FSM_STATE <= GC_START;
            GetCommand_VALID_INTERNAL <= '0';
          elsif Zarlink_FIFO_WRITE = '1' then
            FSM_STATE <= WD_START;
          else
            FSM_STATE <= SM_WAIT;
            WaitSignal <= '0';
          end if;
          Zarlink_ConnectionTimeoutCounter <= (others => '0');
          Z_Done <= '0';
------------------------------------------------------------------------------

        -- Wait until the power on reset is done...
        -- (500ms wait, 24 Bit Counter)
        when ZC_FIRST_RESET =>
          if Zarlink_ResetCounter = Zarlink_WaitForReset then
            FSM_STATE <= ZC_INITIATE_RESET;
          else
            FSM_STATE <= ZC_FIRST_RESET;
          end if;
          Zarlink_ResetCounter <= Zarlink_ResetCounter + 1;
          Z_Done <= '0';

        -- INITIATE RESET
        -- (dec95 <= dec82)
        when ZC_INITIATE_RESET =>
          if (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_SECOND_RESET;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_INITIATE_RESET;
            Z_Address <= std_logic_vector(to_unsigned(95,7));
            ZDOR <= std_logic_vector(to_unsigned(82,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_INITIATE_RESET;
          end if;

        -- Wait until firmware initiated reset is done...
        -- (500ms wait, 24 Bit Counter)
        when ZC_SECOND_RESET =>
          if Zarlink_ResetCounter = Zarlink_WaitForReset then
            FSM_STATE <= ZC_SPI_SPEED_A;
          else
            FSM_STATE <= ZC_SECOND_RESET;
          end if;
          Zarlink_ResetCounter <= Zarlink_ResetCounter + 1;
          Zarlink_ConnectionTimeoutCounter <= (others => '0');

        -- Set SPI Speed
        -- (dec43 <= "01100000" (dec96))
        when ZC_SPI_SPEED_A =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_WATCHDOG_OFF;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_SPI_SPEED_A;
            Z_Address <= std_logic_vector(to_unsigned(43,7));
            ZDOR <= "0" & Zarlink_SPISpeedFlags & "00000";
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_SPI_SPEED_A;
          end if;

        -- Disable Watchdog
        -- (dec26 <= 0x0F)
        when ZC_WATCHDOG_OFF =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_SPI_SPEED_B;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_WATCHDOG_OFF;
            Z_Address <= std_logic_vector(to_unsigned(26,7));
            ZDOR <= std_logic_vector(to_unsigned(15,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_WATCHDOG_OFF;
          end if;

        -- Set SPI Speed
        -- (dec43 <= "01100000")
        when ZC_SPI_SPEED_B =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_REQ;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_SPI_SPEED_B;
            Z_Address <= std_logic_vector(to_unsigned(43,7));
            ZDOR <= "0" & Zarlink_SPISpeedFlags & "00000";
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_SPI_SPEED_B;
          end if;

        -- (dec50 => [3])
        when ZC_WAIT_IRQ_RADIOREADY_A_REQ =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_REQ;
            Z_Address <= std_logic_vector(to_unsigned(50,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_REQ;
          end if;

        -- (dec50 => [3])
        when ZC_WAIT_IRQ_RADIOREADY_A_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Valid_In = '0' then
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_WAIT;
          elsif (Zarlink_Data_In(3) = '1') AND (Zarlink_Valid_In = '1') then
            -- Continue
            FSM_STATE <= ZC_REG_MAC_MODUSER;
          else
            -- Return
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_PAUSE;
          end if;
          Z_Done <= '0';
          Zarlink_Retry_Counter <= (others => '0');

        -- Waiting loop
        when ZC_WAIT_IRQ_RADIOREADY_A_PAUSE =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Retry_Counter = Zarlink_Retry then
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_REQ;
          else
            FSM_STATE <= ZC_WAIT_IRQ_RADIOREADY_A_PAUSE;
          end if;
          Z_Done <= '0';

        -- (dec23 <= "00001111")
        -- User Bits = "0000"
        -- RX 4FSK = "11"
        -- TX 4FSK = "11"
        -- = dec15
        when ZC_REG_MAC_MODUSER =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_REG_MAC_CHANNEL;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_REG_MAC_MODUSER;
            Z_Address <= std_logic_vector(to_unsigned(23,7));
            ZDOR <= "0" & USERDEFINEDBITS & Zarlink_Connection_Mode;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_REG_MAC_MODUSER;
          end if;

        -- (dec25 <= dec5)
        when ZC_REG_MAC_CHANNEL =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_REG_REG_TXBUFFER_BSIZE_A;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_REG_MAC_CHANNEL;
            Z_Address <= std_logic_vector(to_unsigned(25,7));
            ZDOR <= std_logic_vector(to_unsigned(5,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_REG_MAC_CHANNEL;
          end if;

        -- (dec4 <= dec14)
        when ZC_REG_REG_TXBUFFER_BSIZE_A =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_REG_REG_RXBUFFER_BSIZE_A;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_REG_REG_TXBUFFER_BSIZE_A;
            Z_Address <= std_logic_vector(to_unsigned(4,7));
            ZDOR <= std_logic_vector(to_unsigned(14,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_REG_REG_TXBUFFER_BSIZE_A;
          end if;

        -- (dec2 <= dec30)
        when ZC_REG_REG_RXBUFFER_BSIZE_A =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_REG_TXBUFF_MAXPACKSIZE_A;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_REG_REG_RXBUFFER_BSIZE_A;
            Z_Address <= std_logic_vector(to_unsigned(2,7));
            ZDOR <= std_logic_vector(to_unsigned(30,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_REG_REG_RXBUFFER_BSIZE_A;
          end if;

        -- (dec5 <= dec 31)
        when ZC_REG_TXBUFF_MAXPACKSIZE_A =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_ALLOW_HOUSEKEEPINGMESSAGES;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_REG_TXBUFF_MAXPACKSIZE_A;
            Z_Address <= std_logic_vector(to_unsigned(5,7));
            ZDOR <= std_logic_vector(to_unsigned(31,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_REG_TXBUFF_MAXPACKSIZE_A;
          end if;

        -- (dec34 <= dec 2)
        when ZC_ALLOW_HOUSEKEEPINGMESSAGES =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_IMD_COPY_ID_A;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_ALLOW_HOUSEKEEPINGMESSAGES;
            Z_Address <= std_logic_vector(to_unsigned(34,7));
            ZDOR <= std_logic_vector(to_unsigned(2,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_ALLOW_HOUSEKEEPINGMESSAGES;
          end if;


        -- (dec66 <= IMD_A)
        when ZC_IMD_COPY_ID_A =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_IMD_COPY_ID_B;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_IMD_COPY_ID_A;
            Z_Address <= std_logic_vector(to_unsigned(66,7));
            ZDOR <= IMD_A;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_IMD_COPY_ID_A;
          end if;

        -- (dec67 <= IMD_B)
        when ZC_IMD_COPY_ID_B =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_IMD_COPY_ID_C;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_IMD_COPY_ID_B;
            Z_Address <= std_logic_vector(to_unsigned(67,7));
            ZDOR <= IMD_B;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_IMD_COPY_ID_B;
          end if;

        -- (dec68 <= IMD_C)
        when ZC_IMD_COPY_ID_C =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_IMD_ID_A;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_IMD_COPY_ID_C;
            Z_Address <= std_logic_vector(to_unsigned(68,7));
            ZDOR <= IMD_C;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_IMD_COPY_ID_C;
          end if;

        -- (dec19 <= IMD_A)
        when ZC_IMD_ID_A =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_IMD_ID_B;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_IMD_ID_A;
            Z_Address <= std_logic_vector(to_unsigned(19,7));
            ZDOR <= IMD_A;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_IMD_ID_A;
          end if;

        -- (dec20 <= IMD_B)
        when ZC_IMD_ID_B =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_IMD_ID_C;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_IMD_ID_B;
            Z_Address <= std_logic_vector(to_unsigned(20,7));
            ZDOR <= IMD_B;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_IMD_ID_B;
          end if;

        -- (dec21 <= IMD_C)
        when ZC_IMD_ID_C =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_REG_MAC_INITCOM_A_IMD;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_IMD_ID_C;
            Z_Address <= std_logic_vector(to_unsigned(21,7));
            ZDOR <= IMD_C;
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_IMD_ID_C;
          end if;

        -- (dec42 <= dec0)
        when ZC_REG_MAC_INITCOM_A_IMD =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_REQ;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_REG_MAC_INITCOM_A_IMD;
            Z_Address <= std_logic_vector(to_unsigned(42,7));
            ZDOR <= std_logic_vector(to_unsigned(0,8));
            Z_RD <= '0';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_REG_MAC_INITCOM_A_IMD;
          end if;


        -- (dec50 => [2])
        when ZC_WAIT_IRQ_LINKREADY_A_REQ =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_REQ;
            Z_Address <= std_logic_vector(to_unsigned(50,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_REQ;
          end if;

        -- (dec50 => [2])
        when ZC_WAIT_IRQ_LINKREADY_A_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Valid_In = '0' then
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_WAIT;
          elsif (Zarlink_Data_In(2) = '1') AND (Zarlink_Valid_In = '1') then
            -- Continue
            FSM_STATE <= SM_WAIT;
          else
            -- Return
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_PAUSE;
          end if;
          Z_Done <= '0';
          Zarlink_Retry_Counter <= (others => '0');

        -- Waiting loop
        when ZC_WAIT_IRQ_LINKREADY_A_PAUSE =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Retry_Counter = Zarlink_Retry then
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_REQ;
          else
            FSM_STATE <= ZC_WAIT_IRQ_LINKREADY_A_PAUSE;
          end if;
          Z_Done <= '0';

------------------------------------------------------------------------------

        -- Check ZL INIT!!!!
        -- (dec23 => [6:4])
        when CC_CHECKHARDWARE_REG =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= CC_CHECKHARDWARE_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= CC_CHECKHARDWARE_REG;
            Z_Address <= std_logic_vector(to_unsigned(23,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= CC_CHECKHARDWARE_REG;
          end if;

        -- (dec50 => [2])
        when CC_CHECKHARDWARE_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Valid_In = '0' then
            FSM_STATE <= CC_CHECKHARDWARE_WAIT;
          elsif (Zarlink_Data_In(6 downto 4) = USERDEFINEDBITS) AND (Zarlink_Valid_In = '1') then
            -- Continue
            FSM_STATE <= CC_WAIT_IRQ_LINKREADY_A_REQ;
          else
            FSM_STATE <= SM_WAIT;
          end if;
          Z_Done <= '0';

        -- Check Connection!!!!
        -- (dec50 => [2])
        when CC_WAIT_IRQ_LINKREADY_A_REQ =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= CC_WAIT_IRQ_LINKREADY_A_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= CC_WAIT_IRQ_LINKREADY_A_REQ;
            Z_Address <= std_logic_vector(to_unsigned(50,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= CC_WAIT_IRQ_LINKREADY_A_REQ;
          end if;

        -- (dec50 => [2])
        when CC_WAIT_IRQ_LINKREADY_A_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Valid_In = '0' then
            FSM_STATE <= CC_WAIT_IRQ_LINKREADY_A_WAIT;
          elsif (Zarlink_Data_In(2) = '1') AND (Zarlink_Valid_In = '1') then
            -- Continue
            FSM_STATE <= SM_WAIT;
            CheckConnection_OK_INTERNAL <= '1';
          else
            FSM_STATE <= SM_WAIT;
            CheckConnection_OK_INTERNAL <= '0';
          end if;
          Z_Done <= '0';

------------------------------------------------------------------------------

        -- (dec1 => CheckIngoingBuffer_SIZE_INTERNAL)
        when CB_INGOING_START =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= CB_INGOING_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= CB_INGOING_START;
            Z_Address <= std_logic_vector(to_unsigned(1,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= CB_INGOING_START;
          end if;

        -- (dec1 => CheckIngoingBuffer_SIZE_INTERNAL)
        when CB_INGOING_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Valid_In = '0' then
            FSM_STATE <= CB_INGOING_WAIT;
          elsif  Zarlink_Valid_In = '1' then
            -- Continue
            FSM_STATE <= SM_WAIT;
            CheckIngoingBuffer_OK_INTERNAL <= '1';
            CheckIngoingBuffer_SIZE_INTERNAL <= Zarlink_Data_In(6 downto 0);
          else
            FSM_STATE <= SM_WAIT;
            CheckIngoingBuffer_OK_INTERNAL <= '0';
          end if;
          Z_Done <= '0';

------------------------------------------------------------------------------

        -- (dec3 => CheckOutgoingBuffer_SIZE_INTERNAL)
        when CB_OUTGOING_START =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= CB_OUTGOING_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= CB_OUTGOING_START;
            Z_Address <= std_logic_vector(to_unsigned(3,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '0';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= CB_OUTGOING_START;
          end if;

        -- (dec3 => CheckOutgoingBuffer_SIZE_INTERNAL)
        when CB_OUTGOING_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Valid_In = '0' then
            FSM_STATE <= CB_OUTGOING_WAIT;
          elsif  Zarlink_Valid_In = '1' then
            -- Continue
            FSM_STATE <= SM_WAIT;
            CheckOutgoingBuffer_OK_INTERNAL <= '1';
            CheckOutgoingBuffer_SIZE_INTERNAL <= Zarlink_Data_In(6 downto 0);
          else
            FSM_STATE <= SM_WAIT;
            CheckOutgoingBuffer_OK_INTERNAL <= '0';
          end if;
          Z_Done <= '0';
------------------------------------------------------------------------------
        -- (45 => GetCommand_DATA_INTERNAL)
        when GC_START =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= GC_WAIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= GC_START;
            Z_Address <= std_logic_vector(to_unsigned(45,7));
            ZDOR <= (others => '0');
            Z_RD <= '1';
            Z_BM <= '1';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= GC_START;
          end if;

        -- (45 => GetCommand_DATA_INTERNAL)
        when GC_WAIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
            GetCommand_VALID_INTERNAL <= '0';
          elsif Zarlink_Wait = '0' then
            FSM_STATE <= SM_WAIT;
            GetCommand_VALID_INTERNAL <= '0';
          elsif Zarlink_Wait = '1' then
            -- Continue
            FSM_STATE <= GC_WAIT;
            GetCommand_VALID_INTERNAL <= Zarlink_Valid_In;
            GetCommand_DATA_INTERNAL <= Zarlink_Data_In;
          else
            FSM_STATE <= SM_WAIT;
            GetCommand_VALID_INTERNAL <= '0';
          end if;
          Z_Done <= '0';

------------------------------------------------------------------------------

        -- (45 <= Zarlink_FIFO_DATA)
        when WD_START =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif (Zarlink_Wait = '1') AND (Z_Done = '1') then
            -- Message was accepted and now continue
            FSM_STATE <= WD_DOIT;
            Z_Done <= '0';
          elsif (Zarlink_Wait = '0') then
            -- Repeat the message until it was accepted
            FSM_STATE <= WD_START;
            Z_Address <= std_logic_vector(to_unsigned(45,7));
            ZDOR <= Zarlink_FIFO_DATA;
            Z_RD <= '0';
            Z_BM <= '1';
            Z_WE <= '1';
            Z_Done <= '1';
          else
            -- Wait until we can send our message
            FSM_STATE <= WD_START;
          end if;

        when WD_DOIT =>
          if Zarlink_TimeOut = '1' then
            -- It took too long for the connection. Let's end this!
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Wait = '0' then
            -- We are done. Let's go home
            FSM_STATE <= SM_WAIT;
          elsif Zarlink_Wait = '1' then
            -- Continue	with the next 8 bit
            FSM_STATE <= WD_DOIT;
            ZDOR <= Zarlink_FIFO_DATA;
          else
            -- if we are here then something
            -- is very very wrong...
            FSM_STATE <= SM_WAIT;
          end if;

        when others =>
          FSM_STATE <= SM_WAIT;
      end case;

-----------------------------------------

      -- Check if it took too long
      if Zarlink_ConnectionTimeoutCounter = Zarlink_ConnectionTimeout then
        Zarlink_TimeOut <= '1';
      elsif Zarlink_ConnectionTimeoutCounter = to_unsigned(0,32) then
        Zarlink_TimeOut <= '0';
      else
        Zarlink_TimeOut <= Zarlink_TimeOut;
      end if;

    end if;
  end process FSM;

  -- DEBUGGING FSM
  DEBUG_PROCESS : process (FSM_STATE)
  begin
    case FSM_STATE is
      when SM_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 0,6));
      when ZC_FIRST_RESET =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 1,6));
      when ZC_INITIATE_RESET =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 2,6));
      when ZC_SECOND_RESET =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 3,6));
      when ZC_SPI_SPEED_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 4,6));
      when ZC_WATCHDOG_OFF =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 5,6));
      when ZC_SPI_SPEED_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 6,6));
      when ZC_WAIT_IRQ_RADIOREADY_A_REQ =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 7,6));
      when ZC_WAIT_IRQ_RADIOREADY_A_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 8,6));
      when ZC_WAIT_IRQ_RADIOREADY_A_PAUSE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 9,6));
      when ZC_REG_MAC_MODUSER =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(10,6));
      when ZC_REG_MAC_CHANNEL =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(11,6));
      when ZC_REG_REG_TXBUFFER_BSIZE_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(12,6));
      when ZC_REG_REG_RXBUFFER_BSIZE_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(13,6));
      when ZC_REG_TXBUFF_MAXPACKSIZE_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(14,6));
      when ZC_ALLOW_HOUSEKEEPINGMESSAGES =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(15,6));
      when ZC_IMD_COPY_ID_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(16,6));
      when ZC_IMD_COPY_ID_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(17,6));
      when ZC_IMD_COPY_ID_C =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(18,6));
      when ZC_IMD_ID_A =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(19,6));
      when ZC_IMD_ID_B =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(20,6));
      when ZC_IMD_ID_C =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(21,6));
      when ZC_REG_MAC_INITCOM_A_IMD =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(22,6));
      when ZC_WAIT_IRQ_LINKREADY_A_REQ =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(23,6));
      when ZC_WAIT_IRQ_LINKREADY_A_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(24,6));
      when ZC_WAIT_IRQ_LINKREADY_A_PAUSE =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(25,6));
      when CC_CHECKHARDWARE_REG => 
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(26,6));
      when CC_CHECKHARDWARE_WAIT => 
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(27,6));
      when CC_WAIT_IRQ_LINKREADY_A_REQ =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(28,6));
      when CC_WAIT_IRQ_LINKREADY_A_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(29,6));
      when CB_INGOING_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(30,6));
      when CB_INGOING_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(31,6));
      when CB_OUTGOING_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(32,6));
      when CB_OUTGOING_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(33,6));
      when GC_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(34,6));
      when GC_WAIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(35,6));
      when WD_START =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(36,6));
      when WD_DOIT =>
        DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned(37,6));
      when others =>
        DEBUG_M_FSM0_STATE <= "111111";
    end case;
  end process DEBUG_PROCESS;

end architecture Behavioral;
