library IEEE;
library UNISIM;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use UNISIM.VComponents.all;


entity TiggerIn_MuxControl is
    Port ( 

		TriggerIn_Address 		: out STD_LOGIC_VECTOR(4 downto 0);
		TriggerIn_Data 			: in STD_LOGIC;
	 
		Wait_TriggerIn 			: in STD_LOGIC_VECTOR (16 downto 0);

		EXTERN_PlusFullPacket_Trigger : out STD_LOGIC;
		FIFO_wr_en_External		: out STD_LOGIC;
		FIFO_din_External			: out STD_LOGIC_VECTOR (16 downto 0);
		FIFO_FULL 					: in STD_LOGIC;

		TimeStamp 					: in STD_LOGIC_VECTOR (63 downto 0);			
		PollingIsEnabled			: in STD_LOGIC;
		
		-- DEBUG SECTION
		DEBUG_M_FSM0_STATE		: out STD_LOGIC_VECTOR(1 downto 0);
						
		
		Clk 							: in  STD_LOGIC;
		Reset 						: in  STD_LOGIC
	);
end TiggerIn_MuxControl;

architecture Behavioral of TiggerIn_MuxControl is

	signal Mux_Counter : unsigned (4 downto 0) := (others => '0');
	signal TriggerChannelsIn_Copy : STD_LOGIC_VECTOR (31 downto 0) := (others => '0');
	signal TriggerChannelsIn_Buffer : STD_LOGIC_VECTOR (31 downto 0) := (others => '0');

	type STATE_TYPE_COLLECT is (
		-- Reset State 
		STATE_RESET,
		STATE_WAIT,
		STATE_COLLECT
	);
	signal FSM_STATE_COLLECT : STATE_TYPE_COLLECT;
	

	type STATE_TYPE is (
		STATE_RESET,
		STATE_IDLE,
		STATE_PREPARE_HEADER,
		STATE_PREPARE_LENGTH,
		STATE_PREPARE_RESPONSEID,
		STATE_PREPARE_TIMESTAMP_64_A,
		STATE_PREPARE_TIMESTAMP_64_B,
		STATE_PREPARE_TIMESTAMP_64_C,
		STATE_PREPARE_TIMESTAMP_64_D,
		STATE_TRIGGER_A,
		STATE_TRIGGER_B,
		STATE_COOLDOWN,
      STATE_NETWORK_COMMUNICATE_WAIT_A,
      STATE_NETWORK_COMMUNICATE_WAIT_B,
      STATE_NETWORK_COMMUNICATE_WAIT_C,
      STATE_NETWORK_COMMUNICATE_CRC00,
      STATE_NETWORK_COMMUNICATE_CRC01,
      STATE_NETWORK_COMMUNICATE_CRC10,
      STATE_NETWORK_COMMUNICATE_CRC11
	);
	signal FSM_STATE : STATE_TYPE;
	
	signal Wait_Counter : unsigned (16 downto 0) := (others => '0');

   signal CRC_Reset_Internal : STD_LOGIC := '0';
   signal CRC_01_Copy : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal CRC_10_Copy : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal CRC_11_Copy : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');

   signal CRC_00 : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal CRC_01 : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal CRC_10 : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
   signal CRC_11 : STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
	
	-- 0x55AA
	constant NET_HEADER : STD_LOGIC_VECTOR(15 downto 0) := std_logic_vector(to_unsigned(21930,16));
	-- 0x8004
	constant NET_RESPONSEID : STD_LOGIC_VECTOR(15 downto 0) := std_logic_vector(to_unsigned(32772,16));
	-- Not constant because the length can change 
	constant NET_LENGTH : STD_LOGIC_VECTOR(15 downto 0) := std_logic_vector(to_unsigned(26,16));

	signal TimeStamp_Base 			: STD_LOGIC_VECTOR (63 downto 0) := (others => '0');
	signal TriggerChannelsIn_Transmit : STD_LOGIC_VECTOR (31 downto 0) := (others => '0');
	
	
	signal FIFO_wr_en 				: STD_LOGIC := '0';
	signal FIFO_din 					: STD_LOGIC_VECTOR (16 downto 0) := (others => '0');
		
begin
	FIFO_wr_en_external <= FIFO_wr_en;
	FIFO_din_external <= FIFO_din;
	
	FSM : process (Reset, Clk)
	begin
		if Reset = '1' then
			FSM_STATE <= STATE_RESET;
			FIFO_wr_en <= '0';
			FIFO_din <= (others => '0');	

			TimeStamp_Base <= (others => '0');
			TriggerChannelsIn_Transmit <= (others => '0');

         CRC_01_Copy <= (others => '0');
         CRC_10_Copy <= (others => '0');
         CRC_11_Copy <= (others => '0');
         CRC_Reset_Internal <= '0';

		elsif rising_edge(Clk) then	

			-- Default:
			FSM_STATE <= STATE_RESET;
			FIFO_wr_en <= '0';
			FIFO_din <= (others => '0');	

			TimeStamp_Base <= TimeStamp_Base;
			TriggerChannelsIn_Transmit <= TriggerChannelsIn_Transmit;

         CRC_Reset_Internal <= '0';
         CRC_01_Copy <= (others => '0');
         CRC_10_Copy <= (others => '0');
         CRC_11_Copy <= (others => '0');			
			
			EXTERN_PlusFullPacket_Trigger <= '0';

			case FSM_STATE is
				when STATE_RESET =>
					FSM_STATE <= STATE_IDLE;			
			
				when STATE_IDLE =>
					if (TriggerChannelsIn_Transmit = TriggerChannelsIn_Buffer) OR (PollingIsEnabled = '0') then 
						FSM_STATE <= STATE_IDLE;
					else
						FSM_STATE <= STATE_PREPARE_HEADER;
					end if;

					TriggerChannelsIn_Transmit <= TriggerChannelsIn_Buffer;
					TimeStamp_Base <= TimeStamp;
			
				when STATE_PREPARE_HEADER => 
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_HEADER;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & NET_HEADER;				
						FSM_STATE <= STATE_PREPARE_LENGTH;
               end if;
					
				when STATE_PREPARE_LENGTH =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_LENGTH;  
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & NET_LENGTH;
						FSM_STATE <= STATE_PREPARE_RESPONSEID;					
               end if;

				when STATE_PREPARE_RESPONSEID => 
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_RESPONSEID;					
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & NET_RESPONSEID;
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_A;
               end if;

				-- 64 Bit Timestamp 	
				when STATE_PREPARE_TIMESTAMP_64_A =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_A;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & TimeStamp_Base(63 downto 48);
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_B;
					end if;

				when STATE_PREPARE_TIMESTAMP_64_B =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_B;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & TimeStamp_Base(47 downto 32);
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_C;
					end if;

				when STATE_PREPARE_TIMESTAMP_64_C =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_C;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & TimeStamp_Base(31 downto 16);
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_D;
					end if;

				when STATE_PREPARE_TIMESTAMP_64_D =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_PREPARE_TIMESTAMP_64_D;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & TimeStamp_Base(15 downto 0);
						FSM_STATE <= STATE_TRIGGER_A;
					end if;

				when STATE_TRIGGER_A =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_TRIGGER_A;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & TriggerChannelsIn_Transmit(31 downto 16);
						FSM_STATE <= STATE_TRIGGER_B;
					end if;

				when STATE_TRIGGER_B =>
					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_TRIGGER_B;
               else
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & TriggerChannelsIn_Transmit(15 downto 0);
						FSM_STATE <= STATE_COOLDOWN;
					end if;

				when STATE_COOLDOWN =>
					FSM_STATE <= STATE_NETWORK_COMMUNICATE_WAIT_A;

            -- Wait for the CRC - Values I.
            when STATE_NETWORK_COMMUNICATE_WAIT_A =>
					FSM_STATE <= STATE_NETWORK_COMMUNICATE_WAIT_B;

            -- Wait for the CRC - Values II.
            when STATE_NETWORK_COMMUNICATE_WAIT_B =>
					FSM_STATE <= STATE_NETWORK_COMMUNICATE_WAIT_C;

            -- Wait for the CRC - Values III.
            when STATE_NETWORK_COMMUNICATE_WAIT_C =>
					FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC00;

            -- Communicate with the network (CRC 00)
            when STATE_NETWORK_COMMUNICATE_CRC00 =>
					CRC_01_Copy <= CRC_01;
					CRC_10_Copy <= CRC_10;
					CRC_11_Copy <= CRC_11;

					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC00;
					else
						FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC01;
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & CRC_00;				
               end if;

            -- Communicate with the network (CRC 01)
            when STATE_NETWORK_COMMUNICATE_CRC01 =>
					CRC_01_Copy <= CRC_01_Copy;
					CRC_10_Copy <= CRC_10_Copy;
					CRC_11_Copy <= CRC_11_Copy;

					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC01;
					else
						FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC10;
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & CRC_01_Copy;				
					end if;

             -- Communicate with the network (CRC 10)
             when STATE_NETWORK_COMMUNICATE_CRC10 =>
					CRC_10_Copy <= CRC_10_Copy;
					CRC_11_Copy <= CRC_11_Copy;

					if FIFO_FULL = '1' then
						FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC10;
					else
						FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC11;
						FIFO_wr_en <= '1';
						FIFO_din <= "0" & CRC_10_Copy;				
					end if;

              -- Communicate with the network (CRC 11)
              when STATE_NETWORK_COMMUNICATE_CRC11 =>
						CRC_11_Copy <= CRC_11_Copy;

						if FIFO_FULL = '1' then
							FSM_STATE <= STATE_NETWORK_COMMUNICATE_CRC11;
						else
							FSM_STATE <= STATE_IDLE;
							FIFO_wr_en <= '1';
							FIFO_din <= "1" & CRC_11_Copy;
							EXTERN_PlusFullPacket_Trigger	<= '1';
						end if;
					when others =>
						FSM_STATE <= STATE_RESET;
			end case;
		end if;
	end process FSM;
	
	TriggerIn_Address <= std_logic_vector(Mux_Counter);
	
	MuxFSM : process (Reset, Clk)
	begin
		
		if Reset = '1' then
			
			Mux_Counter <= (others => '0');	
			TriggerChannelsIn_Copy <= (others => '0');	
			TriggerChannelsIn_Buffer <= (others => '0');	
			Wait_Counter <= (others => '0');	
			
			FSM_STATE_COLLECT <= STATE_RESET;
			
		elsif rising_edge(Clk) then
			TriggerChannelsIn_Copy <= TriggerChannelsIn_Copy;
			TriggerChannelsIn_Buffer <= TriggerChannelsIn_Buffer;

			Mux_Counter <= Mux_Counter;
			Wait_Counter <= (others => '0');	
			
			case FSM_STATE_COLLECT is
				
				when STATE_RESET =>
					FSM_STATE_COLLECT <= STATE_WAIT;
				
				when STATE_WAIT => 
					Wait_Counter <= Wait_Counter + 1;
					
					if Wait_Counter < unsigned(Wait_TriggerIn) then
						FSM_STATE_COLLECT <= STATE_WAIT;
					else
						FSM_STATE_COLLECT <= STATE_COLLECT;
					end if;
				
				when STATE_COLLECT => 
					if std_logic_vector(Mux_Counter) = "11111" then
						TriggerChannelsIn_Buffer <= TriggerIn_Data & TriggerChannelsIn_Copy(31 downto 1);
						TriggerChannelsIn_Copy <= (others => '0');	
						Mux_Counter <= (others => '0');
					else
						TriggerChannelsIn_Copy <= TriggerIn_Data & TriggerChannelsIn_Copy(31 downto 1); 
						Mux_Counter <= Mux_Counter + 1;
					end if;
					FSM_STATE_COLLECT <= STATE_WAIT;
					
				when others => 
					FSM_STATE_COLLECT <= STATE_RESET;	
			end case;
			
		end if;
	
	end process MuxFSM;

	-- DEBUGGING FSM
	DEBUG_PROCESS : process(FSM_STATE_COLLECT)
	begin
		case FSM_STATE_COLLECT is 
			when STATE_RESET => 
				DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 0,2));
			when STATE_WAIT =>
				DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 1,2));
			when STATE_COLLECT =>
				DEBUG_M_FSM0_STATE <= std_logic_vector(to_unsigned( 2,2));
			when others => 
				DEBUG_M_FSM0_STATE <= (others => '1');
		end case;
	end process DEBUG_PROCESS;

    MyCRC : entity work.CRC_Checker
    Port MAP(
          Network_Data => FIFO_din(15 downto 0),
          Network_Valid => FIFO_wr_en,
          CRC_Reset => CRC_Reset_INTERNAL,
          CRC_00 => CRC_00,
          CRC_01 => CRC_01,
          CRC_10 => CRC_10,
          CRC_11 => CRC_11,
          Reset => Reset,
          Clk => Clk
    );
	 
end architecture Behavioral;

